#!/usr/bin/env python
# Copyright (c) 2005-2007 XenSource, Inc. 2008 Citrix Systems.
# All use and distribution of this copyrighted material is governed by 
# and subject to terms and conditions as licensed by XenSource, Inc. All 
# other rights reserved.  Xen, XenSource and XenEnterprise are either 
# registered trademarks or trademarks of XenSource Inc. in the United 
# States and/or other countries.
#
#
# CSLGHandler: A class to handle both internal and external CSLG requests 
#

""" The cslg SR implementation """

import SR, VDI, ISCSISR, HBASR, SRCommand, util
import xml.dom.minidom
import xs_errors
import socket
from socket import SOCK_DGRAM, AF_INET
import scsiutil
import syslog
import fcntl, struct
import os, time, copy, glob
import zlutil
import xmlrpclib
import re
from iSL import sl_adt_netapp
from iSL import sl_adt_smis
from iSL import sl_adt_delleql
from iSL import sl_adt_remote

TRACEON = 1

SYSTEM_SUPPORTS_PROVISION_FULL         = 0x00000100
SYSTEM_SUPPORTS_PROVISION_THIN         = 0x00000200
SYSTEM_SUPPORTS_MAPPING                = 0x00000400
SYSTEM_SUPPORTS_MULTIPLE_STORAGE_POOLS = 0x00000800
SYSTEM_SUPPORTS_LUN_GROUPING           = 0x00001000
SYSTEM_SUPPORTS_POOL_LEVEL_DEDUPLICATION = 0x00002000
SYSTEM_SUPPORTS_VOLUME_LEVEL_DEDUPLICATION = 0x00004000
SYSTEM_SUPPORTS_DIFF_SNAPSHOT          = 0x00010000
SYSTEM_SUPPORTS_PLEX_SNAPSHOT          = 0x00020000
SYSTEM_SUPPORTS_REMOTE_REPLICATION     = 0x00040000

SYSTEM_SUPPORTS_CLONE                  = 0x00080000
SYSTEM_SUPPORTS_RESIZE                 = 0x00100000
SYSTEM_REQUIRES_STORAGE_POOL_CLEANUP   = 0x00200000
SYSTEM_SUPPORTS_CUSTOM_ENDPOINTS       = 0x00400000
SYSTEM_SUPPORTS_OPTIMIZED_ISCSI_LOGIN  = 0x00800000
SYSTEM_SUPPORTS_LUN_PER_IQN            = 0x01000000
SYSTEM_SUPPORTS_INSTANT_CLONE          = 0x02000000
SYSTEM_SUPPORTS_CLONE_OF_SNAPSHOT      = 0x04000000
SYSTEM_REQUIRES_ISCSI_PRE_LOGIN           = 0x08000000
SYSTEM_SUPPORTS_SNAPSHOT_OF_SNAPSHOT   = 0x10000000
SYSTEM_SUPPORTS_CHAP                   = 0x20000000

SYSTEM_SUPPORTS_FIBRE_CHANNEL          = 0x00000004
SYSTEM_SUPPORTS_ISCSI                  = 0x00000002
SYSTEM_SUPPORTS_NFS                    = 0x00000008
SYSTEM_SUPPORTS_CIFS                   = 0x00000010

NODE_TYPE_UNKNOWN                      = 0x0000
NODE_TYPE_STORAGE_VOLUME               = 0x0001
NODE_TYPE_CIFS_SHARE                   = 0x0002
NODE_TYPE_NFS_SHARE                    = 0x0004

SNAPSHOT_TYPE_DIFF                     = 0x010000
SNAPSHOT_TYPE_PLEX                     = 0x020000
SNAPSHOT_TYPE_REMOTE_REPLICATION       = 0x040000
SNAPSHOT_TYPE_CLONE                    = 0x080000
SNAPSHOT_TYPE_IS_DEFAULT               = 0x100000
SNAPSHOT_TYPE_REMOTE_REPLICATION_SYNC  = 0x200000

SNAPSHOT_CAPABILITY_THIN_PROVISIONED_TARGET    = 0x0010
SNAPSHOT_CAPABILITY_REQ_PRE_PROVISIONED_TARGET = 0x0020
SNAPSHOT_CAPABILITY_AUTO_PROVISIONED_TARGET    = 0x0040
SNAPSHOT_CAPABILITY_POOL_SPECIFICATION         = 0x0080

SYSTEM_SUPPORTS_RAID_UNKNOWN           = 0x0000
SYSTEM_SUPPORTS_RAID0_CONCAT           = 0x0004
SYSTEM_SUPPORTS_RAID0                  = 0x0008
SYSTEM_SUPPORTS_RAID1                  = 0x0010
SYSTEM_SUPPORTS_RAID3                  = 0x0020
SYSTEM_SUPPORTS_RAID4                  = 0x0040
SYSTEM_SUPPORTS_RAID5                  = 0x0080
SYSTEM_SUPPORTS_RAID6                  = 0x0100
SYSTEM_SUPPORTS_RAID10                 = 0x0400
SYSTEM_SUPPORTS_RAID50                 = 0x1000

SYSTEM_DEFAULT_RAID_LEVEL = SYSTEM_SUPPORTS_RAID5 

SYSTEM_RAID_LEVEL_MAP = { \
                'RAID0'  : SYSTEM_SUPPORTS_RAID0, \
                'RAID1'  : SYSTEM_SUPPORTS_RAID1, \
                'RAID3'  : SYSTEM_SUPPORTS_RAID3, \
                'RAID4'  : SYSTEM_SUPPORTS_RAID4, \
                'RAID5'  : SYSTEM_SUPPORTS_RAID5, \
                'RAID6'  : SYSTEM_SUPPORTS_RAID6, \
                'RAID10' : SYSTEM_SUPPORTS_RAID10,\
                'RAID50' : SYSTEM_SUPPORTS_RAID50 }

SYSTEM_SUPPORTS_PROVISION_FULL         = 0x00000100
SYSTEM_SUPPORTS_PROVISION_THIN         = 0x00000200

SNAP_OP_CREATE                         = 0x02
SNAPSHOT_TYPE_CLONE                    = 0x080000
PROVISION_THICK                        = 0x000100
SNAP_STATE_NORMAL                      = 0x01

HOST_OS_TYPE_UNKNOWN                   = 0x001 
HOST_OS_TYPE_LINUX                     = 0x002 
HOST_OS_TYPE_WINDOWS                   = 0x003 
HOST_OS_TYPE_SOLARIS                   = 0x004 
HOST_OS_TYPE_AIX                       = 0x005 
HOST_OS_TYPE_HPUX                      = 0x006 
HOST_OS_TYPE_VMWARE                    = 0x007 
HOST_OS_TYPE_WINDOWS_2K8               = 0x008 

PROVISION_OPTION_NONE                  = 0x000000
PROVISION_OPTION_DEDUPLICATION         = 0x002000

ADAPTER_CALL_SUCCESS                   = 0

XML_HEADER                             = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
SS_LIST_TAG_NAME                       = "csl__storageSystemInfoList"
SNAP_LIST_TAG_NAME                     = "csl__snapshotMethodInfoList"
SP_LIST_TAG_NAME                       = "csl__storagePoolInfoList"
VOL_LIST_TAG_NAME                      = "csl__storageVolumeInfoList"
SA_LIST_TAG_NAME                       = "csl__storageAssignmentInfoList"
INITIATORS_LIST_TAG_NAME               = "csl__initiatorPortInfoList"
TARGETS_LIST_TAG_NAME                  = "csl__targetPortInfoList"
SVLUN_LIST_TAG_NAME                    = "csl__storageVolumeLunInfoList"
SS_TAG_NAME                            = "csl__storageSystemInfo"
SNAP_TAG_NAME                          = "csl__snapshotMethodInfo"
SP_TAG_NAME                            = "csl__storagePoolInfo"
VOL_TAG_NAME                           = "csl__storageVolumeInfo"
SA_TAG_NAME                            = "csl__storageAssignmentInfo"
INITIATORS_TAG_NAME                    = "csl__initiatorPortInfo"
TARGETS_TAG_NAME                       = "csl__targetPortInfo"
SVLUN_TAG_NAME                         = "csl__storageVolumeLunInfo"
SYS_CAP_TAG_NAME                       = "systemCapabilities"
PROT_SUPPORT_TAG_NAME                  = "protocolSupport"

FRIENDLY_NAME                          = "friendlyName"
DISPLAY_NAME                           = "displayName"
VENDOR                                 = "vendor"
MODEL                                  = "model"
SERIAL_NUM                             = "serialNum"
SSID                                   = "storageSystemId"

#SL Error Codes
SL_ERR={}
SL_ERR[6] = "Object was not found"
SL_ERR[27] = "Failed to collect storage information for server"

def getRaidLevelValue( raidLevel ):
    if raidLevel in SYSTEM_RAID_LEVEL_MAP:
        return SYSTEM_RAID_LEVEL_MAP[ raidLevel ]
    else:
        return SYSTEM_SUPPORTS_RAID_UNKNOWN


class CSLGHandler:    
    def __init__(self):
        return
    def getStorageSystemInfo(self, ssid):
        return
    
    def getStoragePoolInfo(self, ssid, pool):
        return
    
    def getStoragePoolInfoByName(self, ssid, poolName):
        return
    
    def createStoragePool(self, name, ssid, pool, size, provisioning_type, provision_options):
        return
    
    def deleteStoragePool(self, ssid, pool):
        return
    
    def createStorageNode(self, name, size, ssid, spid, raid_type, group_name, provisioning_type, provisioning_options):
        return
    
    def getStorageVolumeInfo(self, ssid, svid):
        return
    
    def getStorageVolumeInfoByName(self, ssid, svName):
        return

    def deleteStorageNode(self, svid):
        return

    def resizeStorageNode(self, ssid, svid,size):
        return
        
    def enumStorageSystemTargetPorts(self, ssid):
        return
    
    def assignStorageNode(self, svid, hostname, os_type, protocol):
        return

    def unassignStorageNode(self, ssid, svid, hostname, optflags, protocol):
        return
            
    def listStoragePools(self):
        return

    def listStorageSystems(self):
        return

    def listStorageVolumes(self):
        return

    def getSnapMirrorLunName(self, ssid, source_lun_name, dest_loc):
        return
    
class CSLGHandlerInternal(CSLGHandler):
    def __init__(self, srobj):
        """ Initialize for Internal Storage Adapter Calls """
        _trace("entry CSLGHandler.CSLGHandlerInternal.__init__()")

        CSLGHandler.__init__(self)
        _trace("adapter id: %s" %srobj.adapter)
        self.adapter_inf = None

        if srobj.adapter == "NEXENTA":
            self.adapter_inf = sl_adt_remote
            self.adapter = self.adapter_inf.CXSRemoteStorageAdapter()
        elif srobj.adapter == "NETAPP":
            self.adapter_inf = sl_adt_netapp
            self.adapter = self.adapter_inf.CXSSANetappStorageAdapter()
        elif srobj.adapter == "EMC_CLARIION" or srobj.adapter == "SMIS_STORAGE_SYSTEM":
            self.adapter_inf = sl_adt_smis
            self.adapter = self.adapter_inf.CXSSASmisStorageAdapter()
        elif srobj.adapter == "DELL_EQUALLOGIC":
            self.adapter_inf = sl_adt_delleql
            self.adapter = self.adapter_inf.CXSSADellEqlStorageAdapter()

        _trace("Debug: adapter id : %s"% srobj.adapter)
        if self.adapter_inf is not None:
            self.cred    = self._construct_adapter_credentials(srobj)
        self.session = util.get_localAPI_session()
        try:
            self.sr_ref  = srobj.sr_ref
        except:
            pass

    def _construct_adapter_credentials(self, srobj):
        """ construct adapter credentials object """
        _trace("entry CSLG._construct_adapter_credentials()")
        _trace("(adapterid:%s, ip:%s, username:%s, password:****)" % (srobj.adapter, srobj.cslg, srobj.cslguser) )

        try:
            if srobj.adapter == "NEXENTA":
                cred = {}
                cred['storageAdapterId'] = srobj.adapter
                cred['uuid'] = ""
                cred['friendlyName'] = ""
                cred['ipAddress'] = srobj.ip
                cred['portNumber'] = int(srobj.cslport)
                cred['username'] = srobj.username
                cred['password'] = srobj.password
                cred['context'] = "" 
                cred['ns'] = ""
            else:
                cred = self.adapter_inf.CXSSAManagementCredentials()
                cred.storageAdapterId = srobj.adapter
                cred.uuid = ""
                cred.friendlyName = ""
                cred.ipAddress = srobj.cslg
                cred.portNumber = int(srobj.cslport)
                cred.username = srobj.cslguser
                cred.password = srobj.cslgpassword
                cred.context = ""
                cred.ns = ""
                if srobj.adapter == "EMC_CLARIION" or srobj.adapter == "SMIS_STORAGE_SYSTEM":
                    cred.ns = "interop"
        except:
            pass

        return cred


    def getStorageAdapterInfo(self):
        """ Get Storage Adapter Info """
        _trace("entry CSLGHandler.CSLGHandlerInternal.getStorageAdapterInfo")

        sa_info =  self.adapter_inf.CXSSAStorageAdapterInfo()
        err = self.adapter.getStorageAdapterInfo(sa_info)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGProbe', \
                    opterr='FAILED: CSLGHandler.CSLGHandlerInternal.getStorageAdapterInfo(): ' + err.get_msg())
            return

        name = sa_info.name
        fndly_name = sa_info.friendlyName
        desc = sa_info.desc
        type = sa_info.type
        repType = sa_info.replicationType
        verMaj = sa_info.versionMajor
        verMin = sa_info.versionMinor

        return (name, fndly_name, desc, type, repType, verMaj, verMin)

    def getStorageSystemInfo(self, ssid):
        """ Get Storage System Info """
        _trace("entry CSLGHandler.CSLGHandlerInternal.getStorageSystemInfo()")
       
        ssInfo = self.adapter_inf.CXSSAStorageSystemInfo()
        err = self.adapter.getStorageSystemInfo(self.cred, ssid, ssInfo) 
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGSysInfo', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.getStorageSystemInfo(): ' + err.get_msg())
            return
        
        serial_number = ssInfo.serialNum
        capablities   = ssInfo.capabilities

        #allowed_protocols
        allowed_protocols = ""
        if capablities & SYSTEM_SUPPORTS_FIBRE_CHANNEL :
            allowed_protocols = allowed_protocols + "fc," 
        if capablities & SYSTEM_SUPPORTS_ISCSI :
            allowed_protocols = allowed_protocols + "iscsi," 
        if capablities & SYSTEM_SUPPORTS_NFS :
            allowed_protocols = allowed_protocols + "nfs," 
        if capablities & SYSTEM_SUPPORTS_CIFS :
            allowed_protocols = allowed_protocols + "cifs," 
        allowed_protocols = allowed_protocols[:-1]

        #req_pool_cleanup
        req_pool_cleanup = False
        if capablities & SYSTEM_REQUIRES_STORAGE_POOL_CLEANUP :
            req_pool_cleanup = True        

        #supports_clone
        supports_clone = False
        if capablities & SYSTEM_SUPPORTS_CLONE_OF_SNAPSHOT :
            supports_clone = True        

        #supports_resize 
        supports_resize = False
        if capablities & SYSTEM_SUPPORTS_RESIZE :
            supports_resize = True

        #support LUN per IQN
        supports_LUNIQN = False
        if capablities & SYSTEM_SUPPORTS_LUN_PER_IQN :
            supports_LUNIQN = True

        #support CHAP
        supports_CHAP = False
        if capablities & SYSTEM_SUPPORTS_CHAP :
            supports_CHAP = True

        return (serial_number, allowed_protocols, req_pool_cleanup, supports_resize, supports_clone, supports_LUNIQN, supports_CHAP)


    def getStoragePoolInfo(self, ssid, pool):
        """ Get Storage Pool Info """
        _trace("entry CSLGHandler.CSLGHandlerInternal.getStoragePoolInfo()")

        spInfo = self.adapter_inf.CXSSAStoragePoolInfo()
        err = self.adapter.getStoragePoolInfo(self.cred, ssid, pool, spInfo)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGPoolInfo', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.getStoragePoolInfo(): ' + err.get_msg())

        friendlyName = spInfo.friendlyName
    
        #supported_raid
        supported_raid = []
        raid_capablities = spInfo.supportedRaidTypes
        if raid_capablities & SYSTEM_SUPPORTS_RAID0_CONCAT :
            supported_raid.append("RAID0")  
        if raid_capablities & SYSTEM_SUPPORTS_RAID0 :
            supported_raid.append("RAID0")  
        if raid_capablities & SYSTEM_SUPPORTS_RAID1 :
            supported_raid.append("RAID1")  
        if raid_capablities & SYSTEM_SUPPORTS_RAID3 :
            supported_raid.append("RAID3")  
        if raid_capablities & SYSTEM_SUPPORTS_RAID4 :
            supported_raid.append("RAID4")  
        if raid_capablities & SYSTEM_SUPPORTS_RAID5 :
            supported_raid.append("RAID5")  
        if raid_capablities & SYSTEM_SUPPORTS_RAID6 :
            supported_raid.append("RAID6")  
        if raid_capablities & SYSTEM_SUPPORTS_RAID10 :
            supported_raid.append("RAID10")  
        if raid_capablities & SYSTEM_SUPPORTS_RAID50 :
            supported_raid.append("RAID50")  

        size = int(spInfo.sizeInMB)*1024*1024
        freeSpace = int(spInfo.freeSpaceInMB)*1024*1024
        availableFreeSpace = int(spInfo.availableFreeSpaceInMB)*1024*1024
        parentStoragePoolId = spInfo.parentPool

        # netapp has better info then the generic array which doesn't support
        # thin provisioning
        if ssid.find("NETAPP_") == 0 or ssid.find("Compellent__StorageCenter") == 0:
            physical_size = availableFreeSpace + size
            physical_utilisation = size - freeSpace
            virtual_allocation = size - freeSpace
        else:
            physical_size = size
            physical_utilisation = size - availableFreeSpace
            virtual_allocation = size - availableFreeSpace

        return( physical_size, physical_utilisation, virtual_allocation, parentStoragePoolId, friendlyName, supported_raid)
  

    def createStoragePool(self, name, ssid, pool, size, provisioning_type='', provision_options=''):
        """ Create Storage Pool """
        _trace("entry CSLGHandler.CSLGHandlerInternal.createStoragePool()")

        spInfo = self.adapter_inf.CXSSAStoragePoolInfo()
        prov_type =SYSTEM_SUPPORTS_PROVISION_FULL
        if provisioning_type == "THIN":
            prov_type = SYSTEM_SUPPORTS_PROVISION_THIN
        prov_options=PROVISION_OPTION_NONE
        if provision_options == "DEDUPLICATION":
            prov_options =PROVISION_OPTION_DEDUPLICATION
        err = self.adapter.createStoragePool(self.cred, name, ssid, pool, size, 0, prov_type, prov_options, spInfo)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGPoolCreate', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.createStoragePool(): ' + err.get_msg())
            return

        storagePoolId = spInfo.storagePoolId
        if storagePoolId == '':
            raise xs_errors.XenError('CSLGPoolCreate', \
                                 opterr='FAILED: CSLGHandler.CSLGHandlerInternal.createStoragePool()')

        return( storagePoolId)

    
    def deleteStoragePool(self, ssid, pool):
        """ Delete Storage Pool """
        _trace("entry CSLGHandler.CSLGHandlerInternal.deleteStoragePool()")

        err = self.adapter.deleteStoragePool(self.cred, ssid, pool)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGPoolDelete', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.deleteStoragePool(): ' + err.get_msg())
            return

    def createStorageNode(self, name, size, ssid, spid, sr_uuid = '', justxml = False, raid_type='', group_name='', provisioning_type='', provisioning_options=''):
        """ Create Storage Volume """
        _trace("entry CSLGHandler.CSLGHandlerInternal.createStorageNode()")

        svInfo = self.adapter_inf.CXSSAStorageVolumeInfo()
        size /= 1024*1024
        util.SMlog("from self.adapter.createStorageNode b4 call: %s, %s, %s, %s, %s, %s " % (name, ssid, spid, size, raid_type, group_name))
        prov_type =SYSTEM_SUPPORTS_PROVISION_FULL
        if provisioning_type == "THIN":
            prov_type = SYSTEM_SUPPORTS_PROVISION_THIN
        prov_options=PROVISION_OPTION_NONE
        if provisioning_options == "DEDUPLICATION":
            prov_options =PROVISION_OPTION_DEDUPLICATION

        raid_level = getRaidLevelValue( raid_type )
        if raid_level == SYSTEM_SUPPORTS_RAID_UNKNOWN:
            raid_level = SYSTEM_DEFAULT_RAID_LEVEL

        create_args = [name, ssid, spid, size, raid_level, prov_type, prov_options, group_name]
        util.SMlog( "adapter createStorageNode args: %s" % create_args )

        err = self.adapter.createStorageNode(self.cred, name, ssid, spid, size, raid_level, prov_type, prov_options, group_name, svInfo)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGAllocate', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.createStorageNode(): ' + err.get_msg())
            return

        storageVolumeId = svInfo.storageVolumeId
        lunIqn = svInfo.lunIqn
        if lunIqn == '':
            lunIqn = None

        util.SMlog("from self.adapter.createStorageNode with parameters: %s, %s " % (storageVolumeId, lunIqn))

        return( storageVolumeId, lunIqn)

        
    def getStorageVolumeInfo(self, ssid, svid):
        """ Get Storage Volume Info """
        _trace("entry CSLGHandler.CSLGHandlerInternal.getStorageVolumeInfo()")

        svInfo = self.adapter_inf.CXSSAStorageVolumeInfo()
        util.SMlog("from self.adapter.getStorageVolumeInfo with parameters: %s, %s " % (ssid, svid))
        err = self.adapter.getStorageVolumeInfo(self.cred, ssid, svid, svInfo)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGLunInfo', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.getStorageVolumeInfo(): ' + err.get_msg())
            _trace("FAILED: Error:%d, CSLGHandler.CSLGHandlerInternal.getStorageVolumeInfo()" % err)
            return

        #isSnapshot
        isSnapshot = "No"
        if svInfo.isSnapshot == 1:
            isSnapshot = "Yes"

        size          = int(svInfo.sizeInMB)*1024*1024
        usedSpace     = int(svInfo.usedSpaceInMB)*1024*1024
        displayName   = svInfo.friendlyName
        lunIqn        = svInfo.lunIqn

        return( size, usedSpace, displayName, lunIqn, isSnapshot)

    
    def deleteStorageNode(self, svid):
        """ Delete Storage Volume """
        _trace("entry CSLGHandler.CSLGHandlerInternal.deleteStorageNode()")

        err = self.adapter.deleteStorageNode(self.cred, svid)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGDeallocate', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.deleteStorageNode(): ' + err.get_msg())
            return

        return


    def resizeStorageNode(self, ssid, svid,size):
        """ Resize Storage Volume """
        _trace("entry CSLGHandler.CSLGHandlerInternal.resizeStorageNode()")

        svInfo = self.adapter_inf.CXSSAStorageVolumeInfo()
        err = self.adapter.resizeStorageNode(self.cred, ssid, svid, size, svInfo)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGResizeLun', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.resizeStorageNode(): ' + err.get_msg())
            return

        return

    
    def enumStorageSystemTargetPorts(self, ssid, LUNperIQN):
        """ Enumerate Storage Target Ports """
        _trace("entry CSLGHandler.CSLGHandlerInternal.enumStorageSystemTargetPorts(ssid:%s, LUNperIQN:%s)"%(ssid, LUNperIQN))

        TargetInfoList = self.adapter_inf.ListOfTargetPortInfo()
        filters = self.adapter_inf.ListOfStrings()
        err = self.adapter.enumerateTargetPorts(self.cred, ssid, 1, filters, TargetInfoList)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGTargetPorts', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.enumStorageSystemTargetPorts(): ' + err.get_msg())
            return

        targets = []
        ports   = []
        for targetCount in range(TargetInfoList.size()):
            if TargetInfoList[targetCount].protocolType == SYSTEM_SUPPORTS_ISCSI :
                ports.append(TargetInfoList[targetCount].portId)
                if not LUNperIQN:
                    if TargetInfoList[targetCount].portIqn not in targets:
                        targets.append(TargetInfoList[targetCount].portIqn)

        return (ports, targets)

        
    def assignStorageNode(self, ssid, svid, hostname, flags='', name='', protocol='', optflags = [], chap_user='', chap_pass=''):
        """ Assign Storage Node """
        _trace("entry CSLGHandler.CSLGHandlerInternal.assignStorageNode()")

        #extract IQN or WWN
        iqn = []
        hostwwn = ''
        items = optflags.split(" ")
        # mask password (if any) with *s before logging...
        # ...there isn't an inline if in python 2.4
        _trace("optflags: %s" % map(
            lambda x:
                x.startswith('chap-pass=') and
                    ''.join([x.split('=')[0], '=', '*' * len(x.split('=')[1])])
                or x,
            items))
        for item in items:
            if item.find("host") != -1:
                cmd = item.split("=")
                _trace("hostwwn: %s"% cmd)
                if len(cmd) > 1:
                    hostwwn += cmd[1]
            elif item.find("iqn") != -1 or  item.find("wwn") != -1:
                cmd = item.split("=")
                _trace("iqn/wwn: %s"% cmd)
                if len(cmd) > 1:
                    if cmd[1].find(",") != -1:
                        iqn = cmd[1].split(",")
                    else:
                        iqn.append(cmd[1])
        _trace("Extracted IQN: %s" % iqn)
        if len( hostwwn ) > 0:
            _trace("Extracted HostWWN: %s" % hostwwn)

        #extract protocol
        connection = SYSTEM_SUPPORTS_ISCSI
        if protocol == "FC":
            connection = SYSTEM_SUPPORTS_FIBRE_CHANNEL

        InitiatorList = self.adapter_inf.ListOfStrings()
        _trace( "  iqn sz = %s" % str( len( iqn ) ), priority=syslog.LOG_DEBUG )
        for value in iqn:
            _trace( "    value  = %s" % value, priority=syslog.LOG_DEBUG )
            if len( hostwwn ) > 0:
                value = hostwwn + ":" + value
                _trace( "    value+ = %s" % value, priority=syslog.LOG_DEBUG )
            _trace( "    append: %s" % value, priority=syslog.LOG_DEBUG )
            InitiatorList.append(value)

        SAList = self.adapter_inf.ListOfStorageAssignmentInfo()
        _trace("ssid=%s, svid=%s, hostname=%s protocol=%s" % (ssid, svid, hostname, protocol))

        err = self.adapter.assignStorageNode(self.cred, ssid, svid, hostname, InitiatorList, "", "", chap_user, chap_pass, HOST_OS_TYPE_LINUX, connection, SAList)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGAssign', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.assignStorageNode(): ' + err.get_msg())
            return

        salist_size = SAList.size()
        util.SMlog( "CSLGhandler.assignStorageNode: Returned SAList from self.adapter.assignStorageNode, size = %d" % (salist_size) )

        lun = ""
        lunid = ""
        if SAList.size() > 0 :
            for listCount in range(SAList[0].storageVolumeLunInfoList.size()):
                if SAList[0].storageVolumeLunInfoList[listCount].storageVolumeId == svid:
                    lun = SAList[0].storageVolumeLunInfoList[listCount].lunIqn
                    lunid = SAList[0].storageVolumeLunInfoList[listCount].lun
                    break
            if lunid == "":
                util.SMlog( "CSLGhandler.assignStorageNode: storageVolumeLunInfoList doesn't contain %s info"%svid )
        else:
            util.SMlog( "CSLGhandler.assignStorageNode: EMPTY SAList back from self.adapter.assignStorageNode" )

        #respond back in xml format
        xml_resp=''
        if SAList.size() > 0 :
            xml_resp = XML_HEADER
            xml_resp += create_xml_start_tag(SA_LIST_TAG_NAME)
            for listCount in range(SAList.size()):
                _trace("listCount is %d : friendlyName is %s" % (listCount, SAList[listCount].friendlyName))
                xml_resp += create_xml_start_tag(SA_TAG_NAME)
                xml_resp += create_xml_value_tag(DISPLAY_NAME, SAList[listCount].friendlyName)
                xml_resp += create_xml_value_tag("name", SAList[listCount].friendlyName)
                xml_resp += create_xml_value_tag(FRIENDLY_NAME, SAList[listCount].friendlyName)
                xml_resp += create_host_support_tag(SAList[listCount].flags)
                xml_resp += create_initiators_list_tag(SAList[listCount].initiatorPortList)
                xml_resp += create_target_list_tag(SAList[listCount].targetPortList)
                xml_resp += create_volumeLun_list_tag(SAList[listCount].storageVolumeLunInfoList)
                xml_resp += create_xml_end_tag(SA_TAG_NAME)
            xml_resp += create_xml_end_tag(SA_LIST_TAG_NAME)

        util.SMlog("from self.adapter.assignStorageNode with parameters: %s " % (xml_resp))

        return (xml_resp, lun, lunid)


    def unassignStorageNode(self, ssid, svid, hostname, optflags, protocol):
        """ Unassign Storage Node """
        _trace("entry CSLGHandler.CSLGHandlerInternal.unassignStorageNode()")

        #extract IQN or WWN
        iqn = []
        items = optflags.split(" ")
        _trace("optflags: %s" % optflags)
        for item in items:
            cmd = item.split("=")
            _trace("iqn/wwn: %s"% cmd)
            if len(cmd) > 1:
                if cmd[0] == 'iqn' or cmd[0] == 'wwn':
                   if cmd[1].find(",") != -1:
                       iqn = cmd[1].split(",")
                   else:
                       iqn.append(cmd[1])
        _trace("Extracted IQN: %s" % iqn)

        InitiatorList = self.adapter_inf.ListOfStrings()
        for value in iqn:
            InitiatorList.append(value)
        
        #extract protocol
        connection = SYSTEM_SUPPORTS_ISCSI
        if protocol == "FC":
            connection = SYSTEM_SUPPORTS_FIBRE_CHANNEL

        err = self.adapter.unassignStorageNode(self.cred, ssid, svid, hostname, InitiatorList, "", connection)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGUnassign', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.unassignStorageNode(): ' + err.get_msg())
        
        return

    
    def createCloneOrSnapshot(self, operation, ssid, src_svid, provisioning_type):
        """ Create Clone or Snapshot """
        _trace("entry CSLGHandler.CSLGHandlerInternal.createCloneOrSnapshot()")

        _trace("src_svid is %s" %src_svid)

        sessionInfo = self.adapter_inf.CXSSASnapshotSessionInfo()
        sessionInfo.sessionId = "1"
        sessionInfo.sourceStorageSystemId = ssid
        sessionInfo.sourceStorageVolumeIdList.append(src_svid)
        sessionInfo.targetStorageSystemId = ssid
        #sessionInfo.targetStorageVolumeIdList.append("")
        sessionInfo.targetStorageVolumeBackingStoreList.append("")
        sessionInfo.snapshotOperation = SNAP_OP_CREATE
        if operation == "sv-clone-create":
            sessionInfo.snapshotType = SNAPSHOT_TYPE_CLONE
        else:
            sessionInfo.snapshotType = SNAPSHOT_TYPE_DIFF
        if provisioning_type == "THIN":
            sessionInfo.provisioningType = SYSTEM_SUPPORTS_PROVISION_THIN
        else:
            sessionInfo.provisioningType = SYSTEM_SUPPORTS_PROVISION_FULL
        sessionInfo.snapshotState = SNAP_STATE_NORMAL
        sessionInfo.targetRaidLevel = ""
        sessionInfo.targetStoragePoolId = ""
        #sessionInfo.targetFriendlyNameList.append("")

        ssaInfoList = self.adapter_inf.ListOfSnapshotAssocInfo()
        err = self.adapter.setupSnapshot( self.cred, self.cred, sessionInfo, ssaInfoList)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGSnapClone', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.createCloneOrSnapshot.setupSnapshot(): ' + err.get_msg())
            return
            
        err = self.adapter.createSnapshot( self.cred, self.cred, sessionInfo, ssaInfoList, 1)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGSnapClone', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.createCloneOrSnapshot.createSnapshot(): ' + err.get_msg())
            return

        _trace("DEBGU: SVID: %s" % ssaInfoList[0].targetStorageVolumeId)
        targetStorageVolume = 0
        if ssaInfoList.size() > 0 :
            targetStorageVolume = ssaInfoList[0].targetStorageVolumeId

        return targetStorageVolume
   
    def getStoragePoolIds(self, ssid):
        """ List of Storage Pool Ids"""  
        _trace("entry CSLGHandler.CSLGHandlerInternal.getStoragePoolIds()")

        listSPIDs = []
        SPList = self.adapter_inf.ListOfStoragePoolInfo()
        filters = self.adapter_inf.ListOfStrings()
        err = self.adapter.enumerateStoragePools( self.cred, ssid, filters, SPList)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGPoolList', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.getStoragePoolIds(): ' + err.get_msg())
            return []

        for listCount in range(SPList.size()):
            listSPIDs.append(SPList[listCount].storagePoolId)

        return listSPIDs

    def listStoragePools(self, ssid):
        """ List Storage Pools in XML"""
        _trace("entry CSLGHandler.CSLGHandlerInternal.listStoragePools()")

        SPList = self.adapter_inf.ListOfStoragePoolInfo()
        filters = self.adapter_inf.ListOfStrings()
        err = self.adapter.enumerateStoragePools( self.cred, ssid, filters, SPList)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGPoolList', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.listStoragePools(): ' + err.get_msg())
            return

        #Storage Pool Info in XML format
        xml_resp = XML_HEADER
        xml_resp += create_xml_start_tag(SP_LIST_TAG_NAME)
        for listCount in range(SPList.size()):
            xml_resp += create_xml_start_tag(SP_TAG_NAME)
            xml_resp += create_xml_value_tag(DISPLAY_NAME, SPList[listCount].friendlyName)
            xml_resp += create_xml_value_tag(FRIENDLY_NAME, SPList[listCount].friendlyName)
            xml_resp += create_xml_value_tag("storagePoolId", SPList[listCount].storagePoolId)
            xml_resp += create_xml_value_tag("parentStoragePoolId", SPList[listCount].parentPool)
            xml_resp += create_xml_value_tag("storageSystemId", SPList[listCount].storageSystemId)
            xml_resp += create_xml_value_tag("sizeInMB", SPList[listCount].sizeInMB)
            xml_resp += create_xml_value_tag("freeSpaceInMB", SPList[listCount].freeSpaceInMB)
            xml_resp += create_xml_value_tag("availableFreeSpaceInMB", SPList[listCount].availableFreeSpaceInMB)
            xml_resp += create_xml_value_tag("isDefault", (lambda:"No", lambda:"Yes") [SPList[listCount].isDefault]())
            xml_resp += create_xml_value_tag("status", SPList[listCount].status)
            xml_resp += create_xml_start_tag("provisioningOptions")
            xml_resp += create_raid_support_tag(SPList[listCount].supportedRaidTypes)
            xml_resp += create_node_support_tag(SPList[listCount].supportedNodeTypes)
            xml_resp += create_prov_support_tag(SPList[listCount].supportedProvisioningTypes)
            xml_resp += create_xml_end_tag("provisioningOptions")
            xml_resp += create_xml_end_tag(SP_TAG_NAME)
        xml_resp += create_xml_end_tag(SP_LIST_TAG_NAME)

        return xml_resp 

    def getStoragePoolInfoByName(self, ssid, poolName):
        """ get storage pool info by providing name """
        _trace("entry CSLGHandler.CSLGHandlerInternal.getStoragePoolInfoByName()")

        resp ={}
        SPList = self.adapter_inf.ListOfStoragePoolInfo()
        filters = self.adapter_inf.ListOfStrings()
        err = self.adapter.enumerateStoragePools( self.cred, ssid, filters, SPList)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGPoolList', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.getStoragePoolInfoByName(): ' + err.get_msg())
            return

        for listCount in range(SPList.size()):
            if SPList[listCount].friendlyName.find(poolName)==0:
                resp = SPList[listCount]

        return resp

    def destroyStoragePool(self, ssid, spid):
        """ Destroy Storage Pools """
        _trace("entry CSLGHandler.CSLGHandlerInternal.storagePoolDestroy()")

        err = self.adapter.deleteStoragePool( self.cred, ssid, spid)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGPoolDelete', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.destroyStoragePool(): ' + err.get_msg())
            return

        return


    def getStorageSystemIds(self, adapterid, ip, username, password):
        """ Get Storage System IDs """
        _trace("entry CSLGHandler.CSLGHandlerInternal.getStorageSystemIds")
        listSSIDs = []
        cred = self.adapter_inf.CXSSAManagementCredentials()
        cred.storageAdapterId = adapterid
        cred.uuid = ""
        cred.friendlyName = ""
        cred.ipAddress = ip
        cred.portNumber = 0
        cred.username = username
        cred.password = password
        cred.context = ""
        cred.ns = ""
 
        filters = self.adapter_inf.ListOfStrings()
        list = self.adapter_inf.ListOfStorageSysInfo()
        err = self.adapter.enumerateStorageSystems( cred, filters, list)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGSysInfo', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.getStorageSystemIds(): ' + err.get_msg())
            return []

        for listCount in range(list.size()):
            listSSIDs.append(list[listCount].storageSystemId)

        return listSSIDs

    def listStorageSystems(self):
        """ List Storage Systems """
        _trace("entry CSLGHandler.CSLGHandlerInternal.listStorageSystems()")

        filters = self.adapter_inf.ListOfStrings()
        list = self.adapter_inf.ListOfStorageSysInfo()
        err = self.adapter.enumerateStorageSystems( self.cred, filters, list)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGSysInfo', \
                     opterr='FAILED: %s'%SL_ERR[err] + ' (' + err.get_msg() \
                     + ')')
            return

        #Storage System in XML format
        xml_resp = XML_HEADER
        xml_resp += create_xml_start_tag(SS_LIST_TAG_NAME)
        for listCount in range(list.size()):
            xml_resp += create_xml_start_tag(SS_TAG_NAME)
            xml_resp += create_xml_value_tag(FRIENDLY_NAME, list[listCount].friendlyName)
            xml_resp += create_xml_value_tag(DISPLAY_NAME, (list[listCount].vendor + " " + list[listCount].model + " (" +list[listCount].friendlyName +")" ))
            xml_resp += create_xml_value_tag(VENDOR, list[listCount].vendor)
            xml_resp += create_xml_value_tag(MODEL, list[listCount].model)
            xml_resp += create_xml_value_tag(SERIAL_NUM, list[listCount].serialNum)
            xml_resp += create_xml_value_tag(SSID, list[listCount].storageSystemId)
            xml_resp += create_sys_capablities_tag(list[listCount].capabilities)        
            xml_resp += create_protocol_support_tag(list[listCount].capabilities)        
            xml_resp += create_snapMethod_tag(list[listCount].snapshotMethods)        
            #xml_resp += create_tags(list[listCount].capabilities)        
            #xml_resp += create_raidtypes_tag(list[listCount].capabilities)
            xml_resp += create_xml_end_tag(SS_TAG_NAME)        
        xml_resp += create_xml_end_tag(SS_LIST_TAG_NAME)

        return xml_resp


    def listStorageVolumes(self, ssid, spid, as_svids_names=True, as_xml=False):
        """ List Storage Volumes """
        _trace("entry CSLGHandler.CSLGHandlerInternal.listStorageVolumes()")

        filters = self.adapter_inf.ListOfStrings()
        SVList = self.adapter_inf.ListOfStorageVolumeInfo()
        err = self.adapter.enumerateStorageNodes( self.cred, ssid, spid, filters, SVList)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGLunList', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.listStorageVolumes(): ' + err.get_msg())
            return

        #Storage Volume info in XML format
        if as_xml == True:
            xml_resp = XML_HEADER
            xml_resp += create_xml_start_tag(VOL_LIST_TAG_NAME)
            for listCount in range(SVList.size()):
                xml_resp += create_xml_start_tag(VOL_TAG_NAME)
                xml_resp += create_xml_value_tag(DISPLAY_NAME, SVList[listCount].friendlyName)
                xml_resp += create_xml_value_tag(FRIENDLY_NAME, SVList[listCount].friendlyName)
                xml_resp += create_xml_value_tag("sizeInMB", SVList[listCount].sizeInMB)
                xml_resp += create_xml_value_tag("usedSpaceInMB", SVList[listCount].usedSpaceInMB)
                xml_resp += create_xml_value_tag("accessPermissions", (lambda:"RW", lambda:"RO")[SVList[listCount].accessPermissions & 1]())
                xml_resp += create_xml_value_tag("serialNum", SVList[listCount].serialNum)
                xml_resp += create_xml_value_tag("storageVolumeId", SVList[listCount].storageVolumeId)
                xml_resp += create_xml_value_tag("storageSystemId", SVList[listCount].storageSystemId)
                xml_resp += create_node_support_tag(SVList[listCount].nodeType)
                xml_resp += create_xml_value_tag("isSnapshot", (lambda:"No", lambda:"Yes") [SVList[listCount].isSnapshot]())
                xml_resp += create_xml_value_tag("lunIqn", SVList[listCount].lunIqn)
                xml_resp += create_xml_end_tag(VOL_TAG_NAME)
            xml_resp += create_xml_end_tag(VOL_LIST_TAG_NAME)
            return xml_resp
        elif as_svids_names == True:
            listSVIDs = []
            listSViqns = []
            listNames = []
            for listCount in range(SVList.size()):
                listSVIDs.append(SVList[listCount].storageVolumeId)
                listSViqns.append(SVList[listCount].lunIqn)
                listNames.append(SVList[listCount].friendlyName)
            return (listSVIDs,listNames, listSViqns)

    def getStorageVolumeInfoByName(self, ssid, spid, svName):
        """ Get Storage Volume Info By Name """
        _trace("entry CSLGHandler.CSLGHandlerInternal.getStorageVolumeInfoByName()")

        svid = None
        (svids, svNames) = self.listStorageVolumes(ssid, spid)
        count=0
        for sv_name in svNames:
            if sv_name == svName:
                svid = svids[count]
                break
            count= count+1
        return svid

    def getSnapMirrorLunName(self, ssid, source_lun_name, dest_loc):
        """ get snap Mirrored Volume name """
        _trace("entry CSLGHandler.CSLGHandlerInternal.getSnapMirrorLunName()")
        
        outParams = self.adapter_inf.CXSSACustomParamSet()
        inParams = self.adapter_inf.CXSSACustomParamSet()
        lunName_value = self.adapter_inf.CXSSAValue(source_lun_name, 1)
        volLoc_value = self.adapter_inf.CXSSAValue(dest_loc, 1)

        inParams.add("lunName", lunName_value)
        inParams.add("volLocation", volLoc_value)

        err = self.adapter.customApi( self.cred, ssid, "getSnapMirrorLunName", inParams, outParams)
        if err > ADAPTER_CALL_SUCCESS:
            raise xs_errors.XenError('CSLGSnapClone', \
                     opterr='FAILED: CSLGHandler.CSLGHandlerInternal.customApi(getSnapMirrorLunName)(): ' + err.get_msg())
            return

        label_value = self.adapter_inf.CXSSAValue()
        outParams.getValue("lun_label", label_value)
        label = label_value.getValueStr()

        return label

# function
def _trace(string, errorcase = 0, priority=syslog.LOG_INFO):
    """ simple function to control trace on/off """
    if TRACEON or errorcase:
        util.SMlog(string, priority=priority)

def create_xml_value_tag(tag_name, value):
    xml_out = ""
    if tag_name != "":
        xml_out = "<" + tag_name + ">" + str(value) + "</" + tag_name + ">"

    return xml_out


def create_xml_start_tag(tag_name):
    xml_out = ""
    if tag_name != "":
        xml_out = "<" + tag_name + ">"

    return xml_out


def create_xml_end_tag(tag_name):
    xml_out = ""
    if tag_name != "":
        xml_out = "</" + tag_name + ">\n"

    return xml_out


def compare_create_xml_tag(capablities, comp_value, tag, value):
    xml_out = ""
    if capablities & comp_value:
        xml_out += create_xml_start_tag(tag)
        xml_out += value
        xml_out += create_xml_end_tag(tag)

    return xml_out


def create_host_support_tag(capablities):
    xml_out = ""
    if capablities != "":
        xml_out += create_xml_start_tag("hostOsType")
        xml_out += compare_create_xml_tag(capablities, HOST_OS_TYPE_UNKNOWN, "hostType",  "Unknown")
        xml_out += compare_create_xml_tag(capablities, HOST_OS_TYPE_LINUX, "hostType",  "Linux")
        xml_out += compare_create_xml_tag(capablities, HOST_OS_TYPE_WINDOWS, "hostType",  "Windows")
        xml_out += compare_create_xml_tag(capablities, HOST_OS_TYPE_SOLARIS, "hostType",  "Solaris")
        xml_out += compare_create_xml_tag(capablities, HOST_OS_TYPE_AIX, "hostType",  "AIX")
        xml_out += compare_create_xml_tag(capablities, HOST_OS_TYPE_HPUX, "hostType",  "HPUX")
        xml_out += compare_create_xml_tag(capablities, HOST_OS_TYPE_VMWARE, "hostType",  "VMWare")
        xml_out += create_xml_end_tag("hostOsType")

    return xml_out


def create_volumeLun_list_tag(svLunList):
    xml_out = create_xml_start_tag(SVLUN_LIST_TAG_NAME)
    for listCount in range(svLunList.size()):
        xml_out += create_xml_start_tag(SVLUN_TAG_NAME)
        xml_out += create_xml_value_tag("storageVolumeId", svLunList[listCount].storageVolumeId)
        xml_out += create_xml_value_tag("lunIqn", svLunList[listCount].lunIqn)
        xml_out += create_xml_value_tag("lun", svLunList[listCount].lun)
        xml_out += create_xml_end_tag(SVLUN_TAG_NAME)
    xml_out += create_xml_end_tag(SVLUN_LIST_TAG_NAME)

    return xml_out


def create_target_list_tag(targetPortList):
    xml_out = create_xml_start_tag(TARGETS_LIST_TAG_NAME)
    for listCount in range(targetPortList.size()):
        xml_out += create_xml_start_tag(TARGETS_TAG_NAME)
        xml_out += create_xml_value_tag("portId", targetPortList[listCount].portId)
        xml_out += create_xml_value_tag("friendlyName", targetPortList[listCount].friendlyName)
        xml_out += create_xml_value_tag("displayName", targetPortList[listCount].friendlyName)
        xml_out += create_xml_value_tag("storageSystemId", targetPortList[listCount].storageSystemId)
        xml_out += create_xml_value_tag("protocolType", (lambda:"ISCSI", lambda:"FC")[targetPortList[listCount].protocolType & SYSTEM_SUPPORTS_FIBRE_CHANNEL]())
        if targetPortList[listCount].protocolType & SYSTEM_SUPPORTS_ISCSI :
            xml_out += create_xml_value_tag("portIqn", targetPortList[listCount].portIqn)
        xml_out += create_xml_end_tag(TARGETS_TAG_NAME)
    xml_out += create_xml_end_tag(TARGETS_LIST_TAG_NAME)

    return xml_out


def create_initiators_list_tag(initiatorPortList):
    xml_out = create_xml_start_tag(INITIATORS_LIST_TAG_NAME)
    for listCount in range(initiatorPortList.size()):
        xml_out += create_xml_start_tag(INITIATORS_TAG_NAME)
        xml_out += create_xml_value_tag("portId", initiatorPortList[listCount].portId)
        xml_out += create_xml_value_tag("friendlyName", initiatorPortList[listCount].friendlyName)
        xml_out += create_xml_value_tag("displayName", initiatorPortList[listCount].friendlyName)
        xml_out += create_xml_value_tag("storageSystemId", initiatorPortList[listCount].storageSystemId)
        #xml_out += create_xml_value_tag("protocolType", (lambda:"ISCSI", lambda:"FC")[initiatorPortList[listCount].protocolType & SYSTEM_SUPPORTS_FIBRE_CHANNEL]())
        xml_out += create_xml_value_tag("hostType", initiatorPortList[listCount].hostType)
        xml_out += create_xml_end_tag(INITIATORS_TAG_NAME)
    xml_out += create_xml_end_tag(INITIATORS_LIST_TAG_NAME)

    return xml_out


def create_sys_capablities_tag(capablities):
    xml_out = ""
    if capablities != "":
        xml_out += create_xml_start_tag(SYS_CAP_TAG_NAME)
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_PROVISION_FULL, "capabilities",  "PROVISIONING")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_PROVISION_THIN, "capabilities",  "THIN_PROVISIONING")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_MAPPING, "capabilities",  "MAPPING")    
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_MULTIPLE_STORAGE_POOLS, "capabilities",  "MULTIPLE_STORAGE_POOLS")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_LUN_GROUPING, "capabilities",  "LUN_GROUPING")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_POOL_LEVEL_DEDUPLICATION, "capabilities",  "DEDUPLICATION")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_VOLUME_LEVEL_DEDUPLICATION, "capabilities",  "VOLUME_LEVEL_DEDUPLICATION")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_DIFF_SNAPSHOT, "capabilities",  "DIFF_SNAPSHOT")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_PLEX_SNAPSHOT, "capabilities",  "PLEX_SNAPSHOT")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_REMOTE_REPLICATION, "capabilities",  "REMOTE_REPLICATION")    
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_CLONE, "capabilities",  "CLONE")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RESIZE, "capabilities",  "RESIZE")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_REQUIRES_STORAGE_POOL_CLEANUP, "capabilities",  "REQUIRES_STORAGE_POOL_CLEANUP")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_CUSTOM_ENDPOINTS, "capabilities",  "SUPPORTS_CUSTOM_ENDPOINTS")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_OPTIMIZED_ISCSI_LOGIN, "capabilities",  "SUPPORTS_OPTIMIZED_ISCSI_LOGIN")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_LUN_PER_IQN, "capabilities",  "SUPPORTS_LUN_PER_IQN")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_CHAP, "capabilities",  "SUPPORTS_CHAP")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_INSTANT_CLONE, "capabilities",  "SUPPORTS_INSTANT_CLONE")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_CLONE_OF_SNAPSHOT, "capabilities",  "SUPPORTS_CLONE_OF_SNAPSHOT")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_REQUIRES_ISCSI_PRE_LOGIN, "capabilities",  "REQUIRES_ISCSI_PRE_LOGIN")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_SNAPSHOT_OF_SNAPSHOT, "capabilities",  "SUPPORTS_SNAPSHOT_OF_SNAPSHOT")
        xml_out += create_xml_end_tag(SYS_CAP_TAG_NAME)
    return xml_out


def create_prov_support_tag(capablities):
    xml_out = ""
    if capablities != "":
        xml_out += create_xml_start_tag("supportedProvisioningTypes")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_PROVISION_FULL, "provisioningType",  "FULL")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_PROVISION_THIN, "provisioningType",  "THIN")
        xml_out += create_xml_end_tag("supportedProvisioningTypes")
    return xml_out


def create_raid_support_tag(capablities):
    xml_out = ""
    if capablities != "":
        xml_out += create_xml_start_tag("supportedRaidTypes")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID0_CONCAT, "raidType",  "RAID0")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID0, "raidType",  "RAID0")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID1, "raidType",  "RAID1")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID3, "raidType",  "RAID3")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID4, "raidType",  "RAID4")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID5, "raidType",  "RAID5")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID6, "raidType",  "RAID6")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID10, "raidType",  "RAID10")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_RAID50, "raidType",  "RAID50")
        xml_out += create_xml_end_tag("supportedRaidTypes")
    return xml_out


def create_protocol_support_tag(capablities):
    xml_out = ""
    if capablities != "":
        xml_out += create_xml_start_tag(PROT_SUPPORT_TAG_NAME)
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_FIBRE_CHANNEL, "capabilities",  "FC")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_ISCSI, "capabilities",  "ISCSI")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_NFS, "capabilities",  "NFS")
        xml_out += compare_create_xml_tag(capablities, SYSTEM_SUPPORTS_CIFS, "capabilities",  "CIFS")
        xml_out += create_xml_end_tag(PROT_SUPPORT_TAG_NAME)
    return xml_out


def create_node_support_tag(capablities):
    xml_out = ""
    if capablities != "":
        xml_out += create_xml_start_tag("supportedNodeTypes")
        xml_out += compare_create_xml_tag(capablities, NODE_TYPE_UNKNOWN, "nodeType",  "Unknown")
        xml_out += compare_create_xml_tag(capablities, NODE_TYPE_NFS_SHARE, "nodeType",  "NFS")
        xml_out += compare_create_xml_tag(capablities, NODE_TYPE_CIFS_SHARE, "nodeType",  "CIFS")
        xml_out += compare_create_xml_tag(capablities, NODE_TYPE_STORAGE_VOLUME, "nodeType",  "STORAGE_VOLUME")
        xml_out += create_xml_end_tag("supportedNodeTypes")
    return xml_out


def create_snap_type_support_tag(capablities):
    xml_out = ""
    if capablities != "":
        xml_out += create_xml_start_tag("snapshotTypeList")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_TYPE_DIFF, "snapshotType",  "DIFF_SNAPSHOT")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_TYPE_PLEX, "snapshotType",  "MIRROR_BREAK_OFF")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_TYPE_REMOTE_REPLICATION, "snapshotType",  "REMOTE_REPLICATION")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_TYPE_IS_DEFAULT, "snapshotType",  "IS_DEFAULT")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_TYPE_CLONE, "snapshotType",  "CLONE")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_TYPE_REMOTE_REPLICATION_SYNC, "snapshotType",  "REMOTE_REPLICATION_SYNC")
        xml_out += create_xml_end_tag("snapshotTypeList")
    return xml_out


def create_snap_support_tag(capablities):
    xml_out = ""
    if capablities != "":
        xml_out += create_xml_start_tag("snapshotCapabilities")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_CAPABILITY_THIN_PROVISIONED_TARGET, "capabilities",  "THIN_PROVISIONED_TARGET")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_CAPABILITY_REQ_PRE_PROVISIONED_TARGET, "capabilities",  "REQUIRES_PRE_PROVISIONED_TARGET")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_CAPABILITY_AUTO_PROVISIONED_TARGET, "capabilities",  "AUTO_PROVISIONED_TARGET")
        xml_out += compare_create_xml_tag(capablities, SNAPSHOT_CAPABILITY_POOL_SPECIFICATION, "capabilities",  "POOL_SPECIFICATION_ALLOWED")
        xml_out += create_xml_end_tag("snapshotCapabilities")
    return xml_out


def create_snapMethod_tag(sm_info):
    xml_out = create_xml_start_tag(SNAP_LIST_TAG_NAME)
    for listCount in range(sm_info.size()):
        xml_out += create_xml_start_tag(SNAP_TAG_NAME)
        xml_out += create_xml_value_tag("name", sm_info[listCount].name)
        xml_out += create_xml_value_tag("displayName", sm_info[listCount].name)
        xml_out += create_xml_value_tag("maxSnapshots", sm_info[listCount].maxSnapshots)
        xml_out += create_node_support_tag(sm_info[listCount].nodeTypes)
        xml_out += create_snap_type_support_tag(sm_info[listCount].snapshotType)
        xml_out += create_snap_support_tag(sm_info[listCount].snapshotCapabilities)
        xml_out += create_xml_end_tag(SNAP_TAG_NAME)
    xml_out += create_xml_end_tag(SNAP_LIST_TAG_NAME)

    return xml_out

def error_handler(methodName, err):
    if err > ADAPTER_CALL_SUCCESS:
        if SL_ERR.has_key(err):
            raise xs_errors.XenError('CSL_Integrated', \
                opterr='FAILED: %s'%SL_ERR[err] + ' (' + err.get_msg() + ')')
        else:
            raise xs_errors.XenError('CSL_Integrated', \
                opterr='FAILED: Failed in %s'%methodName + ' (' \
                + err.get_msg() + ')')
