#!/usr/bin/env python
# Copyright (c) 2005-2007 XenSource, Inc. 2008 Citrix Systems.
# All use and distribution of this copyrighted material is governed by 
# and subject to terms and conditions as licensed by XenSource, Inc. All 
# other rights reserved.  Xen, XenSource and XenEnterprise are either 
# registered trademarks or trademarks of XenSource Inc. in the United 
# States and/or other countries.
#
#
# CSLGSR: CSLG storage repository
# 

""" The cslg SR implementation """

import SR, VDI, ISCSISR, HBASR, SRCommand, util
import xml.dom.minidom
import xs_errors
import socket
from socket import SOCK_DGRAM, AF_INET
import scsiutil
import syslog
import scsi_host_rescan
import fcntl, struct
import os, time, copy, glob
import zlutil
import xmlrpclib
import re
import CSLGHandler
from srmetadata import NAME_LABEL_TAG, NAME_DESCRIPTION_TAG, UUID_TAG, \
    IS_A_SNAPSHOT_TAG, SNAPSHOT_OF_TAG, TYPE_TAG, VDI_TYPE_TAG, READ_ONLY_TAG, \
    MANAGED_TAG, SNAPSHOT_TIME_TAG, METADATA_OF_POOL_TAG, SLMetadataHandler, \
    METADATA_OBJECT_TYPE_VDI, METADATA_OBJECT_TYPE_SR, \
    METADATA_UPDATE_OBJECT_TYPE_TAG, SVID_TAG, requiresUpgrade

from xmlrpclib import DateTime

CAPABILITIES = ["SR_PROBE", "SR_UPDATE", "VDI_CREATE", "VDI_DELETE", "VDI_ATTACH",
                "VDI_DETACH", "VDI_RESIZE", "VDI_RESIZE_ONLINE",
#                "VDI_DETACH",
                "VDI_SNAPSHOT", "VDI_CLONE", "VDI_INTRODUCE", "VDI_UPDATE" ]

CONFIGURATION = [ [ 'target', 'IP address or hostname of the array MGT interface' ],
                  [ 'username', 'Login username to be used by the adapter'        ],
                  [ 'password', 'Login password to be used by the adapter'        ],
                  [ 'port', 'port number to connect to storage target [optional]' ],
                  [ 'adapterid', 'storage system adapter ID'                      ],
                  [ 'storageSystemId', 'storage system id'                        ],
                  [ 'storagePoolId', 'storage pool id, e.g RAID group'            ],
                  [ 'protocol', 'xen storage protocol (fc, iscsi) [optional]'     ],
                  [ 'chapuser', 'ISCSI CHAP username [optional]'                  ],
                  [ 'chappassword', 'ISCSI CHAP password [optional]'              ],
                  [ 'provision-type', 'thick/thin [optional]'                     ],
                  [ 'provision-options', 'options (dedup) [optional]'             ],
                  [ 'raid-type', 'Raid type [optional]'                           ]
                ]

DRIVER_INFO = {
    'name': 'cslg',
    'description': 'CSLG SR plugin',
    'vendor': 'Citrix Systems Inc',
    'copyright': '(C) 2011 Citrix Systems Inc',
    'driver_version': '1.0',
    'required_api_version': '1.0',
    'capabilities': CAPABILITIES,
    'configuration': CONFIGURATION
    }

TYPE = 'cslg'
MAX_TIMEOUT = 15
#>>>> There is no way to query minimum size from adapter. so, we use static value
MIN_ALLOCATION = (15*1024*1024)

TRACEON = 1
MDVOLUME_PREFIX = "SL_XenStorage_"
MDVOLUME_NAME   = "MGT"

class CSLGSR(SR.SR):
    """CSLG storage repository"""

    def handles(something):
        """ handles function """
        _trace("entry CSLG.handles()")
        if something == TYPE:
            return True
        return False
    handles = staticmethod(handles)

    def load(self, sr_uuid):
        """ load method, is called by all __init__ of sr """
        _trace("entry CSLG.load()")

        _trace("sr-uuid from probe: %s" % sr_uuid)
        self.sr_vditype = 'aio'
        self.adapter = ''
        self.cslg = ''
        self.cslguser = ''
        self.cslgpassword = ''
        self.cslport = 0
        self.ssid = ''
        self.pool = ''
        self.parentpool = ''
        # either is default (blank) unless overridden
        self.protocol = ''
        # defaults are empty, unless overridden
        self.provision_type = ''
        self.provision_options = ''
        self.raid_type = ''
        self.allowed_protocols = ''
        self.chapusername = ''
        self.chappassword = ''
        self.req_pool_cleanup = 0
        self.supports_resize = 0
        self.serial_number = ''
        self.poolcount = 0
        self.exists = False
        self.uuid = sr_uuid
        self.md_vdi_iqn = ''
        self.ports = None
        self.targets = None

        # if there is no cslg server listed, attempt to derive the
        # cslg server from a combination of SR or POOL parameters
        # 
        # Note that for the only required dconf param is the xen
        # storage server -- which is either passed in or derived.
        # all other params are options for discover, and required
        # for all other operations.

        if self.dconf.has_key('adapterid') and self.dconf['adapterid']:
            self.adapter = self.dconf['adapterid']

        if self.dconf.has_key('target') and self.dconf['target']:
            try:
                if self.dconf['target'].find(':') != -1:
                    conn = self.dconf['target'].split(':')
                    self.cslg = util._convertDNS(conn[0])
                    self.cslport = conn[1]
                else:
                    self.cslg = util._convertDNS(self.dconf['target'])
            except:
                raise xs_errors.XenError('CSLGServer')

        if self.dconf.has_key('cslport') and \
           self.dconf['cslport']:
            self.cslport = self.dconf['cslport']

        if self.dconf.has_key('username') and \
           self.dconf['username']:
            self.cslguser = self.dconf['username']

        if self.dconf.has_key('password') and self.dconf['password']:
            self.cslgpassword = self.dconf['password']
        elif self.dconf.has_key('password_secret'):
            self.cslgpassword = util.get_secret(self.session, self.dconf['password_secret'])

        if self.dconf.has_key('storageSystemId') and \
           self.dconf['storageSystemId']:
            self.ssid = self.dconf['storageSystemId']

        if self.dconf.has_key('storagePoolId') and \
           self.dconf['storagePoolId']:
            self.pool = self.dconf['storagePoolId']

        if self.dconf.has_key('protocol') and \
           self.dconf['protocol']:
            self.protocol = self.dconf['protocol'].upper()            

        # iscsi, fc and 'none'--> any available, are supported
        if self.protocol not in ['ISCSI', 'FC', '', 'AUTO']:
            raise xs_errors.XenError('CSLGInvalidProtocol')
        if self.protocol == 'AUTO':
            self.protocol = ''

        if self.dconf.has_key('provision-type') and \
           self.dconf['provision-type']:
            self.provision_type = self.dconf['provision-type'].upper()

        if self.dconf.has_key('provision-options') and \
           self.dconf['provision-options']:
            self.provision_options = self.dconf['provision-options'].upper()

        if self.dconf.has_key('raid-type') and \
           self.dconf['raid-type']:
            self.raid_type = self.dconf['raid-type'].upper()

        if self.dconf.has_key('chapuser') and \
           self.dconf['chapuser']:
            self.chapusername = self.dconf['chapuser']

        if self.dconf.has_key('chappassword') and self.dconf['chappassword']:
            self.chappassword = self.dconf['chappassword']
        elif self.dconf.has_key('chappassword_secret'):
            self.chappassword = util.get_secret(self.session, self.dconf['chappassword_secret'])

        self.isMaster = False
        if self.dconf.has_key('SRmaster') and self.dconf['SRmaster'] == 'true':
            self.isMaster = True

        os.environ['LANG'] = 'en_US.UTF-8'        
 
        try:
            self.sm_config =  self.session.xenapi.SR.get_sm_config(self.sr_ref)
        except:
            pass
            
        self.metadataVolName = '%s_%s' % (MDVOLUME_PREFIX, sr_uuid.replace("-","_"))
        self.metadataVolumePath = None

        self.cslghandler = CSLGHandler.CSLGHandlerInternal(self)
            
    def create(self, sr_uuid, size):
        """ create method """
        _trace("entry CSLG.create()")

        assert(sr_uuid)
        self._check_required_params()

    def createMetadataVolume(self):
        _trace("entry createMetadataVolume()")
        # At this point create a new storage node on the array
        raid_type = ''
        self.md_svid = None
        if self.dconf.has_key('raid-type'):
            raid_type = self.dconf['raid-type']

        if self.sm_config.has_key('requires_subpool') and self.sm_config['requires_subpool'] == 'True':
            #get pool info, to check if it's already there
            modifiedPoolName = "Xen_"+(self.metadataVolName.replace("-", "_"))
            poolInfo = self.cslghandler.getStoragePoolInfoByName(self.ssid, modifiedPoolName)
            if poolInfo == {}:
                self.pool = _createNetappPool(self, self.metadataVolName, self.uuid, MIN_ALLOCATION)
            else:
                self.pool = poolInfo.storagePoolId

            # first remove subpool is already present
            try:
                self.session.xenapi.SR.remove_from_sm_config(self.sr_ref, 'subpool')
            except:
                pass

            # Now put this subpool in the sm_config
            self.session.xenapi.SR.add_to_sm_config(self.sr_ref, 'subpool', self.pool)

        (self.SVids, sv_names, sv_iqns) = self.cslghandler.listStorageVolumes(self.ssid, self.pool)
        count=0
        for sv_name in sv_names:
            if self.metadataVolName in sv_name or self.metadataVolName in sv_name.replace("-", "_"):
               self.md_svid = self.SVids[count]
               self.md_vdi_iqn = sv_iqns[count]
               break
            count= count+1

        if self.md_svid == None:
            create_args = [self.metadataVolName, MIN_ALLOCATION, self.ssid, self.pool, self.uuid, False, raid_type, '', '', self.provision_options]
            _trace( "create_args = %s" % create_args )
            (self.md_svid, self.md_vdi_iqn) = \
                self.cslghandler.createStorageNode(self.metadataVolName, \
                       MIN_ALLOCATION, self.ssid, self.pool, self.uuid, \
                       False, raid_type, '', '', \
                       self.provision_options)
        util.SMlog("Metadata volume %s target IQN is: %s" % (self.md_svid, self.md_vdi_iqn))
        
        # Now put this svid in the sm_config
        self.session.xenapi.SR.add_to_sm_config(self.sr_ref, 'md_svid', self.md_svid)
        # Now put this md_vdi_iqn in the sm_config
        if self.md_vdi_iqn != None:
            self.session.xenapi.SR.add_to_sm_config(self.sr_ref, 'md_vdi_iqn', self.md_vdi_iqn)

        # Put the adapter name, e.g., "SMIS_STORAGE_ADAPTER" into sm_config
        (adptr_name, fndly_name, desc, type, repType, verMaj, verMin) = self.cslghandler.getStorageAdapterInfo()
        _trace("Add adapter info name to sm_config: %s" % adptr_name )
        self.session.xenapi.SR.add_to_sm_config(self.sr_ref, 'adapter_name', adptr_name)

        #refresh sm_config in locally stored member
        self.sm_config =  self.session.xenapi.SR.get_sm_config(self.sr_ref)

    def attachMetadataVolume(self):
        _trace("entry attachMetadataVolume()")
        self.SVids = []
        self.lunID = ''
        if self.sm_config.has_key('md_svid') and self.sm_config['md_svid'] != '':
            self.md_svid = self.sm_config['md_svid']
        else:
            util.SMlog("  md_svid not present in sm_config, looks like an "\
                       "upgrade. Create metadata volume.")
            self.createMetadataVolume()

        if self.sm_config.has_key('md_vdi_iqn') and self.sm_config['md_vdi_iqn'] != '':
            self.md_vdi_iqn = self.sm_config['md_vdi_iqn']

        # a) Check to see if Volume already assigned
        devs = scsiutil._genReverseSCSIidmap(self.md_svid, pathname="csldev")

        # b) If not assigned, assign to this host
        if not len(devs):
            #>>>> adapter uses hostname to find the lun mapping, in case it couldn't find using iqn
            # find the LUN ID from the returned XML, call it lunID
            localHostname = _get_local_hostname()
            optFlags = _getOptFlags(self, self.protocol)

            self.lunID = util.retry( lambda: self.cslghandler.assignStorageNode(self.ssid, self.md_svid, localHostname, \
                    'flags=xml,bridge,hostgroup', '', self.protocol, \
                    optFlags, self.chapusername, self.chappassword)[2], maxretry = 1 )

        # c) Activate transport
        if self.protocol == 'ISCSI':
            ###TODO: Check for LUN per IQN
            lun_iqn = self.md_vdi_iqn
            host_devpath = []
            iscsi = _load_iscsi_transport(self, self.ssid, \
                                                              self.original_srcmd,\
                                                              lun_iqn, \
                                                              self.uuid, self.mpath)

            for instance in iscsi:
                try:
                    # Attach the iscsi session
                    instance.attach(self.uuid)
                    for val in instance.adapter:
                        entry = "%s" % instance.adapter[val]
                        if entry not in host_devpath:
                            host_devpath.append(entry)
                except:
                    pass
            

            # save the device path in self.metadataVolumePath
            self.metadataVolumePath = None
            (self.metadataVolumePath,SCSIid, npaths) = _check_path(self.md_svid, self)
            _trace("  attachMetadataVolume: metadataVolumePath is %s" %self.metadataVolumePath)

            # if the path did not come up, maybe the iscsi session exists
            # so just try refreshing it
            if self.metadataVolumePath == None:
                scsiutil.remove_stale_luns(host_devpath, self.lunID, \
                        getExpectedCslDevPath(self.md_svid), self.mpath)
                util.SMlog("  attachMetadataVolume: Refreshing the iscsi session.")
                for host in host_devpath:
                    scsi_host_rescan.rescan(host)
                util.SMlog("  attachMetadataVolume: Refreshed the iscsi session.")

            # save the device path in self.metadataVolumePath
            self.metadataVolumePath = None
            (self.metadataVolumePath,SCSIid, npaths) = _check_path_and_raise(self.md_svid, self)
            _trace("  attachMetadataVolume: metadataVolumePath is %s" %self.metadataVolumePath)

            self.mpathmodule.refresh(SCSIid, npaths)

        elif self.protocol == 'FC':
            (host_adt, host_devpath) = _load_fc_transport(self.uuid, \
                                            self.original_srcmd)
            scsiutil.remove_stale_luns(host_devpath, self.lunID, \
                        getExpectedCslDevPath(self.md_svid), self.mpath)

        # save the device path in self.metadataVolumePath
        if self.metadataVolumePath == None:
            _rescan_for_LUN(self.md_svid, self, host_devpath)
            self.metadataVolumePath = _check_path_and_raise(self.md_svid, self)[0]
            _trace("metadataVolumePath is %s" %self.metadataVolumePath)

        if requiresUpgrade(self.metadataVolumePath):
            util.SMlog("This SR requires metadata upgrade.")
            self.updateSRMetadata()

    def detachMetadataVolume(self):
        _trace("entry detachMetadataVolume()")
        if self.sm_config.has_key('md_svid') and self.isMaster:
            # first save the scsi id for the metadata lun
            self.metadataVolumePath = _check_path_and_raise(self.sm_config['md_svid'], self)[0]
            SCSIid = scsiutil.getSCSIid(self.metadataVolumePath)

            # remove the device from the host
            if self.mpath == "true" and len(SCSIid):
                self.mpathmodule.reset(SCSIid, True) # explicit unmap
                util.remove_mpathcount_field(self.session, self.host_ref, \
                       self.sr_ref, SCSIid)

            # unmap the LUN from the array
            if self.connection == 'ISCSI':
                ###Check for LUN per IQN
                if self.sm_config.has_key('md_vdi_iqn') and self.sm_config['md_vdi_iqn'] != '':
                    self.md_vdi_iqn = self.sm_config['md_vdi_iqn']
                iscsi = _load_iscsi_transport(self, self.ssid, \
                                                                  self.original_srcmd,\
                                                                  self.md_vdi_iqn, \
                                                                  self.uuid, self.mpath)

                util.retry( lambda: self.cslghandler.unassignStorageNode(self.ssid, self.sm_config['md_svid'], \
                                                         _get_local_hostname(), \
                                                         _getOptFlags(self, self.connection), self.connection), maxretry = 5 )
                for instance in iscsi:
                    instance.detach(self.uuid)
            else:
                util.retry( lambda: self.cslghandler.unassignStorageNode(self.ssid, \
                    self.sm_config['md_svid'], _get_local_hostname(), \
                    _getOptFlags(self, self.connection), self.connection), maxretry = 5 )
                
            # now that this has been unmapped from the array
            if len(SCSIid):
                sgdevs = scsiutil.cacheSCSIidentifiers()
                devs = scsiutil._genReverseSCSIidmap(SCSIid)
                for dev in devs:
                    if sgdevs.has_key(dev):
                        scsiutil.scsi_dev_ctrl(sgdevs[dev], 'remove')
        else:
            util.SMlog("Missing md_svid key")
            return
        
    def updateSRMetadata(self):
        _trace("entry updateSRMetadata")
        try:
            # Add SR specific SR metadata            
            sr_info = \
            { 
              UUID_TAG: self.uuid,
              NAME_LABEL_TAG: util.to_plain_string(\
                                                        self.session.xenapi.SR.get_name_label(self.sr_ref)),
              NAME_DESCRIPTION_TAG: util.to_plain_string(\
                                self.session.xenapi.SR.get_name_description(self.sr_ref))
            }
            
            vdi_info = {}
            for vdi in self.session.xenapi.SR.get_VDIs(self.sr_ref):
                vdi_uuid = self.session.xenapi.VDI.get_uuid(vdi)
                
                # Create the VDI entry in the SR metadata
                vdi_info[vdi_uuid] = \
                {
                UUID_TAG: vdi_uuid, 
                   NAME_LABEL_TAG: util.to_plain_string(\
                        self.session.xenapi.VDI.get_name_label(vdi)),
                   NAME_DESCRIPTION_TAG: util.to_plain_string(\
                        self.session.xenapi.VDI.get_name_description(vdi)),
                   IS_A_SNAPSHOT_TAG: \
                    int(self.session.xenapi.VDI.get_is_a_snapshot(vdi)),
                   SNAPSHOT_OF_TAG: \
                        self.session.xenapi.VDI.get_snapshot_of(vdi),
                   SNAPSHOT_TIME_TAG: '',
                   TYPE_TAG: \
                        self.session.xenapi.VDI.get_type(vdi),
                   VDI_TYPE_TAG: \
                       '',
                    READ_ONLY_TAG: \
                        int(self.session.xenapi.VDI.get_read_only(vdi)),
                    MANAGED_TAG: \
                        int(self.session.xenapi.VDI.get_managed(vdi)),
                    METADATA_OF_POOL_TAG: \
                        self.session.xenapi.VDI.get_metadata_of_pool(vdi),
                    SVID_TAG: \
                       self.session.xenapi.VDI.get_sm_config(vdi)\
                        ['SVID'] 
                }
            
            SLMetadataHandler(self.metadataVolumePath).\
                writeMetadata(sr_info, vdi_info)
            
        except Exception, e:
            raise xs_errors.XenError('MetadataError', \
                         opterr='Error updating SR Metadata: %s' % str(e))
  
    def syncMetadataAndStorage(self):
        #_trace("entry syncMetadataAndStorage()")
        try:
            # if a VDI is present in the metadata but not in the storage
            # then delete it from the metadata                    
            (sr_info, vdi_info) = \
                SLMetadataHandler(self.metadataVolumePath, False).getMetadata()
            if sr_info.has_key(UUID_TAG):
                if(sr_info[UUID_TAG] != self.uuid):
                    _trace("SR in Metadata is corrupt. NO TRUST.")
                    raise
            else:
                _trace("SR in Metadata is empty. NO TRUST.")
                raise

            # Get a set of SVs from the array using the adapter
            if self.SVids == []:
                self.SVids = self.cslghandler.listStorageVolumes(self.ssid,self.pool)[0]

            for key in vdi_info.keys():
                if not vdi_info[key][SVID_TAG] in self.SVids:
                    # IP address could have changed which might have changed the SVID, 
                    # for ex, from NETAPP__LUN__0A5AC465__5033534256346930716B364E to NETAPP__LUN__0A5AC464__5033534256346930716B364E
                    # In that case, only compare the serial Ids
                    serialIDs = _stripSerialIds(self.SVids)
                    if not vdi_info[key][SVID_TAG][vdi_info[key][SVID_TAG].rfind('__')+2:] in serialIDs:
                        # delete this from metadata
                        SLMetadataHandler(self.metadataVolumePath).\
                            deleteVdiFromMetadata(vdi_info[key][UUID_TAG])
                    else:
                        # update this in metadata
                        vdi_info[key][SVID_TAG] = self.ssid + '__' + vdi_info[key][SVID_TAG][vdi_info[key][SVID_TAG].rfind('__')+2:] 
                        SLMetadataHandler(self.metadataVolumePath).\
                            updateVdi(vdi_info[key])
            
        except Exception, e:
            raise xs_errors.XenError('MetadataError', \
                opterr='Error synching SR Metadata and storage: %s' % str(e))

    def syncMetadataAndXapi(self):
        try:
            # get metadata
            (sr_info, vdi_info) = \
                SLMetadataHandler(self.metadataVolumePath, False).getMetadata()

            # First synch SR parameters
            self.update(self.uuid)

            # Now update the VDI information in the metadata if required
            for vdi_offset in vdi_info.keys():
                try:
                    vdi_ref = \
                        self.session.xenapi.VDI.get_by_uuid(\
                                        vdi_info[vdi_offset][UUID_TAG])
                except:
                    # may be the VDI is not in XAPI yet dont bother
                    continue
                
                new_name_label = util.to_plain_string(\
                       self.session.xenapi.VDI.get_name_label(vdi_ref))
                new_name_description = util.to_plain_string(\
                       self.session.xenapi.VDI.get_name_description(vdi_ref))

                if vdi_info[vdi_offset][NAME_LABEL_TAG] != new_name_label or \
                    vdi_info[vdi_offset][NAME_DESCRIPTION_TAG] != \
                    new_name_description:
                    update_map = {}
                    update_map[METADATA_UPDATE_OBJECT_TYPE_TAG] = \
                        METADATA_OBJECT_TYPE_VDI
                    update_map[UUID_TAG] = vdi_info[vdi_offset][UUID_TAG]
                    update_map[NAME_LABEL_TAG] = new_name_label
                    update_map[NAME_DESCRIPTION_TAG] = new_name_description
                    SLMetadataHandler(self.metadataVolumePath).\
                        updateMetadata(update_map)
        except Exception, e:
            raise xs_errors.XenError('MetadataError', \
                opterr='Error synching SR Metadata and XAPI: %s' % str(e))
                        
    def delete(self, sr_uuid):
        """ delete method """
        _trace("entry CSLG.delete()")
        assert(sr_uuid)

        VDIs = self.session.xenapi.SR.get_VDIs(self.sr_ref)
        if(len(VDIs)):
            raise xs_errors.XenError('SRNotEmpty', \
                opterr='Remove VDIs in this SR and retry')

        self.detach(sr_uuid)

        sm_config = self.session.xenapi.SR.get_sm_config(self.sr_ref)
        if self.sm_config.has_key('md_svid'):
            self.cslghandler.deleteStorageNode(self.sm_config['md_svid'])
        
        if self.sm_config.has_key('subpool'):
            _deleteNetappPool(self, self.sm_config['subpool']) 

    def attach(self, sr_uuid):
        """ attach method """
        _trace("entry CSLG.attach()")
        assert(sr_uuid)

        self._check_required_params()
        self._setMultipathableFlag()
        _insertUdevRules()

        # load storage system info
        try:
            self._load_ss_info()
            self._load_sr_info()
            self._load_sp_details()
        except:
            # SSID may have changed?
            raise xs_errors.XenError('CSLGInvalidSSID')

        ##Connection Protocol
        #cross check with allowed protocols from adapter and store in pbd
        pbd_connection = self.protocol
        if pbd_connection == 'FC':
            if not _check_host_fc(self):
                raise xs_errors.XenError('CSLGInvalidProtocol', \
                    opterr='Protocol(%s) not supported in Xenserver'%pbd_connection)
        else:
            pbd_connection = 'ISCSI'

        #Add connection protocol as PBD key for this host:
        protocol_list = self.allowed_protocols.split(",")
        if pbd_connection.lower() in protocol_list:
            pbd = None
            try:
                pbd = util.find_my_pbd(self.session, self.host_ref, self.sr_ref)
            except:
                pass

            _trace("pbd_connection : %s" %pbd_connection)
            if pbd != None:
                _trace("pbd_connection : %s" %pbd_connection)
                other_config = self.session.xenapi.PBD.get_other_config(pbd)
                other_config['connection'] = pbd_connection
                if pbd_connection == 'ISCSI':
                    (ports, targets) = _find_iscsi_targets(self.cslghandler, self.ssid, self.LUNperIQN)
                    other_config['ports'] = _toString(ports)
                    if not self.LUNperIQN:
                        other_config['targets'] = _toString(targets)
                self.session.xenapi.PBD.set_other_config(pbd, other_config)
        else:
            raise xs_errors.XenError('CSLGInvalidProtocol', \
                                    opterr='Protocol(%s) not supported in Storage System'%pbd_connection)
        self.protocol = pbd_connection

        ##subPool creation flag
        self.sm_config = self.session.xenapi.SR.get_sm_config(self.sr_ref)
        if self.isMaster == True and self.adapter == "NETAPP":
            self.sm_config['requires_subpool'] = (lambda:"True", lambda:"False")[self.parentpool != ""]()

        ##supports_resize Flag
        self.sm_config['supports_resize'] = (lambda:"True", lambda:"False")[self.supports_resize==0]()
        self.session.xenapi.SR.set_sm_config(self.sr_ref, self.sm_config)

        if self.isMaster:
            self.attachMetadataVolume()

            # Synchronise contents
            self.syncMetadataAndStorage()
            self.syncMetadataAndXapi()

            self._db_update()

    def probe(self):
        """ probe method """
        _trace("entry CSLG.probe()")
        xml_resp = self._check_required_params(False)
        if xml_resp != "":
            return xml_resp
        return self._make_srlist()

    def update(self, uuid):
        if self.sm_config.has_key('md_svid'):
            self.metadataVolumePath = \
                _check_path_and_reattach_if_req(self.sm_config['md_svid'], self)[0]

            # synch name_label in metadata with XAPI
            update_map = {METADATA_UPDATE_OBJECT_TYPE_TAG: \
                        METADATA_OBJECT_TYPE_SR,
                        NAME_LABEL_TAG: util.to_plain_string(\
                            self.session.xenapi.SR.get_name_label(self.sr_ref)),
                        NAME_DESCRIPTION_TAG: util.to_plain_string(\
                         self.session.xenapi.SR.get_name_description(self.sr_ref))
                        }
            SLMetadataHandler(self.metadataVolumePath).\
                updateMetadata(update_map)

    def detach(self, sr_uuid):
        """ detach method """
        _trace("entry CSLG.detach()")
        assert(sr_uuid)
 
        pbd = None
        try:
            pbd = util.find_my_pbd(self.session, self.host_ref, self.sr_ref)
        except:
            pass

        if pbd != None:
            ##get protocol
            oconf = self.session.xenapi.PBD.get_other_config(pbd)
            if oconf.has_key('connection'):
                self.connection = oconf['connection']

            ##remove 'connection' key in other_config
            self.session.xenapi.PBD.remove_from_other_config(pbd, 'connection')
            self.session.xenapi.PBD.remove_from_other_config(pbd, 'ports')
            self.session.xenapi.PBD.remove_from_other_config(pbd, 'targets')

        try:
            self.detachMetadataVolume()
        except:
            pass

    ###TODO: Integrate metadata.py utility functions for SR scan synchronisation
    def scan(self, sr_uuid):
        """ scan method """
        _trace("entry CSLG.scan()")
        assert(sr_uuid)

        # get data from service
        self._load_ss_info()
        self._load_sr_info()

        # Now check if there are any VDIs in the metadata, which are not in XAPI
        if self.sm_config.has_key('md_svid'):
            # get VDIs from XAPI
            vdis = self.session.xenapi.SR.get_VDIs(self.sr_ref)
            vdi_uuids = set([])
            for vdi in vdis:
                vdi_uuids.add(self.session.xenapi.VDI.get_uuid(vdi))

            self.metadataVolumePath = _check_path_and_reattach_if_req(self.sm_config['md_svid'], self)[0]
            
            (sr_info, vdi_info) = \
                SLMetadataHandler(self.metadataVolumePath, False).getMetadata()
            if sr_info.has_key(UUID_TAG):
                if(sr_info[UUID_TAG] != sr_uuid):
                    raise xs_errors.XenError('MetadataError', \
                        opterr='Error in SR Metadata storage')
            else:
                raise xs_errors.XenError('MetadataError', \
                    opterr='SR uuid missing in Metadata storage')
            vdiToSnaps = {}
            for key in vdi_info.keys():                
                vdi_uuid = vdi_info[key][UUID_TAG]
                if bool(int(vdi_info[key][IS_A_SNAPSHOT_TAG])):
                    if vdiToSnaps.has_key(vdi_info[key][SNAPSHOT_OF_TAG]):
                        vdiToSnaps[vdi_info[key][SNAPSHOT_OF_TAG]].append(vdi_uuid)
                    else:
                        vdiToSnaps[vdi_info[key][SNAPSHOT_OF_TAG]] = [vdi_uuid]
                            
                if vdi_uuid not in vdi_uuids:
                    util.SMlog("Introduce VDI %s as it is present in "\
                               "metadata and not in XAPI." % vdi_uuid)
                    sm_config = {}
                    sm_config['SVID'] = vdi_info[key][SVID_TAG]
                    ## which we may need to populate here.
                    vdi_ref = self.session.xenapi.VDI.db_introduce(
                                            vdi_uuid,
                                            vdi_info[key][NAME_LABEL_TAG],
                                            vdi_info[key][NAME_DESCRIPTION_TAG],
                                            self.sr_ref,
                                            vdi_info[key][TYPE_TAG],
                                            False,
                                            bool(int(vdi_info[key][READ_ONLY_TAG])),
                                            {},
                                            vdi_uuid,
                                            {},
                                            {})

                    self.session.xenapi.VDI.set_sm_config(vdi_ref,
                                                          sm_config)
                    self.session.xenapi.VDI.set_managed(vdi_ref,
                                        bool(int(vdi_info[key][MANAGED_TAG])))

                    (size, utilisation, label, lun_iqn, snapshot) = _get_usage(self, sm_config['SVID'])
        
                    if lun_iqn != "":
                        self.session.xenapi.VDI.add_to_sm_config(vdi_ref, 'LUNIQN',
                                                 lun_iqn)
            
                    #Add displayName to sm_config
                    if label != "":
                        self.session.xenapi.VDI.add_to_sm_config(vdi_ref, 'displayname',
                                                     label)

                    #Add subpool to sm_config(only for Netapp)
                    if self.sm_config.has_key('requires_subpool') and self.sm_config['requires_subpool'] == 'True':
                        #There must be a flexvol for this lun, get the poolid for that
                        poolName = label[label.find('Xen_'):label.rfind('/')]
                        poolInfo = self.cslghandler.getStoragePoolInfoByName(self.ssid, poolName)
                        pool = poolInfo.storagePoolId
               
                        # Now put this subpool in the sm_config
                        self.session.xenapi.VDI.add_to_sm_config(vdi_ref, 'subpool', pool)

                    ##virtual size
                    self.session.xenapi.VDI.set_virtual_size(vdi_ref, str(size))
                    ##physical utilisation
                    self.session.xenapi.VDI.set_physical_utilisation( \
                        vdi_ref, str(utilisation))

                    self.session.xenapi.VDI.set_is_a_snapshot( \
                        vdi_ref, bool(int(vdi_info[key][IS_A_SNAPSHOT_TAG])))
                    if bool(int(vdi_info[key][IS_A_SNAPSHOT_TAG])):
                        self.session.xenapi.VDI.set_snapshot_time( \
                                vdi_ref, DateTime(vdi_info[key][SNAPSHOT_TIME_TAG]))
                    if vdi_info[key][TYPE_TAG] == 'metadata':
                        self.session.xenapi.VDI.set_metadata_of_pool( vdi_ref, \
                                                vdi_info[key][METADATA_OF_POOL_TAG])

                try:
                    #remove from the vdi_uuids list, so that, we can remove balancing ones
                    vdi_uuids.remove(vdi_uuid)
                except:
                    pass
            
            # Now set the snapshot statuses correctly in XAPI
            for srcvdi in vdiToSnaps.keys():
                try:
                    srcref = self.session.xenapi.VDI.get_by_uuid(srcvdi)
                except:
                    # the source VDI no longer exists, continue
                    continue
                
                for snapvdi in vdiToSnaps[srcvdi]:
                    try:
                        # this might fail in cases where its already set
                        snapref = \
                            self.session.xenapi.VDI.get_by_uuid(snapvdi)
                        self.session.xenapi.VDI.set_snapshot_of(snapref, srcref)
                    except Exception, e:
                        util.SMlog("Setting snapshot failed. "\
                                   "Error: %s" % str(e))
                            
            for vdi_uuid in vdi_uuids:
                util.SMlog("Forget VDI %s as it is not present in \
                           metadata, but in XAPI." % vdi_uuid)
                vdi_ref = self.session.xenapi.VDI.get_by_uuid(vdi_uuid)
                self.session.xenapi.VDI.db_forget(vdi_ref)

        # Update the SR record
        self._db_update()

    def content_type(self, sr_uuid):
        """ content_type method """
        _trace("entry CSLG.content_type()")
        return super(CSLGSR, self).content_type(sr_uuid)

    def vdi(self, uuid):
        """ vdi method """
        _trace("entry CSLG.vdi()")
        assert(uuid)
        return CSLGVDI(self, uuid)


    ## local 

    #####TODO: Review whether any of this code is required
    def _getallvdirecords(self):
        """Helper method which returns a list of all VDI records for this SR
        stored in the XenAPI server"""
        _trace("entry CSLG._getallvdirecords()")

        # Returns a list of (reference, record) pairs: we only need the records
        vdis = self.session.VDI.get_all_records( \
                                 self.session_ref)['Value'].values()
        # We only need the VDIs corresponding to this SR
        return filter(lambda v: v['SR'] == self.sr_ref, vdis)

    def _loadvdis(self):
        """ _loadvdis method """
        _trace("entry CSLG._loadvdis()")
        if self.vdis:
            return

        # for this SR, simply validate the VDIs that Xen believes to be here
        # The list of VDIs comes from the XenAPI because we have no state
        for vdi in self._getallvdirecords():
            vdirec = CSLGVDI(self, vdi['uuid'])
            self.vdis[vdirec.uuid] = vdirec

    def _load_ss_info(self):
        """ get protocol and serial number from storage system """
        _trace("entry CSLG._load_ss_info()")

        (self.serial_number, self.allowed_protocols, \
             self.req_pool_cleanup, self.supports_resize, \
             cloneOfSnapSupported, self.LUNperIQN, self.supports_chap) = \
            self.cslghandler.getStorageSystemInfo(self.ssid)        

    def _load_sr_info(self):
        """ load sr info method from cslg service """
        _trace("entry CSLG._load_sr_info()")

        self.physical_size = 0
        self.physical_utilisation = 0
        self.virtual_allocation = 0
        sm_config = self.session.xenapi.SR.get_sm_config(self.sr_ref)

        if self.req_pool_cleanup and sm_config.has_key('sub-pools'):
            sub_pools = sm_config['sub-pools']
            sub_pool_list = sub_pools.split(",")
            for pool_index in range(0, len(sub_pool_list)-1):
                (physical_size, physical_utilisation, virtual_allocation) = \
                    self._get_sr_usage(self.ssid, sub_pool_list[pool_index])
                self.physical_size += physical_size
                self.physical_utilisation += physical_utilisation
                self.virtual_allocation += virtual_allocation
    
        else:
            (physical_size, physical_utilisation, virtual_allocation) = \
                self._get_sr_usage(self.ssid, self.pool)

            self.physical_size = physical_size
            self.physical_utilisation = physical_utilisation
            self.virtual_allocation = virtual_allocation

    def _load_sp_details(self):
        """ get some details of the pool and place in self """
        _trace("entry CSLG._load_sp_details()")
        supported_raid = []
        (physical_size, physical_utilisation, virtual_allocation, parentStoragePoolId, friendlyName, supported_raid) = self.cslghandler.getStoragePoolInfo(self.ssid, self.pool)
        self.parentpool = parentStoragePoolId

    def _get_sr_usage(self, ssid, pool):
        """ get specific size/usage of pool """
        _trace("entry CSLG._get_sr_usage()")

        tuple = self.cslghandler.getStoragePoolInfo(ssid, pool)
        util.SMlog("Output from getStoragePoolInfo: %s, %s, %s, %s" % (tuple[0], tuple[1], tuple[2], tuple[3]))
        self.parentpool = tuple[3]

        return (tuple[0], tuple[1], tuple[2])

    def _make_srlist(self):
        """ _make_sr_exists method """
        _trace("entry CSLG._make_srlist()")
        #
        # scan for any of the LUNs with following name_label and return all SR UUIDs discovered
        # SL_XenStorage-<SR UUID>
        #
        try:
            name_labels = self.cslghandler.listStorageVolumes(self.ssid, self.pool)[1]

            SRs = []
            Sizes = []
            pattern = "SL_XenStorage__"
            for label in name_labels:
                pos = label.rfind(pattern)
                if pos != -1:
                    uuid = (label[pos+len(pattern):]).replace("_","-")
                    SRs.append(uuid)

            if SRs == []:
                SR = self.session.xenapi.SR.get_all_records_where("field \"type\" =\"cslg\"")
                for sr in SR:
                    SRs.append(SR[sr]['uuid'])
                    Sizes.append(SR[sr]['physical_size'])
            if Sizes == []:
                Sizes = [0]*len(SRs)

            dom = xml.dom.minidom.Document()
            srlist_tag = dom.createElement("SRlist")
            dom.appendChild(srlist_tag)

            for index in range(len(SRs)):
                sr_tag = dom.createElement("SR")
                srlist_tag.appendChild(sr_tag)
                uuid_tag = dom.createElement("UUID")
                sr_tag.appendChild(uuid_tag)
                size_tag = dom.createElement("Size")
                sr_tag.appendChild(size_tag)

                ##add uuid & size values 
                textnode = dom.createTextNode(str(SRs[index]))
                uuid_tag.appendChild(textnode)

                textnode = dom.createTextNode(str(Sizes[index]))
                size_tag.appendChild(textnode)

            return dom.toprettyxml()

        except Exception, e:
            dom = xml.dom.minidom.Document()
            return dom.toprettyxml()

    def _probe_ssid(self):
        """ _probe_ssid method """
        return self.cslghandler.listStorageSystems()

    def _probe_spid(self):
        """ _probe_spid method """
        return self.cslghandler.listStoragePools(self.ssid, True)

    def _probe_luns(self):
        """ _probe_luns method """
        return self.cslghandler.listStorageVolumes(self.ssid, self.pool, False, True)

    def _check_required_params(self, raiseException=True):
        """ _check_required_params method """
        _trace("entry CSLG._check_required_params()")
        _trace("  self.adapter = %s" % self.adapter )
        if (not self.adapter) or (not self.cslghandler.adapter_inf):
           # Check for adapterID, if not throw supported adapters
           adap_lists = _list_adapter_names()
           if len(adap_lists) ==0 :
               _trace("There's no support for any storage adapters")
               raise xs_errors.XenError('CSLGConfigAdapterMissing', opterr='There is no support for any Storage Adapter')
           else:
               adaps = ""
               try:
                   dom = xml.dom.minidom.Document()
                   adaplist_tag = dom.createElement("AdapterList")
                   dom.appendChild(adaplist_tag)
                   for adap in adap_lists:
                       self.adapter = adap
                       #if adap == 'SMIS_STORAGE_SYSTEM':
                       #    continue
                       self.cslghandler = CSLGHandler.CSLGHandlerInternal(self)
                       (name, fndly_name, desc, type, repType, verMaj, verMin) = self.cslghandler.getStorageAdapterInfo()
                       _trace("Adapter info name: %s, fndly_name: %s, desc: %s, type: %s, repType: %s, verMaj: %s, verMin: %s" %(name, fndly_name, desc, type, repType, verMaj, verMin))

                       adap_tag = dom.createElement("Adapter")
                       adaplist_tag.appendChild(adap_tag)
                       createXMLNode(dom, adap_tag, "Name", name)
                       createXMLNode(dom, adap_tag, "FriendlyName", fndly_name)
                       createXMLNode(dom, adap_tag, "Description", desc)
                       createXMLNode(dom, adap_tag, "Type", type)
                       createXMLNode(dom, adap_tag, "ReplicationType", repType)
                       createXMLNode(dom, adap_tag, "VersionMajor", verMaj)
                       createXMLNode(dom, adap_tag, "VersionMinor", verMin)
                   _trace("Adapter List XML %s" %dom)
                   if raiseException == True:
                       raise xs_errors.XenError('CSLGConfigAdapterMissing', opterr='Supported Adapters are %s' %dom.toprettyxml())
                   else:
                       return dom.toprettyxml()
               except util.CommandException, inst:
                   dom = xml.dom.minidom.Document()
                   raise xs_errors.XenError('CSLGConfigAdapterMissing', opterr='Error in detecting  Adapters %s' %dom.toprettyxml())

        # Check for Target IP address, user and password parameters, else throw:
        #     CSLGAdapterTargetMissing
        #     CSLGAdapterUsernameMissing
        #     CSLGAdapterPasswordMissing
        if not self.cslg:
            raise xs_errors.XenError('CSLGConfigServerMissing')
        if not self.cslguser:
            raise xs_errors.XenError('CSLGConfigUsernameMissing')
        #array passwords can be empty. so, not checking for that
        if not self.cslgpassword:
            raise xs_errors.XenError('CSLGConfigPasswordMissing')
        if not self.ssid:
            xml_resp = self.cslghandler.listStorageSystems()
            _trace("The SSID list XML : %s" %xml_resp)
            if raiseException == True:
                raise xs_errors.XenError('CSLGConfigSSIDMissing', opterr='%s' %xml_resp)
            else:
                return xml_resp
        if not self.pool:
            xml_resp = self.cslghandler.listStoragePools(self.ssid)
            _trace("The SPID list XML : %s" %xml_resp)
            if raiseException == True:
                raise xs_errors.XenError('CSLGConfigPoolIDMissing', opterr='Available SPIDs: %s' %xml_resp)
            else:
                return xml_resp
        if self.protocol == 'FC':
            if not _check_host_fc(self):
                raise xs_errors.XenError('CSLGInvalidProtocol', \
                    opterr='Protocol(%s) not supported in Xenserver'%self.protocol)
        return ""

class CSLGVDI(VDI.VDI):
    """ CSLG VDI class definitions """

    def load(self, vdi_uuid):
        """ load method """
        _trace("entry CSLGVDI.load()")
        self.host_iqn = ''
        self.host_devpath = []
        self.host_adt = {}
        # always use multipath?
        self.multihomed = True
        self.attached = False
        # can be a list!
        self.iscsi=[]
        self.iscsi_targets = []
        self.iscsi_target = ''
        self.svid = ''
        self.lun = None
        self.SCSIid = None
        self.ipname = ''
        self.path = None
        self.vdi_iqn = None
        self.vdi_ref = None
        self.ports = None
        self.targets = None

        _trace("DEBUG: sr.srcmd.params: %s" %self.sr.srcmd.params)

        # vdi record exists yet?
        if self.sr.srcmd.params.has_key('vdi_ref'):
            self.vdi_ref = self.sr.srcmd.params['vdi_ref']
        else:
            try:
                self.vdi_ref = self.session.xenapi.VDI.get_by_uuid(vdi_uuid)
            except:
                self.vdi_ref = None

        if self.vdi_ref:
            # TODO: can repopulate as required?
            u = False
            if self.sr.cmd == "sr_scan":
                u = True
            self._load_vdi_info(u)

        # determine protocol for this host/SR
        self._determine_protocol()
        util.SMlog("SR protocol: %s" % self.sr.protocol)
        
        if self.sr.isMaster:
            self.metadataVolumePath = _check_path_and_reattach_if_req(self.sr.sm_config['md_svid'], self.sr)[0]

    def create(self, sr_uuid, vdi_uuid, size):
        """ create method """
        _trace("entry CSLGVDI.create()")
        assert(sr_uuid)
        assert(vdi_uuid)
        self.uuid = vdi_uuid
        self.label = util.to_plain_string(self.sr.srcmd.params['args'][1])
        self.description = util.to_plain_string(self.sr.srcmd.params['args'][2])
        
        # Check for requires_subpool from adapter
        # if true, then create a pool   
        sr_sm_config = self.session.xenapi.SR.get_sm_config(self.sr.sr_ref)
        if sr_sm_config.has_key('requires_subpool') and sr_sm_config['requires_subpool'] == 'True':
            self.sr.pool = _createNetappPool(self.sr, self.label, vdi_uuid, size) 

        self.sm_config = self.sr.srcmd.params['vdi_sm_config']

        self._allocate(vdi_uuid, size, self.label)
        
        # get new utilisation, in case the actual LUN is not 
        # the exact size we asked for.
        (self.size, self.utilisation, self.vol_label,lun_iqn, snapshot) = \
                             _get_usage(self.sr, self.svid)
        self.managed = True

        SLMetadataHandler(self.metadataVolumePath).ensureSpaceIsAvailableForVdis(1)
        self._db_introduce()

        self.vdi_ref = self.session.xenapi.VDI.get_by_uuid(vdi_uuid)
        if self.sm_config.has_key('SVID'):
            _trace("CSLGVDI.create(), SVID already in sm-config: %s new is %s (removing former)" \
                   % (self.sm_config["SVID"], self.svid))
            self.session.xenapi.VDI.remove_from_sm_config(self.vdi_ref, 'SVID')

        self.session.xenapi.VDI.add_to_sm_config(self.vdi_ref, 'SVID', self.svid)

        if self.vdi_iqn is not None:
            if self.sm_config.has_key('LUNIQN'):
                _trace("CSLGVDI.create(), LUNIQN already in sm-config: %s new is %s (removing former)" \
                       % (self.sm_config["LUNIQN"], self.vdi_iqn))
                self.session.xenapi.VDI.remove_from_sm_config(self.vdi_ref, 'LUNIQN')

            self.session.xenapi.VDI.add_to_sm_config(self.vdi_ref, 'LUNIQN', 
                                                 self.vdi_iqn) 

        # Now put the subpool(if created) in the sm_config
        if sr_sm_config.has_key('requires_subpool') and sr_sm_config['requires_subpool'] == 'True':
            if self.sm_config.has_key('subpool'):
                _trace("CSLGVDI.create(), subpool already in sm_config, so replacing it.")
                self.session.xenapi.VDI.remove_from_sm_config(self.vdi_ref, 'subpool')
            self.session.xenapi.VDI.add_to_sm_config(self.vdi_ref, 'subpool', self.sr.pool)

        # Now put displayName in sm_config
        if self.sm_config.has_key('displayname'):
            _trace("CSLGVDI.create(), displayname already in sm_config, so replacing it.")
            self.session.xenapi.VDI.remove_from_sm_config(self.vdi_ref, 'displayname')
        self.session.xenapi.VDI.add_to_sm_config(self.vdi_ref, 'displayname', self.vol_label)

        # Now add this VDI to metadata
        if self.sr.sm_config.has_key('md_svid'):
            vdi_info = {  UUID_TAG : self.uuid,
                        NAME_LABEL_TAG: self.label,
                        NAME_DESCRIPTION_TAG: self.description,
                        IS_A_SNAPSHOT_TAG: 0,
                        SNAPSHOT_OF_TAG: '',
                        SNAPSHOT_TIME_TAG: '',
                        TYPE_TAG: self.sr.srcmd.params['vdi_type'],
                        VDI_TYPE_TAG: 'aio',
                        READ_ONLY_TAG: int(self.read_only),
                        MANAGED_TAG: int(self.managed),
                        METADATA_OF_POOL_TAG: '',
                        SVID_TAG: \
                             self.session.xenapi.VDI.get_sm_config( \
                                self.vdi_ref)['SVID']
            }
            util.SMlog("calling addVDI with info: %s" % vdi_info)
            SLMetadataHandler(self.metadataVolumePath).addVdi(vdi_info)

        return super(CSLGVDI, self).get_params()

    def update(self, sr_uuid, vdi_uuid):
        if self.sr.sm_config.has_key('md_svid'):
            #Synch the name_label of this VDI on storage with the name_label in XAPI        
            update_map = {}
            update_map[METADATA_UPDATE_OBJECT_TYPE_TAG] = \
                METADATA_OBJECT_TYPE_VDI
            update_map[UUID_TAG] = self.uuid
            update_map[NAME_LABEL_TAG] = util.to_plain_string(\
                self.session.xenapi.VDI.get_name_label(self.vdi_ref))
            update_map[NAME_DESCRIPTION_TAG] = util.to_plain_string(\
                self.session.xenapi.VDI.get_name_description(self.vdi_ref))
            update_map[SNAPSHOT_TIME_TAG] = \
                self.session.xenapi.VDI.get_snapshot_time(self.vdi_ref)
            update_map[METADATA_OF_POOL_TAG] = \
                self.session.xenapi.VDI.get_metadata_of_pool(self.vdi_ref),
            update_map[MANAGED_TAG] = \
                int(self.session.xenapi.VDI.get_managed(self.vdi_ref))
            metadataVolPath = \
                       _check_path_and_reattach_if_req(self.sr.sm_config['md_svid'], self.sr)[0]
            SLMetadataHandler(self.metadataVolumePath).updateMetadata(update_map)

    def delete(self, sr_uuid, vdi_uuid):
        """ delete method """
        _trace("entry CSLGVDI.delete()")
        assert(sr_uuid)
        assert(vdi_uuid)

        if self.attached:
            raise xs_errors.XenError('VDIInUse')

        # Mark VDI as unmanaged in XAPI
        vdi_ref = self.session.xenapi.VDI.get_by_uuid(vdi_uuid)
        self.session.xenapi.VDI.set_managed(vdi_ref, False)
        self.managed = False

        # Mark VDI as unmanaged in Metadata
        if self.sr.sm_config.has_key('md_svid'):
            self.update(self.sr.uuid,vdi_uuid)

        # Now figure out whether to actually delete
        self._deletion_check(vdi_ref)
    
    def _deletion_check(self, vdi_ref):
        """ delete check method """
        _trace("entry CSLGVDI.deletion_check()")

        # Find the tree root
        _trace( "DEBUG: Find tree root for %s"%( self.session.xenapi.VDI.get_uuid(vdi_ref) ) )
        snap = True
        _ref = vdi_ref
        try:
            while snap:
                snap = self.session.xenapi.VDI.get_is_a_snapshot(_ref)
                if snap:
                    _trace( "DEBUG:     ref IS snap: %s"%( self.session.xenapi.VDI.get_uuid(_ref) ) )
                    temp_ref = self.session.xenapi.VDI.get_snapshot_of(_ref)
                    _trace( "DEBUG:       parent is: %s"%( self.session.xenapi.VDI.get_uuid(temp_ref) ) )
                    _ref = temp_ref
                else:
                    _trace( "DEBUG:     ref is NOT snap: %s"%( self.session.xenapi.VDI.get_uuid(_ref) ) )
        except:
            #xapi might throw exception, in case vdi_ref is invalid. so, safe to stop here & consider _ref is primary
            pass
        primary = _ref
        _trace( "DEBUG:   Primary ref = %s"%( self.session.xenapi.VDI.get_uuid(primary) ) )

        # Build the unmanaged snapshot list for pruning
        _trace( "DEBUG: Build unmanaged snapshot list for pruning ..." )
        tree = []
        search = [primary]
        while len(search):
            newsearch = []
            for s in search:
                _trace( "DEBUG: Test vdi %s"%(self.session.xenapi.VDI.get_uuid(s) ) )
                if self._is_all_children_unmanaged(s):
                    if self.session.xenapi.VDI.get_managed(s):
                        _trace( "DEBUG:   vdi IS managed, do not add to delete list: %s"%( self.session.xenapi.VDI.get_uuid(s) ) )
                    else:
                        _trace( "DEBUG:   vdi NOT managed, add to delete list: %s"%( self.session.xenapi.VDI.get_uuid(s) ) )
                        tree.append(s)
                        _trace( "DEBUG:   Append: %s"%( self.session.xenapi.VDI.get_uuid(s) ) )
                newsearch += self.session.xenapi.VDI.get_snapshots(s)
            search = newsearch

        # We can delete the tree
        for _ref in reversed(tree):
            self._delete_vdi(_ref)

        _trace("exit CSLGVDI.deletion_check()")
    
    def _is_all_children_unmanaged(self, vdi_ref):
        children = self.session.xenapi.VDI.get_snapshots(vdi_ref)
        _trace( "DEBUG:   Children of vdi_ref %s: %s"%(self.session.xenapi.VDI.get_uuid(vdi_ref), children ) )
        for _ref in children:
            _trace( "DEBUG:     uuid:%s managed:%s"%( self.session.xenapi.VDI.get_uuid(_ref),
                                                      self.session.xenapi.VDI.get_managed(_ref) ) )
            if self.session.xenapi.VDI.get_managed(_ref):
                _trace( "DEBUG:       _ref IS managed, return False, all children NOT unmanaged" )
                return False
        _trace( "DEBUG:     Return True, all children ARE unmanaged" )
        return True

    def _delete_vdi(self, vdi_ref):
        sm_config = self.session.xenapi.VDI.get_sm_config(vdi_ref)
        vdi_uuid = self.session.xenapi.VDI.get_uuid(vdi_ref)
        if sm_config.has_key('SVID'):
            # Now execute the sv-destroy
            try:
                self._deallocate(sm_config['SVID'])
                if sm_config.has_key('subpool'):
                    _deleteNetappPool(self.sr, sm_config['subpool'])
            except:
                util.SMlog("Failed to deallocate node %s, ignoring..." % sm_config['SVID'])
                pass
            self.session.xenapi.VDI.db_forget(vdi_ref)
            if sm_config.has_key('SCSIid'):
                SCSIid = sm_config['SCSIid']
                key = "scsi-" + SCSIid
                try:
                    self.session.xenapi.SR.remove_from_sm_config(self.sr.sr_ref, key)
                except:
                    pass
        if self.sr.sm_config.has_key('md_svid'):
            metadataVolPath = _check_path_and_reattach_if_req(self.sr.sm_config['md_svid'], self.sr)[0]
            if metadataVolPath != None:
                SLMetadataHandler(self.metadataVolumePath).\
                    deleteVdiFromMetadata(vdi_uuid)

    def attach(self, sr_uuid, vdi_uuid):
        """ attach method """
        _trace("entry CSLGVDI.attach()")
        assert(sr_uuid)
        assert(vdi_uuid)

        attachedby = self.sr.protocol
        self.sr.sr_uuid = sr_uuid

        self.sm_config = self.session.xenapi.VDI.get_sm_config(self.vdi_ref)
        # check for LUN per IQN, iscsi login eschewed

        #assign lun & load transport
        #if self.sm_config.has_key('LUNIQN'):
        optflags = _getOptFlags(self, self.sr.protocol)
        # modify _assign to also populate self.lunID
        self._assign(vdi_uuid, optflags)
        self._load_transport(self.sr.protocol)
        # HP - not sure why this is here, the order is absolutely wrong
        # commenting it until we know if we really need this.
        #else:
            # Load transport session
            #self._load_transport(self.sr.protocol)
            #optflags = _getOptFlags(self, self.sr.protocol)
            #self._assign(vdi_uuid, optflags)

        #wait for lun to appear on host
        devs = []
        if self.sr.protocol == "ISCSI":
            # this will return a list of scsi devs
            _trace( "vdi.attach: protocol is iSCSI, _rescan_for_LUN, svid: %s" % self.svid )
            devs = _rescan_for_LUN(self.svid, self.sr)

        if not len(devs):
            # It seems LUN was not found, so try with rescan
            _trace( "vdi.attach: empty dev list, _rescan_for_LUN, svid: %s, host_devpath: %s" % (self.svid, self.host_devpath) )
            devs = _rescan_for_LUN(self.svid, self.sr, self.host_devpath)

        _trace( "vdi.attach: devs from rescan = %s" % devs )

        if not len(devs):
            # the lun failed to appear, so unassign it and throw an exception
            _trace( "vdi.attach: lun failed to appear, unassign and throw" )
            self._unassign(optflags)
            # Additional debug for troubleshooting if LUNs fail to appear
            cmd = ['ls', '-lR' , '/dev/disk/']
            op = util.pread2(cmd)
            cmd = ['bash', '-c', 'ls -l /dev/sd*']
            op = util.pread2(cmd)
            util.SMlog("vdi.attach: throw VDIUnavailable" )
            raise xs_errors.XenError('VDIUnavailable', \
                      opterr='path for VDI not found')

        # Generate the SCSIid
        (self.path,self.SCSIid, npaths) = _check_path_and_raise(self.svid, self.sr, devs)

        # We store the VDI SCSIid in the SR sm_config
        key = "scsi-" + self.SCSIid
        try:
            self.session.xenapi.SR.add_to_sm_config(self.sr.sr_ref, key, vdi_uuid)
            cmd = ['/opt/xensource/sm/mpathcount.py', self.SCSIid]
            util.pread2(cmd)
        except:
            pass

        # We additionally store some attachtime state about the VDI in the sm_config
        try:
            self.sm_config["SCSIid"] = self.SCSIid
            self.sm_config["attached-by"] = attachedby
            self.session.xenapi.VDI.set_sm_config(self.vdi_ref, self.sm_config)
        except:
            _trace("Warning: failed to update vdi sm_config record")
            pass

        # Query the SCSI page data
        scsiutil.update_XS_SCSIdata(vdi_uuid, \
                                    scsiutil.gen_raw_page_data(self.path))

        # In case this VDI comes from legacy SL, displayName will not be present in sm_config. so, add it.
        if not self.sm_config.has_key('displayname'):
            _trace("CSLGVDI.attach(), displayname not in sm_config, this VDI might be from legacy SL.")
            self.session.xenapi.VDI.add_to_sm_config(self.vdi_ref, 'displayname', _get_usage(self.sr, self.svid)[2])

        self.attached = True

        return super(CSLGVDI, self).attach(sr_uuid, vdi_uuid)

    def detach(self, sr_uuid, vdi_uuid):
        """ detach method """
        _trace("entry CSLGVDI.detach(%s)" % self.svid)
        assert(sr_uuid)
        assert(vdi_uuid)

        vdi_ref = self.sr.srcmd.params['vdi_ref']
        self.sm_config = self.session.xenapi.VDI.get_sm_config(self.vdi_ref)
        util.SMlog("sm_config: %s" % self.sm_config)

        #check if VDI is not attached
        if not self.sm_config.has_key('SVID'):
            util.SMlog('No SVID key')
            return
        svid = _find_LUN(self.sm_config['SVID'])
        if not len(svid):
            util.SMlog('No svid')
            return

        if self.sm_config.has_key('attached-by'):
            attachedby = self.sm_config['attached-by']
            self.session.xenapi.VDI.remove_from_sm_config(self.vdi_ref, 'attached-by')
        else:
            attachedby = self.sr.protocol

        self.sr.sr_uuid = sr_uuid
        #self._load_transport(attachedby)

        optflags = _getOptFlags(self, attachedby)
        SCSIid = ""
        if self.sm_config.has_key('SCSIid'):
            SCSIid = self.sm_config['SCSIid']

        if self.sr.mpath == "true" and len(SCSIid):
            self.sr.mpathmodule.reset(SCSIid, True) # explicit unmap
            util.remove_mpathcount_field(self.sr.session, self.sr.host_ref, \
                                         self.sr.sr_ref, SCSIid)

        # if LUNIQN set, we need to detach before unassignment
        #util.SMlog("attachedby: %s, sm_config: %s" % (attachedby, self.sm_config))
        #if attachedby == "ISCSI":
        #    if self.sm_config.has_key('LUNIQN') == True:
        #        for instance in self.iscsi:
        #            instance.detach(self.uuid)
        #        self._unassign(optflags)
        #    else:
        #        self._unassign(optflags)
        #        for instance in self.iscsi:
        #            instance.detach(self.uuid)
        #else:
        #    self._unassign(optflags)
        if attachedby == "ISCSI":
            self.iscsi = _load_iscsi_transport(self.sr, self.sr.ssid, \
                                                              self.sr.original_srcmd,\
                                                              self.vdi_iqn, \
                                                              self.uuid, self.sr.mpath)
            for instance in self.iscsi:
                instance.detach(self.uuid)
        self._unassign(optflags)
        
        # Now that this has been unmapped on the array, remove it
        self._remove_from_host(SCSIid)
        self.attached = False


    #####TODO: rely on VDI is_a_snapshot status field to determine whether 
    ##### source VDI is a snap. The adpater reports the array capabilities
    ##### so we know whether this op will be supported
    def _check_for_clone_OK(self):
        """ check if OK to clone this """
        _trace("entry CSLGVDI._check_for_clone_OK()")

        # if we are trying to clone a SVID that is already a
        # snapshot, then the array must also support this.
        snapshot = self.sr.cslghandler.getStorageVolumeInfo(self.sr.ssid, self.svid)[4]

        if snapshot == "Yes":
            # source of clone is a snapshot -- see if array 
            # supports cloning a snapshot
            
            try:
                if not self.sr.cslghandler.getStorageSystemInfo(self.sr.ssid)[4]:
                    # raising unimplemented should make xapi do an old-fashioned
                    # clone -- by dd'ing
                    _trace("CSLGVDI._check_for_clone_OK() Array does not support clone from snapshot")
                    raise xs_errors.XenError('Unimplemented')
            except util.CommandException, inst:
                _trace(inst, errorcase = 1)
                raise xs_errors.XenError('CSLGSnapClone', \
                         opterr=inst.reason)
            except:
                raise

    def clone(self, sr_uuid, vdi_uuid):
        """ clone method """
        _trace("entry CSLGVDI.clone()")
        assert(sr_uuid)
        assert(vdi_uuid)

        # will raise if not OK, or other error
        self._check_for_clone_OK()

        return self._snapclone('sv-clone-create')

    def snapshot(self, sr_uuid, vdi_uuid):
        """ snapshot method """
        _trace("entry CSLGVDI.snapshot()")
        assert(sr_uuid)
        assert(vdi_uuid)

        return self._snapclone('sv-snap-create')
 
    def introduce(self, sr_uuid, vdi_uuid):
        """ introduce method """
        _trace("entry CSLGVDI.introduce()")
        assert(sr_uuid)
        assert(vdi_uuid)

        lun_iqn = ""
        # in order to be introduced, we need the SVID
        self.sm_config = self.sr.srcmd.params['vdi_sm_config']
        self.svid = self.sm_config['SVID']
        if self.svid!='':
            try:
                lun_iqn = _validate_svid(self.sr, self.svid)
            except:
                raise xs_errors.XenError('CSLGIntroduce', \
                     opterr='Invalid SVID')
        else:
            raise xs_errors.XenError('CSLGConfigSVIDMissing')
        
        self.uuid = vdi_uuid
        (self.size, self.utilisation, self.label, lun_iqn, snapshot) = \
                             _get_usage(self.sr, self.svid)

        try:
            SLMetadataHandler(self.metadataVolumePath).\
                ensureSpaceIsAvailableForVdis(1)
            self._db_introduce()
            
            self.vdi_ref = self.session.xenapi.VDI.get_by_uuid(vdi_uuid)
            #Add LUNIQN to sm_config
            if lun_iqn != "":
                self.session.xenapi.VDI.add_to_sm_config(self.vdi_ref, 'LUNIQN',
                                                     lun_iqn)
            #Add displayName to sm_config
            if self.label != "":
                self.session.xenapi.VDI.add_to_sm_config(self.vdi_ref, 'displayname',
                                                     self.label)

            #Add subpool to sm_config(only for Netapp)
            if self.sr.sm_config.has_key('requires_subpool') and self.sr.sm_config['requires_subpool'] == 'True':
                #There must be a flexvol for this lun, get the poolid for that
                poolName = self.label[self.label.find('Xen_'):self.label.rfind('/')]
                poolInfo = self.sr.cslghandler.getStoragePoolInfoByName(self.sr.ssid, poolName)
                self.pool = poolInfo.storagePoolId
               
                # Now put this subpool in the sm_config
                self.session.xenapi.VDI.add_to_sm_config(self.vdi_ref, 'subpool', self.pool)

        except Exception, e:
            raise xs_errors.XenError('CSLGIntroduce', opterr=e)

        # The amend vdi record in XAPI
        if self.sr.sm_config.has_key('md_svid'):
            vdi_info = { UUID_TAG: self.uuid,
                                NAME_LABEL_TAG: self.label, 
                                NAME_DESCRIPTION_TAG: self.description,
                                IS_A_SNAPSHOT_TAG: 0,
                                SNAPSHOT_OF_TAG: '',
                                SNAPSHOT_TIME_TAG: '',
                                TYPE_TAG: 'user',
                                VDI_TYPE_TAG: 'aio',
                                READ_ONLY_TAG: 0,
                                MANAGED_TAG: int(self.managed),
                                'metadata_of_pool': '',
                                SVID_TAG: self.sm_config['SVID']
                }
            util.SMlog("calling updateVDI with info: %s" % vdi_info)
            try:
                SLMetadataHandler(self.metadataVolumePath).updateVdi(vdi_info)
            except Exception, e:
                _trace("Exception updateVDI - %s"%e)
                _trace("VDI record may be missing in Metadata, hence add it, calling addVdi")
                SLMetadataHandler(self.metadataVolumePath).addVdi(vdi_info)

        return super(CSLGVDI, self).get_params()

    def resize(self, sr_uuid, vdi_uuid, size):
        """ Resize a LUN """

        #sr_sm_config = self.session.xenapi.SR.get_sm_config(self.sr.sr_ref)
        
        # if the array's plugin does not support this, report an error
        if self.sr.sm_config.has_key('supports_resize') and self.sr.sm_config['supports_resize'] == 'False':
            raise xs_errors.XenError('Unimplemented', \
                     opterr='Array plugin does not support resize')

        # extract the SVID
        if self.vdi_ref == None:
            raise xs_errors.XenError('VDIResize', \
                  opterr='failed to extract SVID from vdi record')

        self._load_vdi_info(True)
        #current_size = int(self.session.xenapi.VDI.get_virtual_size(self.vdi_ref))
        current_size = self.size

        _trace("actual_size:%s requested_size:%s"% (current_size, size))

        # Disallow LUN shrinking
        if size < current_size:
            util.SMlog('vdi_resize: shrinking not supported: ' + \
                       '(current size: %d, new size: %d)' % (current_size, size))
            raise xs_errors.XenError('VDISize', opterr='shrinking not allowed')

        # round up if necessary
        size = util.roundup(MIN_ALLOCATION, size)

        # attempt to resize, if there is not enough space, the 
        # service will return an error, normalize on MB. Roundup
        # will accomodate and rounding issues

        size = size / (1024*1024)
        self.sr.cslghandler.resizeStorageNode(self.sr.ssid, self.svid,size)
        
        # reget get new utilisation
        (self.size, self.utilisation, self.label, lun_iqn, snapshot) = \
                             _get_usage(self.sr, self.svid)

        # check if resize succeeded?
        size = size * (1024*1024)
        if self.size < size:
            raise xs_errors.XenError('VDIResize', opterr='Probably Insufficient disk space in storage')

        self._db_update()
        return super(CSLGVDI, self).get_params()

    def _activate(self, sr_uuid, vdi_uuid):
        """ activate a IQN per LUN VDI """

    ###Replace with generic LUN scanning code
    def _determine_protocol(self):
        """ determine protocol method """
        _trace("entry CSLGVDI._determine_protocol()")

        (pbd_ref,rec) = util.find_my_pbd_record(self.session,self.sr.host_ref,self.sr.sr_ref)
        oconf = rec['other_config']

        # if already set
        if self.sr.protocol==None or self.sr.protocol=='':
            self.sr.protocol = oconf['connection']

        # if 'connection' is set in pbd, then use that
        if oconf.has_key('ports'):
            self.sr.ports = _fromString(oconf['ports'])
        if oconf.has_key('targets'):
            self.sr.targets = _fromString(oconf['targets'])

    def _snapclone(self, operation):
        """ common snap/clone method """
        _trace("entry CSLGVDI._snapclone() --> %s" % operation)

        snap_svid = self.sr.cslghandler.createCloneOrSnapshot(operation, self.sr.ssid, self.svid, self.sr.provision_type)
        (self.size, self.utilisation, label, lun_iqn, snapshot) = \
             _get_usage(self.sr, snap_svid)

        # create new vdi structure
        snap_uuid = util.gen_uuid ()
        snap_vdi = VDI.VDI(self.sr, snap_uuid)
        snap_vdi.size = self.size
        snap_vdi.utilisation = self.utilisation

        name_label = util.to_plain_string(\
                                                self.session.xenapi.VDI.get_name_label(self.vdi_ref))
        #name_label = name_label + ".snap"
        snap_vdi.label = name_label
        snap_vdi.shareable = self.shareable
        snap_vdi.read_only = self.read_only

        # introduce to XAPI
        try:
            snap_vdi.vdi_ref = snap_vdi._db_introduce()
            util.SMlog("vdi_snapshot: introduced VDI: %s" % snap_vdi.vdi_ref)
        except Exception, excp:
            util.SMlog("vdi_snapshot: error VDI.db_introduce: %s" % (str(excp)))
            raise

        # jam svid into VDI
        self.session.xenapi.VDI.add_to_sm_config(snap_vdi.vdi_ref, \
                                                 'SVID', snap_svid)
        # and lun_iqn if present
        if lun_iqn != "":
            self.session.xenapi.VDI.add_to_sm_config(snap_vdi.vdi_ref, \
                                                 'LUNIQN', lun_iqn)
        # Now put displayName in sm_config
        self.session.xenapi.VDI.add_to_sm_config(snap_vdi.vdi_ref, 'displayname', label)
       
        self.vdi_ref = snap_vdi.vdi_ref
        if self.sr.sm_config.has_key('md_svid'):
            if operation == 'sv-snap-create':
                vdi_info = { UUID_TAG: snap_uuid,
                                NAME_LABEL_TAG: snap_vdi.label,
                                NAME_DESCRIPTION_TAG: util.to_plain_string(\
                                  self.session.xenapi.VDI.get_name_description\
                                  (self.vdi_ref)),
                                IS_A_SNAPSHOT_TAG: 1,
                                SNAPSHOT_OF_TAG: self.uuid,
                                SNAPSHOT_TIME_TAG: '',
                                TYPE_TAG: \
                                 self.session.xenapi.VDI.get_type(self.vdi_ref),
                                VDI_TYPE_TAG: 'aio',
                                READ_ONLY_TAG: 0,
                                MANAGED_TAG: int(self.managed),
                                'metadata_of_pool': '',
                                SVID_TAG: snap_svid
                }
            else:
                vdi_info = { UUID_TAG: snap_uuid,
                                NAME_LABEL_TAG: snap_vdi.label,
                                NAME_DESCRIPTION_TAG: util.to_plain_string(\
                                  self.session.xenapi.VDI.get_name_description\
                                  (self.vdi_ref)),
                                IS_A_SNAPSHOT_TAG: 0,
                                SNAPSHOT_OF_TAG: '',
                                SNAPSHOT_TIME_TAG: '',
                                TYPE_TAG: \
                                 self.session.xenapi.VDI.get_type(self.vdi_ref),
                                VDI_TYPE_TAG: 'aio',
                                READ_ONLY_TAG: 0,
                                MANAGED_TAG: int(self.managed),
                                'metadata_of_pool': '',
                                SVID_TAG: snap_svid
                }
            util.SMlog("calling addVDI with info: %s" % vdi_info)
            SLMetadataHandler(self.metadataVolumePath).addVdi(vdi_info)

        return snap_vdi.get_params()

    ####TODO: Only populate this data when required, deprecate this function
    def _load_vdi_info(self, usage=False):
        """ _load_vdi_info method """
        _trace("entry CSLGVDI._load_vdi_info()")

        # get information about the vdi from xapi
        self.sm_config = self.session.xenapi.VDI.get_sm_config(self.vdi_ref)
        if self.sm_config.has_key('LUNIQN'):
            self.vdi_iqn = self.sm_config['LUNIQN']
        self.svid = self.sm_config['SVID']
        #self._check_path()
        if usage:
            (self.size, self.utilisation, self.label, lun_iqn, snapshot) = \
                _get_usage(self.sr, self.svid)

    def _assign(self, vdi_uuid, optflags):
        """ _assign method """
        _trace("entry CSLGVDI._assign()")

        # TODO: Why do we require local IP address? IQN should be sufficient for iSCSI...
        self.ipname = _get_local_hostname()

        (xml, self.lun, self.lunID) = \
            util.retry(lambda: self.sr.cslghandler.assignStorageNode( self.sr.ssid, \
                self.svid, self.ipname, 'flags=xml,bridge', vdi_uuid,
                self.sr.protocol, optflags, self.sr.chapusername,
                self.sr.chapusername), maxretry = 5 )
 
    def _unassign(self, optflags):
        """ _unassign method """
        _trace("entry CSLGVDI._unassign()")

        self.ipname = _get_local_hostname()
        util.retry( lambda: self.sr.cslghandler.unassignStorageNode(self.sr.ssid, \
                                            self.svid, self.ipname, optflags, self.sr.protocol), maxretry = 5 )

    def _allocate(self, vdi_uuid, size, name_label):
        """ _allocate method """
        _trace("entry CSLGVDI._allocate()")

        diskname  = self._gen_diskname(vdi_uuid, name_label)
        spid = self.sr.pool

        # roundup if necessary
        # note: until the service supports array specific allocation
        # sizes via a query, we pick fairly large minimum allocation size

        size = util.roundup(MIN_ALLOCATION, size)

        raid_type = ''
        #if sr_config.has_key('raid-type'):
        #    raid_type = sr_config['raid-type']

        if self.sr.srcmd.params['device_config'].has_key('raid-type'):
            raid_type = self.sr.srcmd.params['device_config']['raid-type']

        create_args = [diskname, size, self.sr.ssid, spid, self.sr.uuid, False, \
                       raid_type, '', self.sr.provision_type, self.sr.provision_options]
                       
        _trace( "create_args = %s" % create_args )

        (self.svid, self.vdi_iqn) = self.sr.cslghandler.createStorageNode(diskname,
                                                                          size,
                                                                          self.sr.ssid,
                                                                          spid,
                                                                          self.sr.uuid,
                                                                          False, 
                                                                          raid_type,
                                                                          '',
                                                                          self.sr.provision_type,                                                                          
                                                                          self.sr.provision_options)

    def _deallocate(self, svid):
        """ _deallocate method """
        _trace("entry CSLGVDI._deallocate()")
        self.sr.cslghandler.deleteStorageNode(svid)
        
    def _remove_from_host(self, SCSIid):
        """ _remove_from_host method """
        _trace("entry CSLGVDI._remove_from_host()")

        if not len(SCSIid):
            util.SMlog("_remove_from_host() Warning - no SCSI identifier, using SVID only")
            sgdevs = scsiutil.cacheSCSIidentifiers()
            devs = scsiutil._genReverseSCSIidmap(self.svid, pathname="csldev")

            for dev in devs:
                if sgdevs.has_key(dev):
                    scsiutil.scsi_dev_ctrl(sgdevs[dev], 'remove')
            return

        sgdevs = scsiutil.cacheSCSIidentifiers()
        devs = scsiutil._genReverseSCSIidmap(SCSIid)
        for dev in devs:
            if sgdevs.has_key(dev):
                scsiutil.scsi_dev_ctrl(sgdevs[dev], 'remove')

    def _load_transport(self, transport):
        """ load ISCSI or FC transport"""
        _trace("entry CSLG._load_transport(transport=%s)"%transport)
        if transport == "ISCSI":
            self.host_devpath = []
            self.iscsi = _load_iscsi_transport(self.sr, self.sr.ssid, \
                                                              self.sr.original_srcmd,\
                                                              self.vdi_iqn, \
                                                              self.uuid, self.sr.mpath)
            for instance in self.iscsi:
                try:
                    # If another iscsi session is attached,
                    # refcount this one, and just refresh it
                    instance.attach(self.uuid)
                    
                    for val in instance.adapter:
                        entry = "%s" % instance.adapter[val]
                        if entry not in self.host_devpath:
                            self.host_devpath.append(entry)
                except:
                    pass
            for host in self.host_devpath:
                scsi_host_rescan.rescan(host)
                
        elif transport == "FC":
            (self.host_adt, self.host_devpath) = _load_fc_transport(self.uuid, \
                                            self.sr.original_srcmd)
            scsiutil.remove_stale_luns(self.host_devpath, self.lunID, \
                        getExpectedCslDevPath(self.svid), self.sr.mpath)


    def _gen_diskname(self, vdi_uuid, label):
        """ generate sanitised diskname from VDI name_label """
        _trace("entry CSLG._gen_diskname()")

        vdi_uuid = vdi_uuid.replace("-","")
        # disk name is max 16 bytes of vmname, padded by a 
        # sanitized vdi guid to 24 bytes total length

        striplist = [" ", "!", "@", "#", "$", "%", "^", "&", "*", 
                     "(", ")", "[", "]", "{", "}", ":", ";", '"',
                     "'", "<", ">", "?", "/"]
        for stripchar in striplist:
            label = label.replace(stripchar,"")

        label = label[0:16]
        vdi_uuid = vdi_uuid.replace("-","")
        diskname = util.to_plain_string(label) + vdi_uuid[0:24-len(label)]
        return diskname

###General Utility Funcs
def createXMLNode(dom, parent_tag, tag, value):
    param_tag = dom.createElement(tag)
    value_tag = dom.createTextNode(str(value))
    param_tag.appendChild(value_tag)
    parent_tag.appendChild(param_tag)
    return

def _rescan_for_LUN(svid, srobj = None, host_devpath = None):
    """ wait for LUN to appear """
    _trace("entry _rescan_for_LUN()")

    devs = None
    if host_devpath != None:
        # Trigger a bus rescan
        _trace("  host_devpath: %s" % (host_devpath))
        for instance in host_devpath:
            _trace("  rescan for %s" % instance )
            scsi_host_rescan.rescan(instance)

    if srobj.mpath == "true":
        _trace( "  _rescan_for_LUN: mpath true, _find_LUN for %s" % svid )
        devs = _find_LUN(svid)
        _trace( "  _rescan_for_LUN: _querySCSIid for %s" % devs )
        SCSIid = _querySCSIid(devs)
        if not SCSIid:
            _trace( "  _rescan_for_LUN: _querySCSIid returned nul" )
            return []
        _trace( "  _rescan_for_LUN: SCSIid = %s" % SCSIid )

        srobj.mpathmodule.refresh(SCSIid, len(devs))

    if devs == None:
        _trace( "  _rescan_for_LUN: _find_LUN for %s" % svid )
        devs = _find_LUN(svid)

    _trace( "  _rescan_for_LUN: return devs: %s" % devs )
    return devs

def getExpectedCslDevPath(svid):
    basepath = "/dev/disk/by-csldev/"
    if svid.startswith("NETAPP_"):
        # special attention for NETAPP SVIDs
        svid_parts = svid.split("__")
        globstr = basepath + "NETAPP__LUN__" + "*" + svid_parts[2] + "*" + svid_parts[-1] + "*"
    else:
        globstr = basepath + svid + "*"

    return globstr
    
def _find_LUN(svid):
    _trace("entry CSLG._find_LUN()")
    _trace("  find svid: %s ..." % svid )
    globstr = getExpectedCslDevPath(svid)
    _trace( "  wait_for_path_multi on %s" % globstr )
    path = util.wait_for_path_multi(globstr, MAX_TIMEOUT)
    if not len(path):
        _trace( "  globstr not found: %s" % globstr )
        _trace("exit CSLG._find_LUN()")
        return []

    svid_to_use = re.sub("-[0-9]*:[0-9]*:[0-9]*:[0-9]*$","",os.path.basename(path))
    util.SMlog("  -- SUCCESS FINDING LUN --" )
    util.SMlog("  Generating SVID: %s" % svid_to_use)
    devs = scsiutil._genReverseSCSIidmap(svid_to_use, pathname="csldev")
    util.SMlog("  returning devs: %s" % devs)
    _trace("exit CSLG._find_LUN()")
    return devs

def _check_path_and_raise(svid, srobj, devs=[]):
    (path ,SCSIid, npaths) = _check_path(svid, srobj, devs)
    if path == None:
        _trace( "_check_path_and_raise: _check_path returned None, throw VDIUnavailable" )
        raise xs_errors.XenError('VDIUnavailable', \
                                  opterr='unable to query SCSIid')

    _trace( "_check_path_and_raise: return path, SCSIid, npaths" )
    return (path ,SCSIid, npaths)

def _check_path_and_reattach_if_req(svid, srobj, devs=[]):
    (path ,SCSIid, npaths) = _check_path(svid, srobj, devs)
    if path == None:
        #path might have lost either in host, array or due to master/slave Failover, so re-assign it
        _trace("Metadata path lost. so, re-attaching to the host")
        if srobj.isMaster:
            srobj.attachMetadataVolume()
            (path ,SCSIid, npaths) = _check_path(svid, srobj, devs)

    return (path ,SCSIid, npaths)

def _check_path(svid, srobj, devs=[]):
    """ check and activate path """
    _trace("entry CSLG._check_path()")

    if not len(devs):
        _trace( "  _check_path: _find_LUN for svid: %s" % svid )
        devs = _find_LUN(svid)

    SCSIid = _querySCSIid(devs)
    if not SCSIid:
        _trace( "  _check_path: _querySCSIid returned nul" )
        return (None, '', 0)
    else:
        _trace( "  _check_path: SCSIid = %s" % SCSIid )

    _trace("exit CSLG._check_path()")
    return [srobj.mpathmodule.path(SCSIid),SCSIid, len(devs)]

def _querySCSIid(devs):
    for dev in devs:
        try:
            SCSIid = scsiutil.getSCSIid(dev)
            return SCSIid
        except:
            pass
    return None

def _check_host_fc(self):
    self.host_adt = {}
    try:
        driver = SR.driver('hba')
        self.module = driver(self.original_srcmd, self.uuid)
        try:
            self.host_adt = self.module._init_hbas()
            _trace("DEBUG: _check_host_fc: host_adt %s"%self.host_adt)
        except:
            _trace("DEBUG: _check_host_fc: exception on _init_hbas() call")
            pass
        try:
            self.host_wwn = self.module._init_hba_hostname()
            _trace("DEBUG: _check_host_fc: host_wwn %s"%self.host_wwn)
        except:
            _trace("DEBUG: _check_host_fc: exception on _init_hostname() call")
            pass
    except:
        pass

    if len(self.host_adt.iterkeys()):
        return True
    else:
        return False

def _find_iscsi_targets(cslghandler, ssid, LUNperIQN=False):
    """ _find_iscsi_targets method """
    _trace("entry CSLGVDI._find_iscsi_targets()")

    return cslghandler.enumStorageSystemTargetPorts(ssid, LUNperIQN)

def _build_iqn_optflag(self):
    """ _build_iqn_optflag method """
    _trace("entry CSLGSR._build_iqn_optflag()")
 
    sr = self
    if type(self).__name__ == 'CSLGVDI':
        sr = self.sr

    if not sr.dconf.has_key('localIQN'):
        try:
            sr.dconf['localIQN'] = sr.session.xenapi.host.get_other_config(sr.host_ref)['iscsi_iqn']
        except:
            raise xs_errors.XenError('ConfigISCSIIQNMissing')

    optflags = "iqn=%s " % sr.dconf['localIQN']
    if sr.chapusername and sr.chappassword:
        optflags = optflags + "chap-user=%s chap-pass=%s" % \
                  (sr.chapusername, sr.chappassword)
    return optflags

def _build_wwn_optflag(self):
    """ _build_wwn_optflag method """
    _trace("entry CSLGSR._build_wwn_optflag()")

    sr = self
    if type(self).__name__ == 'CSLGVDI':
        sr = self.sr

    driver = SR.driver('hba')
    sr.module = driver(sr.original_srcmd, sr.uuid)
    sr.host_adt = sr.module._init_hbas()

    optflags = 'wwn='
    for host in sr.host_adt:
        optflags = optflags + sr.host_adt[host] + ","
    optflags = optflags[:-1]  # truncate last comma
    _trace( "  wwn optflags = %s" % optflags )

    if sr.adapter == "EMC_CLARIION" or sr.adapter == "SMIS_STORAGE_SYSTEM":
        host_wwn = sr.module._init_hba_hostname()
        _trace( "  got host_wwn: %s" % host_wwn )
        if len( host_wwn ) > 0:
            optflags += ' hostwwn=' + host_wwn
    _trace( "  return optflags = %s" % optflags )
    return optflags

def _getOptFlags(self, transport):
    if transport == "ISCSI":
        return _build_iqn_optflag(self)
    elif transport == "FC":
        return _build_wwn_optflag(self)
    else:
        raise xs_errors.XenError('CSLGInvalidProtocol', \
            opterr='protocol = %s' % transport)

def _load_fc_transport(uuid, srcmd):
    _trace("entry _load_fc_transport()")
    driver = SR.driver('hba')
    module = driver(srcmd, uuid)
    module.attach(uuid)
    host_adt = module._init_hbas()
    host_devpath = []
    for node in host_adt:
        host_devpath.append(node.split('host')[1])
    return [host_adt, host_devpath]

def _load_iscsi_transport(sr, ssid,srcmd,vdi_iqn,uuid,mpath):
    _trace("entry _load_iscsi_transport()")
    if sr.ports != None and sr.targets != None:
        ports = sr.ports
        targets = sr.targets
    else:
        (ports, targets) = _find_iscsi_targets(sr.cslghandler, ssid, vdi_iqn)

    if vdi_iqn != None and len(vdi_iqn) > 0:
        targets.append(vdi_iqn)

    util.SMlog("  _load_scsi_transport: ports: %s, targets: %s" % (ports, targets))
    if not len(ports) or not len(targets):
        raise xs_errors.XenError('CSLGISCSIInit', \
                                     opterr='Unable to identify target port/IQN')

    driver = SR.driver('iscsi')
    iscsi = []

    for target in targets:
        for port in ports:
            original_srcmd = copy.deepcopy(srcmd)

            # Parse the port into name/ip and ip port (if there is one)
            port_parts = port.split(':')
            original_srcmd.params['device_config']['target'] = port_parts[0]
            if len(port_parts) > 1:
                original_srcmd.params['device_config']['port'] = port_parts[1]

            original_srcmd.params['device_config']['targetIQN'] = target
            instance = driver(original_srcmd, uuid)
            iscsi.append(instance)

    return iscsi

def _list_adapter_names():
    """ find the list of adapter libs in the directory """
    _trace("in _list_adapter_names()")

    adapList = []
    try:
        dirPath = os.path.join('/opt/xensource/sm/', 'iSL')
        _trace("DEBUG: dirpath %s" % dirPath)
        for file in os.listdir(dirPath):
            _trace("file %s" % file)
            if file.startswith("lib") == True:
                adapName = file[3:file.find(".")]
                adapList.append(adapName.upper())

    except Exception, e:
        _trace("FAILED: system error with installed storage adapters : %s" % e)
        pass

    return adapList

def _get_usage(srobj, svid):
    """ determine protocol method """
    _trace("entry CSLGVDI._get_usage()")
    var = srobj.cslghandler.getStorageVolumeInfo(srobj.ssid, svid)
    util.SMlog(var)
    return var

def _validate_svid(srobj, svid):
    """ _validateSVID function """
    _trace("entry CSLGVDI._validate_svid()")

    return srobj.cslghandler.getStorageVolumeInfo(srobj.ssid, svid)[3]

def _get_local_hostname():
    """ _get_local_hostname function """
    _trace("entry CSLGVDI._get_local_hostname()")

    # try getting hostname ... if not in DNS, exception occurs
    # and return IP address instead
    try:
        ipname = socket.gethostname()
        socket.gethostbyname(ipname)
        return ipname
    except:
        session = util.get_localAPI_session()
        Hostref = util.get_localhost_uuid(session)
        rec = session.xenapi.host.get_record(Hostref)
        return rec['address']

def _trace(string, errorcase = 0, priority=syslog.LOG_INFO):
    """ simple function to control trace on/off """
    if TRACEON or errorcase:
        util.SMlog(string, priority=priority)

def _insertUdevRules():
    rulefile = "/etc/udev/rules.d/96-CSLG.rules"
    scriptfile = "/etc/udev/scripts/CSLGfilt"
    if not os.path.exists(scriptfile):
        fd = open (scriptfile, "w")
        fd.write("#!/bin/bash\n")
        fd.write("/opt/Citrix/StorageLink/bin/csl_devscan --path=${1} | awk '{print $2}'\n")
        fd.close()
        os.chmod(scriptfile, 0755)

    if not os.path.exists(rulefile):
        fd = open (rulefile, "w")
        fd.write("KERNEL==\"sd*[!0-9]\", BUS==\"scsi\",PROGRAM=\"/etc/udev/scripts/CSLGfilt $tempnode\", SYMLINK+=\"disk/by-csldev/%c-%b\"\n")

        fd.close()

def _removeUdevRules():
    rulefile = "/etc/udev/rules.d/96-CSLG.rules"
    scriptfile = "/etc/udev/scripts/CSLGfilt"
    if os.path.exists(rulefile):
        os.unlink(rulefile)
    if os.path.exists(scriptfile):
        os.unlink(scriptfile)

def _toString(li):
    if not len(li):
        return ""
    s = ""
    for idx in range(0,len(li)-1):
        s += (li[idx] + ",")
    s += li[len(li)-1]
    return s

def _fromString(li):
    return li.split(',')

def _createNetappPool(sr, name_label, vdi_uuid, size):
    _trace("->_createNetappPool")
    poolname = "Xen_" + name_label.strip() + vdi_uuid.split("-")[0]
    ssid = sr.srcmd.params['device_config']['storageSystemId']
    sizeMB = size/(1024*1024)
    if int(sizeMB) < 9: #Netapp doesn't allow to create pool size less than 9Mb
        sizeMB = 9 
    _trace("Creating pool with args : poolname:%s, ssid:%s, spid:%s, size:%s" % (poolname, ssid, sr.pool, sizeMB))
    subpool = sr.cslghandler.createStoragePool( poolname, ssid, sr.pool, sizeMB, sr.provision_type, sr.provision_options) 
    _trace("pool created, spid: %s" %subpool)
    return subpool

def _deleteNetappPool(sr, subpool):
    _trace("->_deleteNetappPool")
    ssid = sr.srcmd.params['device_config']['storageSystemId']
    _trace("Delete netapp pool id: %s"%subpool)
    sr.cslghandler.deleteStoragePool( ssid, subpool)
    return

def _stripSerialIds(svids):
    serialids=[]
    for svid in svids:
        serialids.append(svid[svid.rfind('__')+2:])
    return serialids
        
if __name__ == '__main__':
    SRCommand.run(CSLGSR, DRIVER_INFO)
else:
    SR.registerSR(CSLGSR)
