#!/usr/bin/env python
# Copyright (c) 2005-2007 XenSource, Inc. All use and distribution of this
# copyrighted material is governed by and subject to terms and conditions
# as licensed by XenSource, Inc. All other rights reserved.
# Xen, XenSource and XenEnterprise are either registered trademarks or
# trademarks of XenSource Inc. in the United States and/or other countries.
#
#
# EQUALSR: DELL-EQUALLOGIC storage repository
#

import SR, VDI, ISCSISR, SRCommand, util, scsiutil
import os, sys, re, random
import datetime, time
import xml.dom.minidom
import xs_errors
import sshutil
import copy

EXCL_MAP = False
CAPABILITIES = ["SR_PROBE","VDI_CREATE","VDI_DELETE","VDI_ATTACH","VDI_DETACH",
                "VDI_CLONE","VDI_SNAPSHOT","VDI_RESIZE"]

CONFIGURATION = [ [ 'target', 'IP address or hostname of the EQUALLOGIC group' ], \
                  [ 'control', 'IP address or hostname of the control interface for the EQUALLOGIC (optional)' ], \
                  [ 'username', 'Login username to be used by the driver for manipulating volumes on the filer' ], \
                  [ 'password', 'Login password to be used by the driver for manipulating volumes on the filer' ], \
                  [ 'chapuser', 'The username to be used during CHAP authentication (optional)' ], \
                  [ 'chappassword', 'The password to be used during CHAP authentication (optional)' ], \
                  [ 'allocation', 'Valid values are thick or thin (optional, defaults to thick)'], \
                  [ 'storagepool', 'The disk pool on which the Volumes should be stored (optional)'] ]


DRIVER_INFO = {
    'name': 'EqualLogic',
    'description': 'SR driver for mapping of LUNs to VDIs on a EQUALLOGIC array group, providing use of fast snapshot and clone features on the array',
    'vendor': 'Citrix Systems Inc',
    'copyright': '(C) 2008 Citrix Systems Inc',
    'driver_version': '0.1',
    'required_api_version': '0.1',
    'capabilities': CAPABILITIES,
    'configuration': CONFIGURATION
    }

SR_PREFIX = 'XenStorage'
SR_SUFFIX = 'MANAGEMENT'
EQL_SIZE_INCREMENT = 15 * 1024 * 1024
NULL = 0

class EQUALSR(ISCSISR.ISCSISR):
    """Equallogic array storage repository"""
    def handles(type):
        if type == "equal":
            return True
        return False
    handles = staticmethod(handles)

    def load(self, sr_uuid):
        # Required keys:
        #   target, user, password
        self.sr_vditype = 'aio'
        if self.cmd == "sr_content_type":
            return
        
        if not self.dconf.has_key('target') or not self.dconf['target']:
            raise xs_errors.XenError('EQLTarget')
        self.target = self.control = self.dconf['target']
        if self.dconf.has_key('control') and self.dconf['control']:
            self.control = self.dconf['control']
        if not self.dconf.has_key('username') or not self.dconf['username']:
            raise xs_errors.XenError('EQLUsername')
	# Allow empty password string
        if not (self.dconf.has_key('password') or self.dconf.has_key('password_secret')):
            raise xs_errors.XenError('EQLPassword')

        # Optional dconf params
        self.isMaster = False
        if self.dconf.has_key('SRmaster') and self.dconf['SRmaster'] == 'true':
            self.isMaster = True

        self.pool = False
        if self.dconf.has_key('storagepool'):
            self.poolname = self.dconf['storagepool']
            self.pool = True
        self.username = self.dconf['username']
        if self.dconf.has_key('password_secret'):
            self.password = util.get_secret(self.session, self.dconf['password_secret'])
        else:
            self.password = self.dconf['password']
        self.conn_cred = [self.control, self.username, self.password]
        self.mgmtsrname = SR_PREFIX + sr_uuid.replace("-","") + SR_SUFFIX
        self.luns = []
        self.conn = sshutil.SSHSession(self.conn_cred)
        self.conn.connect()
        (rc, grpparams_info, errlog, errmsg) = sshutil.grpparams_detail(self.conn)
        if (rc):
            if (len(errlog)):
                 for id in range(len(errlog)):
                     util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"creating management LUN")
        else:
            group_IP = grpparams_info['Group-Ipaddress']
            if (self.target != group_IP):
                eql_raise_error(self,'','',code='EQLInvalidTargetIP')

        # Test for thick vs thin provisioning conf parameter
        if self.dconf.has_key('allocation') and self.dconf['allocation'] == 'thin':
            self.allocation = False
        else:
            self.allocation = True

        if self.cmd in ['sr_create','sr_attach']:
            if not sshutil.check_version(self.conn):
                eql_exit(self)
                raise xs_errors.XenError('EQLUnsupportedVersion')
            
        if util.match_uuid(sr_uuid):
            self.exists = sshutil.vol_exists(self.conn, self.mgmtsrname)
        else:
            self.exists = 0
            
    def create(self, sr_uuid, size):
        if self.exists:
            eql_raise_error(self,'','',code='SRExists')
        
        if not self.pool:
            print >>sys.stderr,_storagepooltoXML(self._getAllStoragePools())
            eql_raise_error(self,'','',code='EQLStoragePool')

	# Ignore thin provisioning for the Management LUN
        allocate = 'thick'
        size_str = "20m"
	state = 'offline'
        # Create a management LUN
        name = self.mgmtsrname
        uuid = sr_uuid

        (rc, outmsg, iSCSIname, errlog, errmsg) = sshutil.vol_create(self.conn, str(name), \
                              str(size_str), str(allocate), str(sr_uuid), str(state), 0, \
                              "volume-offline", str(self.poolname))
        if (rc):
            if (len(errlog)):
                 for id in range(len(errlog)):
                     util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"creating management LUN")
        eql_exit(self)


    def delete(self, sr_uuid):
        if not self.exists:
            return
        VDI_PREFIX = SR_PREFIX + sr_uuid.replace("-","")
        vdi_pattern = VDI_PREFIX + "[0-9a-f]{20}$"

        (rc, vol_list, errlog, errmsg) = sshutil.list_SR_vols(self.conn, vdi_pattern)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"listing volumes in SR")
        for index in range(len(vol_list)):
            sshutil.vol_offline(self.conn, vol_list[index])
            (rc, output, errlog, errmsg) = sshutil.vol_delete(self.conn, vol_list[index])
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                eql_raise_error(self, errmsg,"deleting vdi")
        (rc, errlog, errmsg) = sshutil.vol_offline(self.conn, self.mgmtsrname)
        (rc, output, errlog, errmsg) = sshutil.vol_delete(self.conn, self.mgmtsrname)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"deleting SR management-volume")
        eql_exit(self)

    def attach(self, sr_uuid):
        if not self.exists:
            eql_raise_error(self,'','',code='SRUnavailable')

        util.SMlog("Attaching SR [%s]" % sr_uuid)
        name = self.mgmtsrname

        (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.conn, str(name))
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            errorstr = "listing volume details VDI (name %s)" % name
            eql_raise_error(self, errmsg, errorstr)
        real_poolname = vol_info['Pool']

        if self.pool:
            if (self.poolname != real_poolname):
                eql_raise_error(self,'','',code='EQLInvalidStoragePool')
        else:
            pbd = None
            try:
                pbd = util.find_my_pbd(self.session, self.host_ref, self.sr_ref)
            except:
                pass

            if pbd != None:
                dconf = self.session.xenapi.PBD.get_device_config(pbd)
                dconf['storagepool'] = real_poolname
                self.session.xenapi.PBD.set_device_config(pbd, dconf)
            self.poolname = real_poolname
            self.pool = True

        conf = self.session.xenapi.SR.get_sm_config(self.sr_ref)
        key1 = 'StoragePool'
        if not conf.has_key(key1):
            self.session.xenapi.SR.add_to_sm_config(self.sr_ref, key1, self.poolname)
        else:
            conf[key1] = real_poolname
            self.session.xenapi.SR.set_sm_config(self.sr_ref, conf)

        key2 = 'Allocation'
        if not conf.has_key(key2):
            if self.allocation:
                self.session.xenapi.SR.add_to_sm_config(self.sr_ref, key2, 'Thick')
            else:
                self.session.xenapi.SR.add_to_sm_config(self.sr_ref, key2, 'Thin')
        else:
            if self.dconf.has_key('allocation'):
                if self.allocation:
                    conf[key2] = 'Thick'
                else:
                    conf[key2] = 'Thin'
                self.session.xenapi.SR.set_sm_config(self.sr_ref, conf)

	self.localIQN = self.session.xenapi.host.get_other_config(self.host_ref)['iscsi_iqn']
        (rc, errlog, errmsg) = sshutil.vol_online(self.conn, name)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"Bringing SR management-volume %s online:" % name)
        (rc, output, errlog, errmsg) = sshutil.vol_create_access_rec_by_initiator(self.conn, name, self.localIQN)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"Creating access record for SR management-volume %s" % name)
        eql_exit(self)
        self._setMultipathableFlag()
        self._addLUNperVDIkey()

    def probe(self):
        XML = util.SRtoXML(self._getAllSRs())
        eql_exit(self)
        return XML

    def detach(self, sr_uuid):
        if not self.exists:
            return
        activePBDs = util.test_activePoolPBDs(self.session, self.host_ref, self.sr_ref)
        name = self.mgmtsrname
        VDI_PREFIX = SR_PREFIX + sr_uuid.replace("-","")
        vdi_pattern = VDI_PREFIX + "[0-9a-f]{20}$"
        self.localIQN = self.session.xenapi.host.get_other_config(self.host_ref)['iscsi_iqn']

        (rc, vol_list, errlog, errmsg) = sshutil.list_SR_vols(self.conn, vdi_pattern)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"listing volumes in SR")
        for index in range(len(vol_list)):
            (rc, vdi_info, errlog, errmsg) = sshutil.vol_detail(self.conn, vol_list[index])
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                eql_raise_error(self, errmsg,"getting vdi detail info")
	    else:
                desc_str = vdi_info['Description']
                if not (match_vol_desc(desc_str)):
                    util.SMlog("EQL: Volume [%s] metadata corrupted" % vol_list[index])
                    eql_exit(self)
                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                try:
                    vdi_uuid = desc_str.split(':')[0].replace("UUID-","")
                except:
                    util.SMlog("EQL: Volume [%s] metadata corrupted" % vol_list[index])
                    eql_exit(self)
                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
                (rc, output, errlog, errmsg) = sshutil.vol_delete_access_rec_by_initiator(self.conn, \
                                                 vdi_name, self.localIQN)
                if not activePBDs:
                    (rc, errlog, errmsg) = sshutil.vol_offline(self.conn, vol_list[index])

        (rc, output, errlog, errmsg) = sshutil.vol_delete_access_rec_by_initiator(self.conn, name, \
                                                 self.localIQN)

        if not activePBDs:
            util.SMlog("No other active PBDs, setting  vol to offline [%s]" % \
                           name)
            (rc, errlog, errmsg) = sshutil.vol_offline(self.conn, name)
        eql_exit(self)

    def scan(self, sr_uuid):
        if not self.exists:
            eql_raise_error(self,'','',code='SRUnavailable')
        self._loadvdis()
        self._getPoolstats()
        self.passthrough = True
        XML = super(EQUALSR, self).scan(sr_uuid)
        eql_exit(self)
        return XML

    def content_type(self, sr_uuid):
        return super(EQUALSR, self).content_type(sr_uuid)
    
    def vdi(self, uuid):
        return EQUALVDI(self, uuid)

    def added_vdi(self, vdi):
        self.vdis[vdi.uuid] = vdi
        #self._addtoserial(vdi.serial)

    def deleted_vdi(self, uuid):
        if uuid in self.vdis:
            #self._delfromserial(self.vdis[uuid].serial)
            del self.vdis[uuid]

    def _loadvdis(self):
        if self.vdis:
            return

        sr_uuid = self.uuid
        VDI_PREFIX = SR_PREFIX + sr_uuid.replace("-","")
        vdi_pattern = VDI_PREFIX + "[0-9a-f]{20}$"
        (rc, vdi_list, errlog, errmsg) = sshutil.list_SR_vols(self.conn, vdi_pattern)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"listing all VDIs in the SR")

        for index in range(len(vdi_list)):
            (rc, vdi_info, errlog, errmsg) = sshutil.vol_detail(self.conn, vdi_list[index])
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                eql_raise_error(self, errmsg,"getting vdi detail info")
	    else:
                desc_str = vdi_info['Description']
                if not (match_vol_desc(desc_str)):
                    util.SMlog("EQL: Volume [%s] metadata corrupted" % vdi_list[index])
                    eql_exit(self)
                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                try:
                    vdi_uuid = desc_str.split(':')[0].replace("UUID-","")
                    vdi_deleted = desc_str.split(':')[1].replace("DELETED-","")
                except:
                    util.SMlog("EQL: Volume [%s] metadata corrupted" % vdi_list[index])
                    eql_exit(self)
                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                if (vdi_deleted == "False"):
                    vol_utilisation = int(str_to_bytes(vdi_info['VolReserveInUse']))
                    snap_reserve_percent = int(vdi_info['Snap-Reserve'].split('%')[0])
                    vol_size = int(str_to_bytes(vdi_info['Size']))

                    snap_reserve = vol_size * (snap_reserve_percent / 100)
                    self.virtual_allocation += vol_size
                    self.physical_utilisation += vol_utilisation
                    self.added_vdi(self.vdi(vdi_uuid))
                (rc, snap_list, errlog, errmsg) = sshutil.list_VDI_snaps(self.conn, vdi_list[index], vdi_pattern)
                if (rc):
                    if (len(errlog)):
                        for id in range(len(errlog)):
                            util.SMlog(errlog[id])
                    eql_raise_error(self, errmsg,"listing all snapshots in VDI")
                else:
                    if (len(snap_list) != 0):
                        for id in range(len(snap_list)):
                            (rc, snap_info, errlog, errmsg) = sshutil.snap_detail(self.conn, vdi_list[index], snap_list[id])
                            if (rc):
                                if (len(errlog)):
                                    for id in range(len(errlog)):
                                        util.SMlog(errlog[id])
                                eql_raise_error(self, errmsg,"listing snapshots details")
                            else:
                                desc_str = snap_info['Description']
                                if not (match_snap_vol_desc(desc_str)):
                                    util.SMlog("EQL: Volume Snapshot [%s] metadata corrupted" % snap_list[id])
                                    eql_exit(self)
                                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                                try:
                                    snap_uuid = desc_str.split(':')[0].replace("SNAP-","")
                                except:
                                    util.SMlog("EQL: Volume Snapshot [%s] metadata corrupted" % snap_list[id])
                                    eql_exit(self)
                                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                                self.added_vdi(self.vdi(snap_uuid))

    ###EQUAL SR Utility Functions###
        
    def _getPoolstats(self):
        sr_uuid = self.uuid
        sr_name = self.mgmtsrname
        (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.conn, sr_name)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            errorstr = "listing volume details of SR management volume (name %s)" % sr_name
            eql_raise_error(self, errmsg, errorstr)

        self.sr_poolname = vol_info['Pool']
        (rc, pool_info, errlog, errmsg) = sshutil.pool_detail(self.conn, self.sr_poolname)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            errorstr = "listing pool details pool (name %s)" % self.sr_poolname
            eql_raise_error(self, errmsg, errorstr)
        total_cap_str = pool_info['TotalCapacity']       
        free_space_str = pool_info['FreeSpace']

        total_cap = str_to_bytes(total_cap_str)
        free_space = str_to_bytes(free_space_str)
        self.physical_size = total_cap
        self.physical_utilisation = (total_cap - free_space)

    def _getAllStoragePools(self):
        (rc, sp_list, errlog, errmsg) = sshutil.all_StoragePools(self.conn)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"listing all Storage pools in the group")

        return sp_list

    def _getAllSRs(self):
        sr_pattern = SR_PREFIX + "[0-9a-f]{32}" + SR_SUFFIX
        (rc, sr_list, errlog, errmsg) = sshutil.list_SR_vols(self.conn, sr_pattern)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self, errmsg,"listing all SRs in the pool")

        SRlist = {}
        for id in range(len(sr_list)):
            (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.conn, str(sr_list[id]))
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "listing volume details VDI (name %s)" % sr_list[id]
                eql_raise_error(self, errmsg, errorstr)
            uuid = vol_info['Description']
            if not util.match_uuid(uuid):
                util.SMlog("EQL: SR Management Volume [%s] metadata corrupted" % sr_list[id])
                eql_exit(self)
                raise xs_errors.XenError('EQLInvalidVolMetaData')
            vol_size_str = vol_info['VolReserve']
            vol_size = str_to_bytes(vol_size_str)
            vol_pool = vol_info['Pool']
            entry = {}
            entry['size'] = vol_size
            entry['storagepool'] = vol_pool
            SRlist[uuid] = entry
        return SRlist



class EQUALVDI(VDI.VDI):
    def load(self, vdi_uuid):
        if not self.sr.exists:
            eql_raise_error(self.sr,'','',code='SRUnavailable')
        self.exists = False
        self.is_snapshot = False
        sr_uuid = self.sr.uuid
        VDI_PREFIX = SR_PREFIX + sr_uuid.replace("-","")
        vdi_pattern = VDI_PREFIX + "[0-9a-f]{20}$"

        vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        (rc, self.vol_info, errlog, errmsg) = sshutil.vol_detail(self.sr.conn, vdi_name)
        if not rc:
            self.exists = True;
        
        if not self.exists:
            if self.sr.cmd == 'sr_scan':
                try:
                    self.sr.srcmd.params['vdi_ref'] = self.sr.session.xenapi.VDI.get_by_uuid(vdi_uuid)
                except:
                    return
            if self.sr.srcmd.params.has_key('vdi_ref'):
                vdi_ref = self.sr.srcmd.params['vdi_ref']
                self.sm_config = self.session.xenapi.VDI.get_sm_config(self.sr.srcmd.params['vdi_ref'])
                if self.sm_config.has_key('snapshot-of'): 
                    self.is_snapshot = True
            else:
                return

        if (self.is_snapshot):
            self.orig_vdi = self.sm_config['snapshot-of']
            vdi_name = "XenStorage" + sr_uuid.replace("-", "") + self.orig_vdi.replace("-", "")[:20]
            snap_name = "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
            self.vol_name = vdi_name
            (rc, self.snap_info, errlog, errmsg) = sshutil.snap_detail(self.sr.conn, str(vdi_name), str(snap_name))
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "listing snap details VDI (name %s)" % snap_name
                eql_raise_error(self.sr, errmsg, errorstr)

            self.exists = True
            self.obj = self.snap_info
            desc_str = self.obj['Description']
            if not (match_snap_vol_desc(desc_str)):
                util.SMlog("EQL: Volume Snapshot [%s] metadata corrupted" % str(snap_name))
                eql_exit(self.sr)
                raise xs_errors.XenError('EQLInvalidVolMetaData')
            try:
                self.parent = desc_str.split(':')[1].replace("SRC-","")
                utilisation_str = desc_str.split(':')[2].replace("UTIL-","")
            except:
                util.SMlog("EQL: Volume Snapshot [%s] metadata corrupted" % str(snap_name))
                eql_exit(self.sr)
                raise xs_errors.XenError('EQLInvalidVolMetaData')
            self.utilisation = int(utilisation_str)
            size_str = self.obj['OriginalVolumeSize']
            self.size = str_to_bytes(size_str)
            if (self.obj['Permission'] == "read-only"):
                self.read_only = True
            else:
                self.read_only = False
            self.sr.added_vdi(self)
        else:
	    vdi_name = self.sr.mgmtsrname.replace(SR_SUFFIX,"") + vdi_uuid.replace("-", "")[:20]
            self.vol_name = vdi_name
            if self.vol_info.has_key('Description'):
                desc_str = self.vol_info['Description']
                if not (match_vol_desc(desc_str)):
                    util.SMlog("EQL: Volume [%s] metadata corrupted" % self.vol_name)
                    eql_exit(self.sr)
                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                try:
                   vol_uuid = desc_str.split(':')[0].replace("UUID-","")
                except:
                    util.SMlog("EQL: Volume [%s] metadata corrupted" % self.vol_name)
                    eql_exit(self.sr)
                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                if (vol_uuid == vdi_uuid):
                    self.obj = self.vol_info
                    self.exists = True
                    self.size = str_to_bytes(self.obj['Size'])
                    self.utilisation = str_to_bytes(self.obj['VolReserveInUse'])
                    if (self.obj['Permission'] == "read-only"):
                        self.read_only = True
                    else:
                        self.read_only = False
                    self.sr.added_vdi(self)
        
        if self.sr.cmd == "vdi_detach":
            eql_exit(self.sr)

    def create(self, sr_uuid, vdi_uuid, size):
        if self.exists:
            eql_raise_error(self.sr,'','',code='VDIExists')
        self.sm_config = self.sr.srcmd.params['vdi_sm_config']
        vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        description = "UUID-%s:DELETED-%s"%(vdi_uuid, "False")
        state = 'offline'
        if self.sm_config.has_key('allocation'):
            if (self.sm_config['allocation'] == "thin"):
                allocate = 'thin'
            else:
                allocate = 'thick'
        else:
            conf = self.session.xenapi.SR.get_sm_config(self.sr.sr_ref)
            if conf.has_key('Allocation'):
                if conf['Allocation'] == 'Thick':
                    allocate = 'thick'
                else:
                    allocate = 'thin'
            else:
                allocate = 'thick'
        snapreserve = 100
        if self.sm_config.has_key('snap-reserve-percentage'):
            try:
                snapreserve = long(self.sm_config['snap-reserve-percentage'])
            except:
                eql_exit(self.sr)
                raise xs_errors.XenError('EQLInvalidSnapReserve')

        snapdepletion = "volume-offline"
        if self.sm_config.has_key('snap-depletion'):
            if not self.sm_config['snap-depletion'] in ['delete-oldest', 'volume-offline']:
                eql_exit(self.sr)
                raise xs_errors.XenError('EQLInvalidSnapDepletionKey')
            snapdepletion = self.sm_config['snap-depletion']

        #Size in Bytes, roundup to nearest MB
        size_mb = util.roundup(EQL_SIZE_INCREMENT, long(size)) / (1024 * 1024)
        size_str = "%dMB"% size_mb
        if self.sr.pool:
            poolname = self.sr.poolname
        else:
            conf = self.session.xenapi.SR.get_sm_config(self.sr.sr_ref)
            if conf.has_key('StoragePool'):
                poolname = conf['StoragePool']
        (rc, outmsg, iSCSIname, errlog, errmsg) = sshutil.vol_create(self.sr.conn, str(vdi_name), \
                                 str(size_str), str(allocate), str(description), \
                                 str(state), snapreserve, str(snapdepletion), str(poolname))
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            errorstr = "creating VDI (size %s)" % size_str
            eql_raise_error(self.sr, errmsg, errorstr)

        (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.sr.conn, str(vdi_name))
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            errorstr = "listing volume details VDI (name %s)" % vdi_name
            eql_raise_error(self.sr, errmsg, errorstr)
        self.size = str_to_bytes(vol_info['Size'])
        self.sr.added_vdi(self)
        self._db_introduce()
        eql_exit(self.sr)
        return super(EQUALVDI, self).get_params()

    def delete(self, sr_uuid, vdi_uuid):
        if not self.exists:
            return
        if (self.is_snapshot):
            vdi_name = "XenStorage" + sr_uuid.replace("-", "") + self.orig_vdi.replace("-", "")[:20]
            snap_name = "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        else:
            vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]

        VDI_PREFIX = SR_PREFIX + sr_uuid.replace("-","")
        vdi_pattern = VDI_PREFIX + "[0-9a-f]{20}$"
        (rc, snap_list, errlog, errmsg) = sshutil.list_VDI_snaps(self.sr.conn, vdi_name, vdi_pattern)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self.sr, errmsg,"listing all snapshots in VDI")

        if (self.is_snapshot):
            if (len(snap_list) > 1):
                (rc, output, errlog, errmsg) = sshutil.snap_delete(self.sr.conn, vdi_name, snap_name)
                if (rc) and errmsg.find('Snapshot %s does not exist' % snap_name) == -1:
                    if (len(errlog)):
                        for id in range(len(errlog)):
                            util.SMlog(errlog[id])
                    errorstr = "Deleting snapshot VDI (name %s)" % snap_name
                    eql_raise_error(self.sr, errmsg, errorstr)
            else:
                (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.sr.conn, str(vdi_name))
                if (rc):
                    if (len(errlog)):
                        for id in range(len(errlog)):
                            util.SMlog(errlog[id])
                    errorstr = "listing volume details VDI (name %s)" % vdi_name
                    eql_raise_error(self.sr, errmsg, errorstr)
                desc_str = vol_info['Description']
                if not (match_vol_desc(desc_str)):
                    util.SMlog("EQL: Volume [%s] metadata corrupted" % str(vdi_name))
                    eql_exit(self.sr)
                    raise xs_errors.XenError('EQLInvalidVolMetaData')
                try:
                    deleted_val = desc_str.split(':')[1].replace("DELETED-","")
                except:
                    util.SMlog("EQL: Volume [%s] metadata corrupted" % str(vdi_name))
                    eql_exit(self.sr)
                    raise xs_errors.XenError('EQLInvalidVolMetaData')

                if (deleted_val == "True"):
                    if (vol_info['Status'] == "online"):
                        (rc, errlog, errmsg) = sshutil.vol_offline(self.sr.conn, vdi_name)
                        if (rc):
                            if (len(errlog)):
                                for id in range(len(errlog)):
                                    util.SMlog(errlog[id])
                            eql_raise_error(self.sr, errmsg,"Bringing down VDI %s offline:" % vdi_name)
                    (rc, output, errlog, errmsg) = sshutil.vol_delete(self.sr.conn, vdi_name)
                    if (rc) and errmsg.find('Volume %s does not exist' % vdi_name) == -1:
                        if (len(errlog)):
                            for id in range(len(errlog)):
                                util.SMlog(errlog[id])
                        errorstr = "Deleting VDI (name %s)" % vdi_name
                        eql_raise_error(self.sr, errmsg, errorstr)
                else:
                    (rc, output, errlog, errmsg) = sshutil.snap_delete(self.sr.conn, vdi_name, snap_name)
                    if (rc) and errmsg.find('Snapshot %s does not exist' % snap_name) == -1:
                        if (len(errlog)):
                            for id in range(len(errlog)):
                                util.SMlog(errlog[id])
                        errorstr = "Deleting snapshot VDI (name %s)" % snap_name
                        eql_raise_error(self.sr, errmsg, errorstr)
        else:
            if (len(snap_list) != 0):
                description = "UUID-%s:DELETED-%s"%(vdi_uuid, "True")
                (rc, errlog, errmsg) = sshutil.vol_set_deleted(self.sr.conn, vdi_name, description)
                if (rc):
                    if (len(errlog)):
                        for id in range(len(errlog)):
                            util.SMlog(errlog[id])
                    errorstr = "Setting Volume deleted field for VDI (name %s)" % vdi_name
                    eql_raise_error(self.sr, errmsg, errorstr)
            else:
                (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.sr.conn, str(vdi_name))
                if (rc):
                    if (len(errlog)):
                        for id in range(len(errlog)):
                            util.SMlog(errlog[id])
                    errorstr = "listing volume details VDI (name %s)" % vdi_name
                    eql_raise_error(self.sr, errmsg, errorstr)
                if (vol_info['Status'] == "online"):
                        (rc, errlog, errmsg) = sshutil.vol_offline(self.sr.conn, vdi_name)
                        if (rc):
                            if (len(errlog)):
                                for id in range(len(errlog)):
                                    util.SMlog(errlog[id])
                            eql_raise_error(self.sr, errmsg,"Bringing down VDI %s offline:" % vdi_name)
                (rc, output, errlog, errmsg) = sshutil.vol_delete(self.sr.conn, vdi_name)
                if (rc) and errmsg.find('Volume %s does not exist' % vdi_name) == -1:                        
                    if (len(errlog)):
                        for id in range(len(errlog)):
                            util.SMlog(errlog[id])
                    errorstr = "Deleting VDI (name %s)" % vdi_name
                    eql_raise_error(self.sr, errmsg, errorstr)

        sm_config = self.session.xenapi.VDI.get_sm_config(self.sr.srcmd.params['vdi_ref'])
        if sm_config.has_key('SCSIid'):
            SCSIid = sm_config['SCSIid']
            key = "scsi-" + SCSIid
            try:
                self.session.xenapi.SR.remove_from_sm_config(self.sr.sr_ref, key)
            except:
                pass
        
        self.sr.deleted_vdi(vdi_uuid)
        #self.sr._updateSerial()
        self._db_forget()
        eql_exit(self.sr)

    def attach(self, sr_uuid, vdi_uuid):
        if not self.exists:
            eql_raise_error(self.sr,'','',code='VDIUnavailable')
        loop = 0
        while True:
            if not EXCL_MAP:
                self.activate(sr_uuid, vdi_uuid)
                eql_exit(self.sr)
            try:
                self.iscsi.attach(vdi_uuid)
            except:
                util.SMlog("Attempt %d; ISCSI attach failed" % loop)
                pass
        
            # Make sure the LUN0 path is available, wait up to an additional 60 secs
            if util.wait_for_path(self.path, 20):
                break
            
            if loop >= 3:
                util.SMlog("LUN path %s not detected after %d retries" % (self.path,loop))
                eql_raise_error(self,'VDIUnavailable','No LUN path detected',code='VDIUnavailable')
            loop += 1

	# Set default scheduler
        vdi_ref = self.sr.srcmd.params['vdi_ref']
        self.other_config = self.session.xenapi.VDI.get_other_config(vdi_ref)
        if self.other_config.has_key('scheduler') and self.other_config['scheduler'] != self.sr.sched:
            self.sr.sched = self.other_config['scheduler']
        realdev = os.path.realpath(self.path).split('/')[-1]
        util.set_scheduler(realdev, self.sr.sched)

        SCSIid = scsiutil.getSCSIid(self.path)
        key = "scsi-" + SCSIid
        try:
            self.session.xenapi.SR.add_to_sm_config(self.sr.sr_ref, key, vdi_uuid)
            cmd = ['/opt/xensource/sm/mpathcount.py',SCSIid]
            util.pread2(cmd)           
        except:
            pass
        try:
            vdi_ref = self.sr.srcmd.params['vdi_ref']
            self.session.xenapi.VDI.add_to_sm_config(vdi_ref, "SCSIid", SCSIid)
            scsiutil.update_XS_SCSIdata(vdi_uuid, \
                                        scsiutil.gen_raw_page_data(self.path))
        except:
            pass
        return super(EQUALVDI, self).attach(sr_uuid, vdi_uuid)

    def activate(self, sr_uuid, vdi_uuid):
        if (self.is_snapshot):
            vdi_name = "XenStorage" + sr_uuid.replace("-", "") + self.orig_vdi.replace("-", "")[:20]
            snap_name = "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        else:
            vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        if (self.is_snapshot):
            if (self.snap_info['Status'] != "online"):
                (rc, errlog, errmsg) = sshutil.snap_online(self.sr.conn, vdi_name, snap_name)
                if (rc):
                    if (len(errlog)):
                        for id in range(len(errlog)):
                            util.SMlog(errlog[id])
                    eql_raise_error(self.sr, errmsg,"Bringing snapshot VDI %s online:" % snap_name)
        else:
            if (self.vol_info['Status'] != "online"):
                (rc, errlog, errmsg) = sshutil.vol_online(self.sr.conn, vdi_name)
                if (rc):
                    if (len(errlog)):
                        for id in range(len(errlog)):
                            util.SMlog(errlog[id])
                    eql_raise_error(self.sr, errmsg,"Bringing SR VDI %s online:" % vdi_name)
        vdi_ref = self.sr.srcmd.params['vdi_ref']
	self.localIQN = self.session.xenapi.host.get_other_config(self.sr.host_ref)['iscsi_iqn']        
        (rc, output, errlog, errmsg) = sshutil.vol_create_access_rec_by_initiator(self.sr.conn, vdi_name, \
                                                                                  self.localIQN)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            eql_raise_error(self.sr, errmsg,"Creating access record for VDI %s" % vdi_name)

        if (self.is_snapshot):
            (rc, snap_info, errlog, errmsg) = sshutil.snap_detail(self.sr.conn, str(vdi_name), str(snap_name))
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "listing  details of snapshot VDI (name %s)" % snap_name
                eql_raise_error(self.sr, errmsg, errorstr)

            if snap_info['Status'] == 'offline-snap-reserve-met':
                util.SMlog("EQL: Volume out of space, failing to attach [%s]" % vdi_name)
                eql_exit(self.sr)
                raise xs_errors.XenError('EQLVolOutofSpace')
            targetIQN = snap_info['iSCSI Name']
        else:
            (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.sr.conn, str(vdi_name))
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "listing volume details VDI (name %s)" % vdi_name
                eql_raise_error(self.sr, errmsg, errorstr)

            targetIQN = vol_info['iSCSI Name']

        # Load the iscsi module
        original_srcmd = copy.deepcopy(self.sr.original_srcmd)
        original_srcmd.params['device_config']['targetIQN'] = targetIQN
        driver = SR.driver('iscsi')
        self.iscsi = driver(original_srcmd, vdi_uuid)
        self.iscsi.targetIQN = targetIQN
        self.iscsi.path = os.path.join("/dev/iscsi", self.iscsi.targetIQN, self.iscsi.address)
        self.path = os.path.join(self.iscsi.path, "LUN0")

    def detach(self, sr_uuid, vdi_uuid):
        if (self.is_snapshot):
            vdi_name = "XenStorage" + sr_uuid.replace("-", "") + self.orig_vdi.replace("-", "")[:20]
            snap_name = "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        else:
            vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]

        if not self.exists:
            eql_raise_error(self.sr,'','',code='VDIUnavailable')
        if not EXCL_MAP:
            self.deactivate(sr_uuid, vdi_uuid)
        loop = 0
        while loop <= 3:
            try:
                self.iscsi.detach(vdi_uuid)
                break
            except:
                util.SMlog("Attempt %d; ISCSI detach failed" % loop)
                loop += 1
                pass
            

        self.localIQN = self.session.xenapi.host.get_other_config(self.sr.host_ref)['iscsi_iqn']

        if (self.is_snapshot):
            (rc, snap_info, errlog, errmsg) = sshutil.snap_detail(self.sr.conn, str(vdi_name), str(snap_name))
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "listing snap details VDI (name %s)" % snap_name
                eql_raise_error(self.sr, errmsg, errorstr)

            if (snap_info['Status'] == "online"):
                if EXCL_MAP:
                    (rc, errlog, errmsg) = sshutil.snap_offline(self.sr.conn, vdi_name, snap_name)
                    if (rc):
                        if (len(errlog)):
                            for id in range(len(errlog)):
                                util.SMlog(errlog[id])
                        eql_raise_error(self.sr, errmsg,"Bringing down snapshot VDI %s offline:" % snap_name)
        else:
            (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.sr.conn, str(vdi_name))
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "listing volume details VDI (name %s)" % vdi_name
                eql_raise_error(self.sr, errmsg, errorstr)

            if (vol_info['Status'] == "online"):
                if EXCL_MAP:
                    (rc, errlog, errmsg) = sshutil.vol_offline(self.sr.conn, vdi_name)
                    if (rc):
                        if (len(errlog)):
                            for id in range(len(errlog)):
                                util.SMlog(errlog[id])
                        eql_raise_error(self.sr, errmsg,"Bringing down VDI %s offline:" % vdi_name)
        eql_exit(self.sr)
        return

    def deactivate(self, sr_uuid, vdi_uuid):
        if (self.is_snapshot):
            vdi_name = "XenStorage" + sr_uuid.replace("-", "") + self.orig_vdi.replace("-", "")[:20]
            snap_name = "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        else:
            vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]

        if (self.is_snapshot):
            (rc, snap_info, errlog, errmsg) = sshutil.snap_detail(self.sr.conn, str(vdi_name), str(snap_name))
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "listing  details of snapshot VDI (name %s)" % snap_name
                eql_raise_error(self.sr, errmsg, errorstr)

            targetIQN = snap_info['iSCSI Name']
        else:
            (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.sr.conn, str(vdi_name))
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "listing volume details VDI (name %s)" % vdi_name
                eql_raise_error(self.sr, errmsg, errorstr)

            targetIQN = vol_info['iSCSI Name']

        # Load the iscsi module
        original_srcmd = copy.deepcopy(self.sr.original_srcmd)
        original_srcmd.params['device_config']['targetIQN'] = targetIQN
        driver = SR.driver('iscsi')
        self.iscsi = driver(original_srcmd, vdi_uuid)
        self.iscsi.targetIQN = targetIQN
        self.iscsi.path = os.path.join("/dev/iscsi", self.iscsi.targetIQN, self.iscsi.address)
        # Explicitly remove the LUN from the scsi bus
        self.path = os.path.join(self.iscsi.path, "LUN0")
        try:
            realpath = os.path.realpath(self.path)
            if os.path.exists(realpath):
                SCSIid = scsiutil.getSCSIid(realpath)
                key = "scsi-" + SCSIid
                self.session.xenapi.SR.remove_from_sm_config(self.sr.sr_ref, key)
            if self.iscsi.devs.has_key(realpath):
                scsiutil.scsi_dev_ctrl(self.iscsi.devs[realpath], 'remove')
        except:
            pass
        return
    
    def resize(self, sr_uuid, vdi_uuid, size):
        if not self.exists:
            eql_raise_error(self.sr,'','',code='VDIUnavailable')
        current_size = self.size

        # Disallow LUN shrinking
        if size < current_size:
            util.SMlog('vdi_resize: shrinking not supported: ' + \
                       '(current size: %d, new size: %d)' % (current_size, size))
            errormsg = "Resize request"
            errorstr = "shrinking not supported"
            eql_raise_error(self.sr, errormsg, errorstr, code='VDISize')
        
        vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]

        #Size in Bytes, roundup to nearest MB
        size_mb = util.roundup(EQL_SIZE_INCREMENT, long(size)) / (1024 * 1024)
        size_str = "%dMB"% long(size_mb)
        if size > current_size:
            diff = size - current_size
            (rc, errlog, errmsg) = sshutil.vol_grow(self.sr.conn, vdi_name, size_str)
        else:
            (rc, errlog, errmsg) = sshutil.vol_shrink(self.sr.conn, vdi_name, size_str)

        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            errorstr = "resizing volume VDI (name %s)" % vdi_name
            eql_raise_error(self.sr, errmsg, errorstr)

        (rc, vol_info, errlog, errmsg) = sshutil.vol_detail(self.sr.conn, str(vdi_name))
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            errorstr = "listing volume details VDI (name %s)" % vdi_name
            eql_raise_error(self.sr, errmsg, errorstr)
        self.size = str_to_bytes(vol_info['Size'])
        self.utilisation = str_to_bytes(self.obj['VolReserveInUse'])
        self._db_update()
        eql_exit(self.sr)
        return super(EQUALVDI, self).get_params()

    def _snapshot(self, sr_uuid, vdi_uuid, dest):
        # Used internally by clone and snapshot
        if not self.exists:
            eql_raise_error(self.sr,'','',code='VDIUnavailable')

        self.sr.srcmd.params.pop("vdi_ref")
        leaf_vdi = VDI.VDI(self.sr, dest) # user-visible leaf VDI
        leaf_vdi.read_only = False
        leaf_vdi.location = dest
        leaf_vdi.size = self.size
        leaf_vdi.utilisation = self.utilisation        
        return leaf_vdi

    def snapshot(self, sr_uuid, vdi_uuid):
        if (self.is_snapshot):
            eql_raise_error(self.sr,'','',code='EQLSnapshotOfSnapshot')
        dest = util.gen_uuid ()
        self.comment = True
        leaf_vdi = self._snapshot(sr_uuid, vdi_uuid, dest)
        description = "SNAP-%s:SRC-%s:UTIL-%s"%(dest, vdi_uuid, str(leaf_vdi.utilisation))        
        vdi_name = "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        (rc, errlog, errmsg) = sshutil.vol_create_snapshot(self.sr.conn, vdi_name, sr_uuid, dest, description)
        if (rc):
            if (len(errlog)):
                for id in range(len(errlog)):
                    util.SMlog(errlog[id])
            errorstr = "Creating snapshot of VDI (name %s)" % vdi_name
            eql_raise_error(self.sr, errmsg, errorstr)

        try:
            leaf_vdi_ref = leaf_vdi._db_introduce()
            sm_config = self.session.xenapi.VDI.get_sm_config(leaf_vdi_ref)
            sm_config['snapshot-of'] = vdi_uuid
            self.session.xenapi.VDI.set_sm_config(leaf_vdi_ref, sm_config)
            #util.SMlog("vdi_snapshot: introduced VDI: %s" % leaf_vdi_ref)
        except Exception, e:
            util.SMlog("vdi_snapshot: caught error during VDI.db_introduce: %s" % (str(e)))
            raise
        eql_exit(self.sr)
        return leaf_vdi.get_params()
            
    def clone(self, sr_uuid, vdi_uuid):
        dest = util.gen_uuid ()
        if not self.exists:
            eql_raise_error(self.sr,'','',code='VDIUnavailable')

        self.comment = False
        vdi_ref = self.sr.srcmd.params['vdi_ref']
        leaf_vdi = self._snapshot(sr_uuid, vdi_uuid, dest)

        self.clone_of_snapshot = False
        self.sm_config = self.session.xenapi.VDI.get_sm_config(vdi_ref)
        if self.sm_config.has_key('snapshot-of'): 
            self.orig_vdi = self.sm_config['snapshot-of']
            self.clone_of_snapshot = True
            vdi_name = "XenStorage" + sr_uuid.replace("-", "") + self.orig_vdi.replace("-", "")[:20]
            snap_name = "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]
        else:
            vdi_name= "XenStorage" + sr_uuid.replace("-", "") + vdi_uuid.replace("-", "")[:20]

        description = "UUID-%s:DELETED-%s"%(dest, "False")
        snapreserve = 100
        snapdepletion = "volume-offline"

        if (self.clone_of_snapshot):
            (rc, output, iscsiname, errlog, errmsg) = sshutil.vol_create_clone_from_snapshot(self.sr.conn, vdi_name, snap_name, sr_uuid, dest, description, snapreserve, str(snapdepletion)) 
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "Creating clone of VDI (name %s)" % vdi_name
                eql_raise_error(self.sr, errmsg, errorstr)

        else:
            (rc, output, iscsiname, errlog, errmsg) = sshutil.vol_create_clone(self.sr.conn, vdi_name, sr_uuid, dest, description, snapreserve, str(snapdepletion)) 
            if (rc):
                if (len(errlog)):
                    for id in range(len(errlog)):
                        util.SMlog(errlog[id])
                errorstr = "Creating clone of VDI (name %s)" % vdi_name
                eql_raise_error(self.sr, errmsg, errorstr)

        try:
            leaf_vdi.read_only = False
            leaf_vdi_ref = leaf_vdi._db_introduce()
            util.SMlog("vdi_clone: introduced VDI: %s" % leaf_vdi_ref)
        except Exception, e:
            util.SMlog("vdi_clone: caught error during VDI.db_introduce: %s" % (str(e)))
            raise
        eql_exit(self.sr)
        return leaf_vdi.get_params()

def str_to_bytes(size_str):
    size_val_str = re.match('^[0-9.]*', size_str).group()
    size_val = float(size_val_str)

    size_unit_str = re.search('[a-zA-Z]+', size_str).group()

    if (size_unit_str == "MB"):
        size_in_bytes = int(size_val * 1024 * 1024)
        return size_in_bytes

    if (size_unit_str == "GB"):
        size_in_bytes = int((size_val * 1024) + 0.99) * (1024 * 1024)
        return size_in_bytes

    if (size_unit_str == "TB"):
        size_in_bytes = int((size_val * 1024) + 0.99) * (1024 * 1024 * 1024)
        return size_in_bytes
    return int(size_val)

def eql_raise_error(srobj, reason, st, code='EIO'):
    try:
        if srobj.has_attr('sr'):
            srobj.sr.conn.close()
        else:
            srobj.conn.close()
    except:
        pass
    err = ''
    if len(reason):
        err = '%s:%s' % (st,reason)
    raise xs_errors.XenError(code, \
                             opterr=err)

def eql_exit(srobj):
    srobj.conn.close()

def match_vol_desc(s):
    regex = re.compile("^UUID-[0-9a-f]{8}-(([0-9a-f]{4})-){3}[0-9a-f]{12}:DELETED-[True|False]")
    return regex.search(s, 0)

def match_snap_vol_desc(s):
    regex = re.compile("^SNAP-[0-9a-f]{8}-(([0-9a-f]{4})-){3}[0-9a-f]{12}:SRC-[0-9a-f]{8}-(([0-9a-f]{4})-){3}[0-9a-f]{12}:UTIL-")
    return regex.search(s, 0)

def _targetIQNtoXML(target, iqn):
    dom = xml.dom.minidom.Document()
    element = dom.createElement("iscsi-target-iqns")
    dom.appendChild(element)

    entry = dom.createElement('TGT')
    element.appendChild(entry)

    subentry = dom.createElement('Index')
    entry.appendChild(subentry)           
    textnode = dom.createTextNode('0')
    subentry.appendChild(textnode)
            
    subentry = dom.createElement('IPAddress')
    entry.appendChild(subentry)           
    textnode = dom.createTextNode(target)
    subentry.appendChild(textnode)

    subentry = dom.createElement('TargetIQN')
    entry.appendChild(subentry)           
    textnode = dom.createTextNode(iqn)
    subentry.appendChild(textnode)
    
    return dom.toprettyxml()

def _storagepooltoXML(Storagepoollist):
    dom = xml.dom.minidom.Document()
    driver = dom.createElement("StoragePoollist")
    dom.appendChild(driver)

    for key in Storagepoollist.keys():
        entry = dom.createElement("StoragePool")
        driver.appendChild(entry)
        
        args = Storagepoollist[key].split('|')
        e = dom.createElement("Name")
        entry.appendChild(e)
        textnode = dom.createTextNode(key)
        e.appendChild(textnode)

        e = dom.createElement("Default")
        entry.appendChild(e)
        textnode = dom.createTextNode(args[0])
        e.appendChild(textnode)

        e = dom.createElement("Members")
        entry.appendChild(e)
        textnode = dom.createTextNode(args[1])
        e.appendChild(textnode)

        e = dom.createElement("Volumes")
        entry.appendChild(e)
        textnode = dom.createTextNode(args[2])
        e.appendChild(textnode)

        e = dom.createElement("Capacity")
        entry.appendChild(e)
        textnode = dom.createTextNode(str(str_to_bytes(args[3])))
        e.appendChild(textnode)

        e = dom.createElement("FreeSpace")
        entry.appendChild(e)
        textnode = dom.createTextNode(str(str_to_bytes(args[4])))
        e.appendChild(textnode)
    return dom.toprettyxml()

def _getDateString():
    d = datetime.datetime.now()
    t = d.timetuple()
    return "%s-%s-%s:%s:%s:%s" % \
          (t[0],t[1],t[2],t[3],t[4],t[5]) 

def _uuid_to_long(uuid):
    return long(uuid.replace('-',''), 16)
    
if __name__ == '__main__':
    SRCommand.run(EQUALSR, DRIVER_INFO)
else:
    SR.registerSR(EQUALSR)
