#!/usr/bin/env python
# Copyright (c) 2005-2007 XenSource, Inc. All use and distribution of this
# copyrighted material is governed by and subject to terms and conditions
# as licensed by XenSource, Inc. All other rights reserved.
# Xen, XenSource and XenEnterprise are either registered trademarks or
# trademarks of XenSource Inc. in the United States and/or other countries.
#
#
# NETAPPSR: Netapp Filer storage repository
#

import SR, VDI, ISCSISR, SRCommand, util, scsiutil
import os, sys, re, random
import datetime, time
import xml.dom.minidom
import xs_errors
sys.path.insert(0, "/opt/xensource/sm/netapp")
import netapplib

# N.B. You *must* restart xapi after making any changes to the EXCL_MAP flag
EXCL_MAP = False
CAPABILITIES = ["SR_PROBE","SR_UPDATE","VDI_CREATE","VDI_DELETE","VDI_ATTACH",
                "VDI_DETACH","VDI_CLONE","VDI_SNAPSHOT","VDI_RESIZE"]
if EXCL_MAP:
    CAPABILITIES.append("VDI_ACTIVATE")
    CAPABILITIES.append("VDI_DEACTIVATE")

CONFIGURATION = [ [ 'target', 'IP address or hostname of the NETAPP filer' ], \
                  [ 'controlport', 'Control connection port for the NETAPP filer (optional, defaults to 80)' ], \
                  [ 'usehttps', 'use OpenSSL for a secure transport connection, true or false (optional, defaults to false)' ], \
                  [ 'username', 'Login username to be used by the driver for manipulating LUNs on the filer' ], \
                  [ 'password', 'Login password to be used by the driver for manipulating LUNs on the filer' ], \
                  [ 'aggregate', 'The aggregate name on which the Flexvol should be created on the filer. (optional)' ], \
                  [ 'FlexVols', 'The number of FlexVols to allocate per-SR. (optional)' ], \
                  [ 'chapuser', 'The username to be used during CHAP authentication (optional)' ], \
                  [ 'chappassword', 'The password to be used during CHAP authentication (optional)' ], \
                  [ 'allocation', 'Valid values are thick or thin (optional, defaults to thick)'], \
                  [ 'asis', 'Enable A-SIS de-duplication for this SR, true or false (optional, defaults to false)'], \
                  [ 'multihomed', 'Enable multi-homing to this target, true or false (optional, defaults to same value as host.other_config:multipathing)'] ]


DRIVER_INFO = {
    'name': 'NetApp',
    'description': 'SR driver for mapping of LUNs to VDIs on a NETAPP server, providing use of fast snapshot and clone features on the filer',
    'vendor': 'Citrix Systems Inc',
    'copyright': '(C) 2008 Citrix Systems Inc',
    'driver_version': '0.1',
    'required_api_version': '0.1',
    'capabilities': CAPABILITIES,
    'configuration': CONFIGURATION
    }

NAPP_VERSION = 7.0
NULL = 0
FVOLNUM_DEFAULT = 8
MAXVOL = 32
FVOL_MULTIPLIER = 2.4
MIN_LUNSIZE = 4 * 1024 * 1024
PORTNUM_DEFAULT = 80
HTTPS_PORTNUM_DEFAULT = 443
# 2^16 Max port number value
MAXPORT = 65535

#Subset of relevant error codes
ZAPI_AUTHENTICATION = 13002
ZAPI_TARGET = 2
INITGROUP_EXISTS = 9004
INITGROUP_MAP_EXISTS = 9023
INITGROUP_HAS_LUN = 9024
INITGROUP_MAP_EXISTS2 = 9039
INITGROUP_HAS_NODE = 9008
LUN_SIZE_TOO_LARGE = 9034
SNAPSHOTEXISTS = 13020
NOMORESNAPS = 135
INVALIDIQN = 9006
VOL_ONLINE = 13060
COMPAT_VOL_ONLINE = 13114
APILICENSE = 13008


class NETAPPSR(ISCSISR.ISCSISR):
    """Netapp filer storage repository"""
    def handles(type):
        if type == "netapp":
            return True
        return False
    handles = staticmethod(handles)

    def load(self, sr_uuid):
        # Required keys:
        #   target, user, password
        
        self.sr_vditype = 'aio'
        
        if not self.dconf.has_key('target') or not self.dconf['target']:
            raise xs_errors.XenError('NAPPTarget')
        if not self.dconf.has_key('username') or not self.dconf['username']:
            raise xs_errors.XenError('NAPPUsername')
	# Allow empty password string
        if not (self.dconf.has_key('password') or self.dconf.has_key('password_secret')):
            raise xs_errors.XenError('NAPPPassword')

        # Optional dconf params
        self.aggregate = ''
        if self.dconf.has_key('aggregate') and self.dconf['aggregate']:
            self.aggregate = self.dconf['aggregate']

        self.isMaster = False
        if self.dconf.has_key('SRmaster') and self.dconf['SRmaster'] == 'true':
            self.isMaster = True
        
        self.fvolnum = FVOLNUM_DEFAULT
        if self.dconf.has_key('FlexVols'):
            try:
                fvols = long(self.dconf['FlexVols'])
            except:
                raise xs_errors.XenError('InvalidArg', \
                        opterr='No. of FlexVols must be a value between 1 and %d' % MAXVOL)
            if fvols > MAXVOL:
                self.fvolnum = MAXVOL
            elif fvols == 0:
                raise xs_errors.XenError('InvalidArg', \
                        opterr='No. of FlexVols must be a value between 1 and %d' % MAXVOL)
            else:
                self.fvolnum = fvols

        self.target = self.dconf['target']
        self.username = self.dconf['username']
        if self.dconf.has_key('password_secret'):
            self.password = util.get_secret(self.session, self.dconf['password_secret'])
        else:
            self.password = self.dconf['password']

        self.fvname = "XenStorage_%s" % sr_uuid.replace("-","_")
        self.remotepath = "/vol/%s" % self.fvname
        self.fvol = []
        self.luns = []
        self.serialstring = ''
        self.exists = 0
        try:
            self.serialstring = scsiutil.get_serial_record(self.session, self.sr_ref)
        except:
            pass
        
        #Attempt to connect
        if self.dconf.has_key('multihomelist'):
            targetlist = self.dconf['multihomelist'].split(',')
        else:
            targetlist = ['%s:' % self.target]
        self.conn = False
        self.port = 0
        if self.dconf.has_key('controlport'):
            try:
                self.port = long(self.dconf['controlport'])
                if self.port > MAXPORT or self.port < 1:
                    raise
            except:
                raise xs_errors.XenError('InvalidArg', \
                        opterr='Controlport argument must be an integer between 1 and %d' % MAXPORT)
        SSL = False
        if self.dconf.has_key('usehttps') and self.dconf['usehttps'] == 'true':
            SSL = True
        for val in targetlist:
            (target,port) = val.split(':')
            try:
                self.sv = na_open_session(target, self.username, self.password, port=self.port, secure=SSL)
                self.target = target
                self.conn = True
                break
            except SR.SRException, e:
                if not e.errno == 124:
                    raise
        if not self.conn and self.cmd in ['sr_detach','vdi_detach']:
            return
        elif not self.conn:
            raise xs_errors.XenError('NAPPTargetFailed')
        
        out = netapplib.na_server_invoke(self.sv, "iscsi-node-get-name", NULL);
        if not na_test_result(out):
            raise xs_errors.XenError('NAPPTargetIQN')
        targetIQN = netapplib.na_child_get_string(out, "node-name")
        if self.dconf.has_key('targetIQN'):
            if self.dconf['targetIQN'] != targetIQN:
                print >>sys.stderr,_targetIQNtoXML(self.target, targetIQN)
                raise xs_errors.XenError('ConfigTargetIQNMissing', \
                             opterr='N.B. This parameter is optional and can safely be excluded')
        else:
            self.original_srcmd.params['device_config']['targetIQN'] = targetIQN

        # Unless explicitly set, we assign the multi-homing status to be the same as Mpath
        if not self.original_srcmd.params['device_config'].has_key('multihomed'):
            self.original_srcmd.params['device_config']['multihomed'] = self.mpath

        # Default to thick provisioning, N.B. _getstats tests for 'space_reservation==none'
        self.allocation = True
        if util.match_uuid(sr_uuid):
            self.exists = self._getFVNum()
        if self.cmd in ["sr_update","vdi_create","vdi_resize","vdi_clone","vdi_snapshot"]:
            self.virtual_allocation = long(self.session.xenapi.SR.get_virtual_allocation(self.sr_ref))

        self.enforce = False
        self.multiplier = FVOL_MULTIPLIER
        try:
            self.other_config = self.session.xenapi.SR.get_other_config(self.sr_ref)
            if self.other_config.has_key('enforce_allocation') and self.other_config['enforce_allocation'] == 'true':
                self.enforce = True
            if self.other_config.has_key('multiplier'):
                self.multiplier = float(self.other_config['multiplier'])
        except:
            pass

        # Load the iscsi module
        driver = SR.driver('iscsi')
        self.iscsi = driver(self.original_srcmd, sr_uuid)
        self.igroup = "XenStorage-%s" % self.iscsi.localIQN
        self.iscsi.targetIQN = targetIQN
        self.iscsi.path = os.path.join("/dev/iscsi",self.iscsi.targetIQN,self.iscsi.address)

    def create(self, sr_uuid, size):
        if self.exists:
            raise xs_errors.XenError('SRExists')

        if not self.aggregate:
            print >>sys.stderr,_aggregatetoXML(self._getAllAggrs())
            raise xs_errors.XenError('NAPPAggregate')

        # Test for thick vs thin provisioning conf parameter
        if self.dconf.has_key('allocation') and self.dconf['allocation'] == 'thin':
            self.allocation = False

        # Test for asis conf parameter, default to False
        asis = False
        if self.dconf.has_key('asis') and self.dconf['asis'] == 'true':
            asis = True
            
        if self.allocation:
            allocate = 'volume'
            # Set the initial size to the min requirement => 20MB
            size_str = "20m"
        else:
            allocate = 'none'
            # Set the volume size to the aggregate size
            self._getAggrstats(self.aggregate)
            size_str = "%dm" % (self.physical_size/(1024 * 1024))
        # Create flexvols
        for i in range(0, self.fvolnum):
            name = "%s_FV%d" % (self.fvname,i)
            path = "/vol/%s" % name
            out = netapplib.volume_create_wrapper(self.sv, str(self.aggregate), \
                                                  str(name), str(size_str), str(allocate));
            if not na_test_result(out):
                #Delete vols that have been created
                self.delete(sr_uuid)
                na_raise_error(out,self.sv,"creating flexvol")
            self.exists += 1
            out = netapplib.volume_snap_reserve_wrapper(self.sv, "0", str(name));
            if not na_test_result(out):
                self.delete(sr_uuid)
                na_raise_error(out,self.sv,"Disabling vol snap reserve")
            out = netapplib.volume_guarantee_wrapper(self.sv, str(allocate), str(name));
            if not na_test_result(out):
                self.delete(sr_uuid)
                na_raise_error(out,self.sv,"Setting vol space guarantee")
            out = netapplib.volume_disableautosnaps_wrapper(self.sv, str(name));
            if not na_test_result(out):
                self.delete(sr_uuid)
                na_raise_error(out,self.sv,"Disabling vol auto snap")

            #Disable NFS export for volume
            out = netapplib.single_export_delete_wrapper(self.sv,str(path));
            if not na_test_result(out) and \
                   netapplib.na_results_errno(out) != APILICENSE:
                self.delete(sr_uuid)
                na_raise_error(out,self.sv,"Disabling NFS export")

            # Check for A-SIS flag
            if asis:
                util.SMlog("Enabling A-SIS")
                out = netapplib.volume_asis_enable(self.sv,str(path));
                if not na_test_result(out):
                    self.delete(sr_uuid)
                    if netapplib.na_results_errno(out) == APILICENSE:
                        raise xs_errors.XenError('NAPPAsisLicense')
                    else:
                        raise xs_errors.XenError('NAPPAsisError')

    def delete(self, sr_uuid):
        if not self.exists:
            return
        tmp = range(0, self.exists)
        tmp.reverse()
        for i in tmp:
            name = "%s_FV%d" % (self.fvname,i)
            out = netapplib.volume_offline_wrapper(self.sv, str(name));
            if not na_test_result(out):
                pass
            out = netapplib.volume_destroy_wrapper(self.sv, str(name));
            if not na_test_result(out):
                na_raise_error(out,self.sv,"destroying flexvol")
        self.iscsi.detach(sr_uuid)
        if not self.iscsi.attached:
            self._destroyIGRP()

    def attach(self, sr_uuid):
        if not self.exists:
            raise xs_errors.XenError('SRUnavailable')
        util.SMlog("Attaching SR [%s]" % sr_uuid)
        util.SMlog("Found %d associated FlexVols" % self.exists)
        for i in range(0, self.exists):
            name = "%s_FV%d" % (self.fvname,i)
            out = netapplib.volume_online_wrapper(self.sv, name)
            if not na_test_result(out) and \
                   netapplib.na_results_errno(out) != VOL_ONLINE and \
                   netapplib.na_results_errno(out) != COMPAT_VOL_ONLINE:
                na_raise_error(out,self.sv,"Bringing volume %s online:" % name)
        out = netapplib.na_server_invoke(self.sv, "iscsi-service-status", NULL);
        if not na_test_result(out):
            na_raise_error(out,self.sv,"Querying iSCSI service status")
        if netapplib.na_child_get_string(out, "is-available") == "false":
            raise xs_errors.XenError('NAPPNoISCSIService', \
                opterr='N.B. Filer admin must logon to target and start iSCSI service')
        self._initIGRP()
        self.iscsi.attach(sr_uuid)
        self._setMultipathableFlag()
        self._addLUNperVDIkey()

    def probe(self):
        return util.SRtoXML(self._getAllSRs())

    def detach(self, sr_uuid):
        if not self.exists:
            return
        self._getLUNs()
        activePBDs = util.test_activePoolPBDs(self.session, self.host_ref, self.sr_ref)
        for i in range(0, self.exists):
            name = "%s_FV%d" % (self.fvname,i)
            for val in self.luns:
                if match_XSFV(val.path,name):
                    #test LUN map
                    LUNmaps = self._getLUNMap(val.path, False)
                    for subval in LUNmaps:
                        entries = val.path.split('/')
                        vdi_uuid = entries.pop()
                        obj = NETAPPVDI(self, vdi_uuid)
                        obj.detach(sr_uuid, vdi_uuid)
            if not activePBDs:
                util.SMlog("No other active PBDs, setting  vol to offline [%s]" % \
                           name)
                netapplib.volume_offline_wrapper(self.sv, name)
        self.iscsi.detach(sr_uuid)
        if not self.iscsi.attached:
            self._destroyIGRP()

    def update(self, sr_uuid):
        if not self.exists:
            raise xs_errors.XenError('SRUnavailable')
        #Stats were already set in get_FVNum()
        self._db_update()
        
    def scan(self, sr_uuid):
        if not self.exists:
            raise xs_errors.XenError('SRUnavailable')
        self._loadvdis()
        self._cleanupsnaps()
        self.passthrough = True
        self._updateSerial()
        if self.enforce:
            for i in range(0, self.exists):
                try:
                    self._synch_vol_size(0, i)
                except:
                    pass
        return super(NETAPPSR, self).scan(sr_uuid)

    def content_type(self, sr_uuid):
        return super(NETAPPSR, self).content_type(sr_uuid)

    def vdi(self, uuid):
        # A subset of VDI commands require a full LUN query
        if self.cmd in ['vdi_create','vdi_resize','vdi_clone','vdi_delete']:
            self._getLUNs()
        return NETAPPVDI(self, uuid)

    def added_vdi(self, vdi):
        self.vdis[vdi.uuid] = vdi
        self._addtoserial(vdi.serial)

    def deleted_vdi(self, uuid):
        if uuid in self.vdis:
            self._delfromserial(self.vdis[uuid].serial)
            del self.vdis[uuid]

    def _loadvdis(self):
        if self.vdis:
            return
        self._getLUNs()
        for val in self.luns:
            self.added_vdi(self.vdi(val.name))

    def _addtoserial(self,serial):
        if self.serialstring.find(serial) == -1:
            self.serialstring = "%s,%s" % (serial,self.serialstring) 

    def _delfromserial(self,serial):
        if self.serialstring.find(serial) != -1:
            newserial = ""
            for val in self.serialstring.split(','):
                if len(val) > 0 and val != serial:
                    newserial = "%s,%s" % (val,newserial)
            self.serialstring = newserial

    def _updateSerial(self):
        scsiutil.add_serial_record(self.session, self.sr_ref, \
                                   self.serialstring)

    def _snapchild(self, snap, volnum):
        if len(self.luns):
            for val in self.luns:
                if val.parentsnap == snap and val.fvnum == volnum:
                    return True
        return False

    
    ###NETAPP SR Utility Functions###
    def _cleanupsnaps(self):
        for i in range(0, self.exists):
            volname = "%s_FV%d" % (self.fvname,i)
            #Query list of snapshots
            out = netapplib.fvol_snaplist_wrapper(self.sv, volname)
            if not na_test_result(out):
                continue
            temp = netapplib.na_elem_child(out, "snapshots");
            iter = netapplib.iterator_wrapper(temp, "name")
            if not iter:
                continue
            for val in iter.replace('|','').split(','):
                if val and util.match_uuid(val) and not self._snapchild(val, i):
                    util.SMlog("Removing unused snap (%s)" % val)
                    out = netapplib.fvol_snapdelete_wrapper(self.sv, val, volname)
                    if not na_test_result(out):
                        pass
        
    def _snapshot(self, fvname, SnapshotID):
        out = netapplib.fvol_snap_wrapper(self.sv, str(SnapshotID), fvname);
        if not na_test_result(out) and not netapplib.na_results_errno(out) == SNAPSHOTEXISTS:
            if netapplib.na_results_errno(out) == NOMORESNAPS:
                raise xs_errors.XenError('NAPPSnapLimit')
            else:
                na_raise_error(out,self.sv,"creating flexvol snapshot")
        return True

    def _getFVNum(self):
        for i in range(0, MAXVOL):
            try:
                self._getstats(self.sv, i)
            except:
                if i > 0:
                    self._getAggrstats(self.fvol[0].aggr)
                return i
        return MAXVOL

    def _getASISstatus(self):
        out = netapplib.na_server_invoke(self.sv, "license-list-info", NULL)
        if not na_test_result(out):
            return False
        temp = netapplib.na_elem_child(out, "licenses")
        iter = netapplib.iterator_wrapper(temp, \
                    "service|code")
        if not iter:
            return False
        for val in iter.split(','):
             if not val:
                 continue
             args = val.split('|')
             if args[0] != "a_sis":
                 continue
             if args[1] != 'null':
                 return True
             else:
                 return False
        return False

    def _getstats(self, handle, subvol):
        path = "%s_FV%d" % (self.remotepath, subvol)
        out = netapplib.volume_info_wrapper(handle,path)
        if not na_test_result(out):
            na_raise_error(out,handle,"Querying vol info")
        obj = NETAPPFVOL(subvol, path)
        temp = netapplib.na_elem_child(out, "volumes");
        iter = netapplib.iterator_wrapper(temp, \
                    "size-used|containing-aggregate|space-reserve|size-total")
        if not iter:
            na_raise_error(out,handle,"Querying vol info")
        args = iter.replace(',','').split('|')
        obj.physical_utilisation = long(args[0])
        obj.aggr = args[1]
        reserve = args[2]
        obj.size = long(args[3])
        if reserve == "none":
            self.allocation = False
        self.fvol.append(obj)

    def _getAllAggrs(self):
        asis = self._getASISstatus()
        out = netapplib.na_server_invoke(self.sv, "aggr-list-info", NULL);
        if not na_test_result(out):
            na_raise_error(out,self.sv,"Querying aggr info")
        temp = netapplib.na_elem_child(out, "aggregates");
        Aggrlist = {}
        iter = netapplib.iterator_wrapper(temp, \
                   "name|type|size-total|disk-count|raid-status")
        if not iter:
            return Aggrlist
        for val in iter.split(','):
            if val:
                args = val.split('|')
                if args[1] == 'aggr':
                    val += str(asis)
                    Aggrlist[args[0]] = val
        return Aggrlist

    def _getAggrstats(self, aggr):
        out = netapplib.aggregate_info_wrapper(self.sv, str(aggr))
        if not na_test_result(out):
            na_raise_error(out, self.sv, "Querying aggregate info")
        temp = netapplib.na_elem_child(out, "aggregates");
        iter = netapplib.iterator_wrapper(temp, \
                                          "size-total|size-used")
        if not iter:
            na_raise_error(out, self.sv, "Querying aggregate info")
        args = iter.replace(',','').split('|')
        self.physical_size = long(args[0])
        self.physical_utilisation = long(args[1])

    def _getAllSRs(self):
        out = netapplib.volume_all_info_wrapper(self.sv)
        if not na_test_result(out):
            na_raise_error(out,self.sv,"Querying vol info")
        temp = netapplib.na_elem_child(out, "volumes");

        aggrlist = self._getAllAggrs()
        SRlist = {}
        iter = netapplib.iterator_wrapper(temp, "name|size-total|containing-aggregate")
        if not iter:
            return SRlist
        for val in filter(match_XS, iter.split(',')):
            entry = {}
            args = val.split('|')
            key = args[0].split('_FV')[:-1]
            uuid = key[0].replace('_','-').replace('XenStorage-','')
            entry['size'] = args[1]
            entry['aggregate'] = args[2]
            if SRlist.has_key(uuid):
                e = SRlist[uuid]
                e['size'] = str(long(e['size'])+long(entry['size']))
                if e['aggregate'].find(entry['aggregate']) != 0:
                    e['aggregate'] += ",%s" % entry['aggregate']
                SRlist[uuid] = e
            else:
                SRlist[uuid] = entry
        return SRlist

    def _updateLUNs(self, child):
        iter = netapplib.iterator_wrapper(child, "path|size|backing-snapshot")
        if not iter:
            return
        count = 0
        for val in iter.split(','):
            if match_XSFV(val, self.fvname):
                args = val.split('|')
                entries = args[0].split('/')
                size = long(args[1])
                lun = entries.pop()
                vol = entries.pop()
                volnum = vol.split('_FV')[1]
                comment = self._getLUNComment(args[0])
                serial = self._getLUNSerial(args[0])
                snapLUN = False
                if is_snap(comment):
                    snapLUN = True
                obj = NETAPPLUN(long(volnum), self.remotepath, lun, size, snapLUN, comment, serial)
                if args[2].find('null') == -1:
                    parentlist = args[2].split('/')
                    id = parentlist.pop()
                    parent = parentlist.pop()
                    obj.parentsnap = parent
                self.luns.append(obj)
                count += 1
        return count


    def _getLUNs(self):
        if len(self.luns):
            return
        out = netapplib.na_server_invoke(self.sv, "lun-list-info", NULL);
        if not na_test_result(out):
            return
        temp = netapplib.na_elem_child(out, "luns");
        self._updateLUNs(temp)
        self.virtual_allocation = 0
        for lun in self.luns:
            if not is_snap(lun.comment):
                self.virtual_allocation += lun.size

    def _getLUN(self, vdi_uuid):
        for i in range(0, self.exists):
            path = "%s_FV%d/%s" % (self.remotepath,i,vdi_uuid)
            out = netapplib.lun_info_wrapper(self.sv, path);
            if na_test_result(out):
                temp = netapplib.na_elem_child(out, "luns");
                if self._updateLUNs(temp):
                    return self.luns[-1]
        return 0

    def _getLUNid(self, path):
        out = netapplib.lun_map_list_wrapper(self.sv, path);
        if not na_test_result(out):
            na_raise_error(out,self.sv,"lun_map_list query")
        temp = netapplib.na_elem_child(out, "initiator-groups");
        iter = netapplib.iterator_wrapper(temp, \
                       "initiator-group-name|lun-id")
        if not iter:
            raise xs_errors.XenError('EIO')
        for subval in iter.split(','):
            if len(subval) and match_XSFV(subval, self.igroup):
                try:
                    LUNid = long(subval.split('|')[1])
                    util.SMlog("Returning LUNid %d" % LUNid)
                    return LUNid
                except:
                    raise xs_errors.XenError('VDIUnavailable')
        raise xs_errors.XenError('VDIUnavailable')

    def _getLUNComment(self, path):
        out = netapplib.lun_get_comment_wrapper(self.sv, path);
        if not na_test_result(out):
            return ""
        else:
            return netapplib.na_child_get_string(out, "comment")

    def _getLUNSerial(self, path):
        out = netapplib.lun_get_serial_wrapper(self.sv, path);
        if not na_test_result(out):
            na_raise_error(out,self.sv,"lun_serial query")
        return netapplib.na_child_get_string(out, "serial-number")

    def _getLUNMap(self, path, allgrps):
        #test LUN map
        LUNmap = []
        out = netapplib.lun_map_list_wrapper(self.sv, path);
        if not na_test_result(out):
            na_raise_error(out,self.sv,"lun_map_list query")
        temp = netapplib.na_elem_child(out, "initiator-groups");
        iter = netapplib.iterator_wrapper(temp, "initiator-group-name")
        if not iter:
            return LUNmap
        for subval in iter.replace('|','').split(','):
            if not len(subval):
                continue
            util.SMlog("GETLUNMAP: [%s]" % subval)
            if not allgrps:
                if match_XSFV(subval, self.igroup):
                    LUNmap.append(subval)
            else:
                LUNmap.append(subval)
        return LUNmap
        
    def _initIGRP(self):
        out = netapplib.igroup_create_wrapper(self.sv, str(self.igroup));
        if not na_test_result(out):
            if long(netapplib.na_results_errno(out)) != INITGROUP_EXISTS:
                na_raise_error(out,self.sv,"creating igroup")
        out = netapplib.igroup_add_wrapper(self.sv, str(self.iscsi.localIQN), \
                                           str(self.igroup));
        if not na_test_result(out):
            if long(netapplib.na_results_errno(out)) == INITGROUP_HAS_NODE:
                return
            elif long(netapplib.na_results_errno(out)) == INVALIDIQN:
                raise xs_errors.XenError('InvalidIQN', \
                        opterr='Value provided was [%s]' % self.iscsi.localIQN)
            else:
                na_raise_error(out,self.sv,"adding node to igroup")

    def _destroyIGRP(self):
        out = netapplib.igroup_destroy_wrapper(self.sv, str(self.igroup));
        if not na_test_result(out):
            #We do not forcibly remove in case disk still in use
            util.SMlog("Igroup destroy failed (not fatal): %s" \
                       % netapplib.na_results_reason(out))
        else:
            util.SMlog("Igroup destroy succeeded %s" % self.igroup)

    def _synch_vol_size(self, size, fvnum, decr=False):
        if not self.allocation:
            return
        if not len(self.luns):
            self._getLUNs()
        f_alloc = 0
        for val in self.luns:
            if val.fvnum == fvnum:
                f_alloc += val.size
        if not decr:
            f_alloc = long(na_size_estimate(f_alloc, self.multiplier) + na_size_estimate(long(size),self.multiplier) + (20 * 1024 * 1024))
        else:
            f_alloc = long(na_size_estimate(f_alloc, self.multiplier) - na_size_estimate(long(size),self.multiplier) + (20 * 1024 * 1024))
        f_actual = long(self.fvol[fvnum].size)
        size_str = "%dk" % long(f_alloc/1024)
        out = None
        if self.enforce and f_actual != f_alloc:
            util.SMlog("Enforce True, F_ALLOC: %d; F_ACTUAL: %d" % (f_alloc,f_actual))
            out = self._volResize("%s_FV%d" % (self.fvname, fvnum), \
                                     size_str)
        elif f_actual < f_alloc:
            out = self._volResize("%s_FV%d" % (self.fvname, fvnum), \
                                     size_str)
        elif decr and f_actual > f_alloc:
            size_str = "-%dk" % (na_size_estimate(long(size),self.multiplier)/1024)
            out = self._volResize("%s_FV%d" % (self.fvname, fvnum), \
                                     size_str)
        if out is not None and not na_test_result(out):
                raise xs_errors.XenError('SRNoSpace', opterr='Requested size %s. [%s]' % \
                                         (size_str,netapplib.na_results_reason(out)))
        return

    def _volResize(self, vol, delta):
        out = netapplib.volume_resize_wrapper(self.sv, str(delta), str(vol));
        if not na_test_result(out):
            util.SMlog("Vol resize failed [%s,%s]: %s" % (vol, delta, \
                                                          netapplib.na_results_reason(out)))
        else:
            util.SMlog("Vol resize succeeded, args [%s,%s]" % (vol, delta))
        return out
        
        


class NETAPPVDI(VDI.VDI):
    def load(self, vdi_uuid):
        if not self.sr.conn and self.sr.cmd == 'vdi_detach':
            return
        if not self.sr.exists:
            raise xs_errors.XenError('SRUnavailable')
        self.exists = False
        self.comment = True
	self.sm_config_keep.append('SCSIid')
        if len(self.sr.luns):
            for val in self.sr.luns:
                if match_XSFV(val.name,vdi_uuid):
                    self.exists = True
                    self.obj = val
                    self.size = self.obj.size
                    self.utilisation = self.obj.utilisation
                    self.snap = self.obj.snap
                    self.serial = self.obj.serial
                    self.sr.added_vdi(self)
                    if is_snap(val.comment) or is_clone(val.comment):
                        self.parent = val.comment.split('|')[2].replace("SRC:","")
                    return
        self.obj = self.sr._getLUN(vdi_uuid)
        if self.obj != 0:
            self.exists = True
            self.size = self.obj.size
            self.utilisation = self.obj.utilisation
            self.snap = self.obj.snap
            self.serial = self.obj.serial
            self.sr.added_vdi(self)
        
    def create(self, sr_uuid, vdi_uuid, size):
        if self.exists:
            raise xs_errors.XenError('VDIExists')
        self.sm_config = self.sr.srcmd.params['vdi_sm_config']
        # SCSIid may be present (eg during vm-import) - remove it, its stale.
        self.sm_config.pop('SCSIid',0)
        vmhash = random.randrange(0, self.sr.exists)
        if self.sm_config.has_key('vmhint'):
            if util.exactmatch_uuid(self.sm_config['vmhint']):
                vmhash = _uuid_to_long(self.sm_config['vmhint']) % self.sr.exists
            else:
                # Try to extract the hint as a number
                try:
                    vmhash = long(self.sm_config['vmhint']) % self.sr.exists
                except:
                    pass

        path = os.path.join("%s_FV%d" % (self.sr.remotepath, vmhash), vdi_uuid)
        if self.sr.allocation:
            self.sr._synch_vol_size(size, vmhash)
            res = "true"
        else:
            res = "false"
        sizebuf = "%d" % util.roundup(MIN_LUNSIZE, size)
        out = netapplib.lun_create_wrapper(self.sr.sv, path, sizebuf, res);
        if not na_test_result(out):
            errorstr = "creating VDI (size %s)" % sizebuf
            if long(netapplib.na_results_errno(out)) == LUN_SIZE_TOO_LARGE:
                try:
                    tmp = netapplib.lun_maxsize_wrapper(self.sr.sv, path)
                    errorstr += ": Max available size [%s]" \
                          % netapplib.na_child_get_string(out, "without-snapshot-reserve")
                except:
                    pass
            na_raise_error(out,self.sr.sv,errorstr)
        self.size = long(netapplib.na_child_get_string(out, "actual-size"));

        #Set the LUN comment field
        comment = "SRC|%s" % _getDateString()
        netapplib.lun_set_comment_wrapper(self.sr.sv, str(comment), path);
        self.serial = self.sr._getLUNSerial(path)
        self.sr.added_vdi(self)
        self.sr._updateSerial()
        self._db_introduce()
        self.sr.virtual_allocation += self.size
        self.sr.update(self.sr.uuid)
        return super(NETAPPVDI, self).get_params()

    def delete(self, sr_uuid, vdi_uuid):
        if not self.exists:
            return
        fvnum = self.obj.fvnum
        path = os.path.join("%s_FV%d" % (self.sr.remotepath, fvnum), self.obj.name)
    	out = netapplib.lun_delete_wrapper(self.sr.sv, path);
        if not na_test_result(out):
            na_raise_error(out,self.sr.sv,"deleting VDI")
        if self.sr.allocation and not self.snap:
            try:
                self.sr._synch_vol_size(self.obj.size, fvnum, decr=True)
            except:
                pass
        try:
            vdi_ref = self.sr.srcmd.params['vdi_ref']
            sm_config = self.session.xenapi.VDI.get_sm_config(vdi_ref)
            if sm_config.has_key('SCSIid'):
                key = "scsi-" + sm_config['SCSIid']
                util.SMlog("Path exists, removing key [%s]" % key)
                self.session.xenapi.SR.remove_from_sm_config(self.sr.sr_ref, key)
        except:
            pass
        self.sr.deleted_vdi(vdi_uuid)
        self.sr._updateSerial()
        self._db_forget()
        util.set_dirty(self.sr.session, self.sr.sr_ref)

    def attach(self, sr_uuid, vdi_uuid):
        if not self.exists:
            raise xs_errors.XenError('VDIUnavailable')
        if not EXCL_MAP:
            self.activate(sr_uuid, vdi_uuid)
        else:
            self.path = "%s/SERIAL-%s" % (self.sr.iscsi.path,self.serial.encode("hex"))
        return super(NETAPPVDI, self).attach(sr_uuid, vdi_uuid)

    def activate(self, sr_uuid, vdi_uuid):
        fvnum = self.obj.fvnum
        path = os.path.join("%s_FV%d" % (self.sr.remotepath, fvnum), self.obj.name)
        if EXCL_MAP:
            # Test for exclusive mapping
            LUNmap = self.sr._getLUNMap(path, True)
            util.SMlog("LUNMap returned: [%s](%d)" % (LUNmap,len(LUNmap)))
            if len(LUNmap) > 0:
                if len(LUNmap) == 1 and LUNmap[0] == self.sr.igroup:
                    pass
                else:
                    raise xs_errors.XenError('NAPPExclActivate')

	retries = 3
	while retries != 0:
            out = netapplib.lun_map_wrapper(self.sr.sv,str(self.sr.igroup),path)
            retries = retries - 1
            if not na_test_result(out):
                if long(netapplib.na_results_errno(out)) == INITGROUP_MAP_EXISTS or \
                       long(netapplib.na_results_errno(out)) == INITGROUP_MAP_EXISTS2:
                    self.LUNid = self.sr._getLUNid(path)
                    util.SMlog("LUN already assigned to igroup [%s,%s]" \
                           % (self.sr.igroup, self.LUNid))
                    break

                # It appears the netapp library is not thread safe for lun mapping - if we collide retry.
                elif long(netapplib.na_results_errno(out)) == INITGROUP_HAS_LUN and retries > 0:
                    pass

                else:
                    na_raise_error(out,self.sr.sv,"creating LUN map")
            else:
                self.LUNid = long(netapplib.na_child_get_string(out, "lun-id-assigned"))
                util.SMlog("LUN assigned to igroup [%s] with id [%d]" \
                       % (self.sr.igroup, self.LUNid))
                break

        util.SMlog("Attaching LUN with Comment: [%s]" % self.sr._getLUNComment(path))
        lunpaths = self.sr.iscsi._attach_LUN_bylunid(self.LUNid)
        if not len(lunpaths):
            util.SMlog("Unable to detect LUN attached to host; id %d" % self.LUNid)
            raise xs_errors.XenError('VDIUnavailable')
        self.path = lunpaths[0]
        vdi_ref = self.sr.srcmd.params['vdi_ref']
        
        # Verify the Serial path
        serialhex = self.serial.encode("hex")
        serialpaths = self.sr.iscsi._attach_LUN_byserialid(serialhex)
        if not len(serialpaths):
            util.SMlog("LUN path detected but not serial [%s,%s]" % \
                       (lunpaths, serialhex))
        if len(lunpaths) != len(serialpaths):
            util.SMlog("Serial to LUN Path inconsistencies: [%s,%s]" % (lunpaths, \
                                                                            serialpaths))
            raise xs_errors.XenError('VDIUnavailable', opterr='Path inconsistencies')
        for i in range(0, len(lunpaths)):
            if os.path.realpath(lunpaths[i]) != os.path.realpath(serialpaths[i]):
                util.SMlog("Serial to LUN Path inconsistencies: [%s,%s]" % (lunpaths, \
                                                                            serialpaths))
                raise xs_errors.XenError('VDIUnavailable', opterr='Path inconsistencies')
        
        self.realpath = os.path.realpath(self.path)
        self.serialpath = serialpaths[0]
        scsiutil.update_XS_SCSIdata(vdi_uuid, \
                                    scsiutil.gen_raw_page_data(self.path))
        self.SCSIid = scsiutil.getSCSIid(self.path)
        key = "scsi-" + self.SCSIid
        try:
            sr_sm_config = self.session.xenapi.SR.get_sm_config(self.sr.sr_ref)
            if not sr_sm_config.has_key(key):
                self.session.xenapi.SR.add_to_sm_config(self.sr.sr_ref, key, vdi_uuid)
            elif sr_sm_config[key] != vdi_uuid:
                util.SMlog("SR sm_config key %s has value %s, expected %s" % (key, sr_sm_config[key], vdi_uuid))
                raise
            cmd = ['/opt/xensource/sm/mpathcount.py',self.SCSIid]
            util.pread2(cmd)
        except:
            raise xs_errors.XenError('VDIUnavailable')

        try:
            vdi_sm_config = self.session.xenapi.VDI.get_sm_config(vdi_ref)
            if not vdi_sm_config.has_key("SCSIid"):
                self.session.xenapi.VDI.add_to_sm_config(vdi_ref, "SCSIid", self.SCSIid)
            elif vdi_sm_config["SCSIid"] != self.SCSIid:
                util.SMlog("VDI sm_config key SCSIid has value %s, expected %s" % (vdi_sm_config["SCSIid"], self.SCSIid))
                raise
        except:
            util.SMlog('Exception adding SCSIid, %s to sm_config' % self.SCSIid)
            raise xs_errors.XenError('VDIUnavailable')

        # Test for Multipath node if required
        self.path = self.sr.mpathmodule.path(self.SCSIid)
        if self.sr.mpath == 'true':
            self.sr.mpathmodule.refresh(self.SCSIid,len(lunpaths))
            if not os.path.exists(self.path):
                util.SMlog("Unable to detect Multipath node")
                raise xs_errors.XenError('VDIUnavailable', \
                        opterr='Multipath node unavailable (%s)' % self.path)
        
	# Set default scheduler
        self.other_config = self.session.xenapi.VDI.get_other_config(vdi_ref)
        if self.other_config.has_key('scheduler') and self.other_config['scheduler'] != self.sr.sched:
            self.sr.sched = self.other_config['scheduler']
        realdev = os.path.realpath(self.path).split('/')[-1]
        util.set_scheduler(realdev, self.sr.sched)

    def detach(self, sr_uuid, vdi_uuid):
        if self.sr.conn and not self.exists:
            raise xs_errors.XenError('VDIUnavailable')
        if not EXCL_MAP:
            self.deactivate(sr_uuid, vdi_uuid)        
        return

    def deactivate(self, sr_uuid, vdi_uuid):
        self.LUNid = -1
        self.SCSIid = ""
        path = ""
        try:
            vdi_ref = self.sr.srcmd.params['vdi_ref']
            sm_config = self.session.xenapi.VDI.get_sm_config(vdi_ref)
            if self.sr.conn:
                fvnum = self.obj.fvnum
                path = os.path.join("%s_FV%d" % (self.sr.remotepath, fvnum), self.obj.name)
                self.LUNid = self.sr._getLUNid(path)
                if sm_config.has_key('SCSIid'):
                    self.SCSIid = sm_config['SCSIid']
            else:
                if sm_config.has_key('SCSIid'):
                    devs = scsiutil._genReverseSCSidtoLUNidmap(sm_config['SCSIid'])
                    if len(devs):
                        for i in devs:
                            li = i.split(':')
                            li.insert(0,1)
                            scsiutil.scsi_dev_ctrl(li, "remove")
        except:
            pass

        if self.LUNid < 0:
            return
        
	if len(self.SCSIid) == 0:
            util.SMlog('Failed to get SCSIid, sm_config is %s' % str(sm_config))

        # Explicitly remove the LUN from the scsi bus
        self.sr.iscsi._detach_LUN_bylunid(self.LUNid, self.SCSIid)

        # Unmap the LUN on the target
        if self.sr.conn:
            out = netapplib.lun_unmap_wrapper(self.sr.sv,str(self.sr.igroup),path)
        return
    
    def resize(self, sr_uuid, vdi_uuid, size):
        if not self.exists:
            raise xs_errors.XenError('VDIUnavailable')
        fvnum = self.obj.fvnum
        current_size = self.obj.size

        # Disallow LUN shrinking
        if size < current_size:
            util.SMlog('vdi_resize: shrinking not supported: ' + \
                       '(current size: %d, new size: %d)' % (current_size, size))
            raise xs_errors.XenError('VDISize', opterr='shrinking not allowed')
        
        if size > current_size:
            diff = size - current_size
            if self.sr.allocation and self.sr.physical_size < \
                   (self.sr.virtual_allocation + diff):
                raise xs_errors.XenError('SRNoSpace')
        path = os.path.join("%s_FV%d" % (self.sr.remotepath, fvnum), self.obj.name)
        sizebuf = "%d" % size

        # Adjust the FV size
        if self.sr.allocation:
            prev = long(na_size_estimate(current_size, self.sr.multiplier))
            new = long(na_size_estimate(size,self.sr.multiplier))
            if new > prev:
                size_str = "+%dk" % long((new - prev)/1024)
            else:
                size_str = "-%dk" % long((prev - new)/1024)
            out = self.sr._volResize("%s_FV%d" % (self.sr.fvname, fvnum), \
                                      size_str)
            if not na_test_result(out):
                raise xs_errors.XenError('SRNoSpace', opterr='Requested size %s. [%s]' % \
                                         (size_str,netapplib.na_results_reason(out)))
        out = netapplib.lun_resize_wrapper(self.sr.sv, path, sizebuf);
        if not na_test_result(out):
            na_raise_error(out,self.sr.sv,"Resizing LUN")
        self.size = long(netapplib.na_child_get_string(out, "actual-size"))
        self.utilisation = self.size
        self._db_update()
        self.sr.virtual_allocation += (self.size - current_size)
        self.sr.update(self.sr.uuid)
        return super(NETAPPVDI, self).get_params()

    def _snapshot(self, sr_uuid, vdi_uuid, dest):
        # Used internally by clone and snapshot
        if not self.exists:
            raise xs_errors.XenError('VDIUnavailable')
        if self.sr.allocation and self.sr.physical_size < \
               (self.sr.virtual_allocation + self.size):
            raise xs_errors.XenError('SRNoSpace')
        fvnum = self.obj.fvnum
        fvname = "%s_FV%d" % (self.sr.fvname, fvnum)
        srcpath = os.path.join("%s_FV%d" % (self.sr.remotepath, fvnum), self.obj.name)
        dstpath = os.path.join("%s_FV%d" % (self.sr.remotepath, fvnum), dest)

        leaf_vdi = VDI.VDI(self.sr, dest) # user-visible leaf VDI
        leaf_vdi.read_only = False
        leaf_vdi.location = dest
        leaf_vdi.size = self.size
        leaf_vdi.utilisation = self.utilisation
        
        if not self.comment and self.sr.allocation:
            res = "true"
        else:
            res = "false"

        # Assign the SnapshotID
        SnapshotID = ''
        if self.sr.srcmd.params.has_key('driver_params'):
	    self.driver_param = self.sr.srcmd.params['driver_params']
            if self.driver_param.has_key('epochhint'):
                SnapshotID = self.driver_param['epochhint']
        if not SnapshotID:
            SnapshotID = util.gen_uuid()
        util.SMlog("Using SnapshotID: %s" % SnapshotID)
        self.sr._snapshot(fvname, SnapshotID)
        out = netapplib.lun_clone_wrapper(self.sr.sv, str(SnapshotID), srcpath, dstpath, res)
        if not na_test_result(out):
            na_raise_error(out,self.sr.sv,"Cloning LUN")
        if self.comment:
            comment = "SNAP|%s|SRC:%s|SNAPID:%s" % (_getDateString(),vdi_uuid,SnapshotID)
            netapplib.lun_set_comment_wrapper(self.sr.sv, str(comment), dstpath);
        util.SMlog("VDI snapshot completed, new VDI [%s] (snapID %s)" % (dest,SnapshotID))

        return leaf_vdi

    def snapshot(self, sr_uuid, vdi_uuid):
        dest = util.gen_uuid ()
        leaf_vdi = self._snapshot(sr_uuid, vdi_uuid, dest)
        
        try:
            leaf_vdi_ref = leaf_vdi._db_introduce()
            util.SMlog("vdi_snapshot: introduced VDI: %s" % leaf_vdi_ref)
        except Exception, e:
            util.SMlog("vdi_snapshot: caught error during VDI.db_introduce: %s" % (str(e)))
            raise
        self.sr.update(self.sr.uuid)
        return leaf_vdi.get_params()
            
    def clone(self, sr_uuid, vdi_uuid):
        dest = util.gen_uuid ()
        if not self.exists:
            raise xs_errors.XenError('VDIUnavailable')
        if self.sr.allocation:
            self.sr._synch_vol_size(self.obj.size, self.obj.fvnum)

        #Create a snapshot and then split the clone
        self.comment = False
        leaf_vdi = self._snapshot(sr_uuid, vdi_uuid, dest)

        dstpath = os.path.join("%s_FV%d" % (self.sr.remotepath, self.obj.fvnum), dest)
        out = netapplib.lun_clone_split_wrapper(self.sr.sv, str(dstpath));
        if not na_test_result(out):
            na_raise_error(out,self.sr.sv,"Splitting LUN")
        comment = "CLONE|%s|SRC:%s" % (_getDateString(),vdi_uuid)
        netapplib.lun_set_comment_wrapper(self.sr.sv, comment, dstpath);
        
        try:
            leaf_vdi.read_only = False
            leaf_vdi_ref = leaf_vdi._db_introduce()
            util.SMlog("vdi_clone: introduced VDI: %s" % leaf_vdi_ref)
        except Exception, e:
            util.SMlog("vdi_clone: caught error during VDI.db_introduce: %s" % (str(e)))
            raise
        self.sr.virtual_allocation += self.obj.size
        self.sr.update(self.sr.uuid)
        return leaf_vdi.get_params()


# NETAPP session handler functions
def na_test_result(out):
    if netapplib.na_results_status(out) != netapplib.NA_OK:
        return False
    return True
    
def na_raise_error(out, sv, st):
    #na_close_session(sv)
    code = netapplib.na_results_errno(out);
    reason = netapplib.na_results_reason(out);
    raise xs_errors.XenError('EIO', \
                             opterr='%s:%s (%d)' % (st,reason,code))

def na_open_session(server, username, password, port=0, secure=False):
    sv = netapplib.na_server_open( str(server), 1, 1);
    if port:
        netapplib.na_server_set_port(sv, port);
    netapplib.na_server_set_style(sv, netapplib.NA_STYLE_LOGIN_PASSWORD);
    netapplib.na_server_set_admin_user(sv, util.to_plain_string(username), \
                                       util.to_plain_string(password));

    if secure:
        netapplib.na_server_set_transport_type(sv, netapplib.NA_SERVER_TRANSPORT_HTTPS, str(0));
        util.SMlog("Set transport type to HTTPS")
        if not port:
            # No port override, set to default SSL port
            netapplib.na_server_set_port(sv, HTTPS_PORTNUM_DEFAULT);

    try:
        util._testHost(server, netapplib.na_server_get_port(sv), 'NAPPTargetFailed')
    except:
        na_close_session(sv)
        raise
        
    out = netapplib.na_server_invoke(sv, "system-get-version", NULL);
    if not na_test_result(out):
        raise xs_errors.XenError('NAPPAuthFailed')
    try:
        version = float(_sanitise_version(netapplib.na_child_get_string(out, "version")))
    except:
         raise xs_errors.XenError('NAPPUnsupportedVersion', \
               opterr='Unable to parse version string, Minimum supported %s' \
                                 % (repr(NAPP_VERSION)))       
    if (version < NAPP_VERSION):
        raise xs_errors.XenError('NAPPUnsupportedVersion', \
                                opterr='Detected: %s, Minimum supported %s' \
                                 % (str(version),repr(NAPP_VERSION)))
    return sv

def na_size_estimate(size, multiplier):
    # The additional estimated space required
    # for full provisioning is:
    # size * 2 * 20% for snap data
    # This multiplier can be adjusted via an
    # other-config key
    return size * multiplier

def _sanitise_version(str):
    for i in str.split():
        if match_version(i):
            i = re.sub("^[A-Z]","",i)
            i = re.sub(":","",i)
            i = re.sub("[A-Z].*$","",i)
            i = re.sub("[a-z]",".",i)
            l = i.split('.')
            return "%s.%s" % (l[0],l[1])
    raise xs_errors.XenError('NAPPUnsupportedVersion')

def na_close_session(handle):
    try:
        netapplib.na_server_close(handle);
    except:
        pass

def match_XS(s):
    regex = re.compile("^XenStorage_[0-9a-f]{8}_(([0-9a-f]{4})_){3}[0-9a-f]{12}_FV")
    return regex.search(s, 0)

def match_XSFV(s, FVname):
    if not len(s):
        return False
    regex = re.compile(FVname)
    return regex.search(s, 0)

def match_version(s):
    regex = re.compile("[0-9].[0-9]")
    return regex.search(s, 0)

def is_snap(s):
    regex = re.compile("^SNAP")
    return regex.search(s, 0)

def is_clone(s):
    regex = re.compile("^CLONE")
    return regex.search(s, 0)

def _targetIQNtoXML(target, iqn):
    dom = xml.dom.minidom.Document()
    element = dom.createElement("iscsi-target-iqns")
    dom.appendChild(element)

    entry = dom.createElement('TGT')
    element.appendChild(entry)

    subentry = dom.createElement('Index')
    entry.appendChild(subentry)           
    textnode = dom.createTextNode('0')
    subentry.appendChild(textnode)
            
    subentry = dom.createElement('IPAddress')
    entry.appendChild(subentry)           
    textnode = dom.createTextNode(target)
    subentry.appendChild(textnode)

    subentry = dom.createElement('TargetIQN')
    entry.appendChild(subentry)           
    textnode = dom.createTextNode(iqn)
    subentry.appendChild(textnode)
    
    return dom.toprettyxml()

def _aggregatetoXML(Aggrlist):
    dom = xml.dom.minidom.Document()
    driver = dom.createElement("Aggrlist")
    dom.appendChild(driver)

    for key in Aggrlist.keys():
        entry = dom.createElement("Aggr")
        driver.appendChild(entry)
        
        args = Aggrlist[key].split('|')
        e = dom.createElement("Name")
        entry.appendChild(e)
        textnode = dom.createTextNode(key)
        e.appendChild(textnode)

        e = dom.createElement("Size")
        entry.appendChild(e)
        textnode = dom.createTextNode(args[2])
        e.appendChild(textnode)

        e = dom.createElement("Disks")
        entry.appendChild(e)
        textnode = dom.createTextNode(args[3])
        e.appendChild(textnode)

        e = dom.createElement("RAIDType")
        entry.appendChild(e)
        textnode = dom.createTextNode(args[4])
        e.appendChild(textnode)

        e = dom.createElement("asis_dedup")
        entry.appendChild(e)
        textnode = dom.createTextNode(args[5])
        e.appendChild(textnode)
    return dom.toprettyxml()

def _getDateString():
    d = datetime.datetime.now()
    t = d.timetuple()
    return "%s-%s-%s:%s:%s:%s" % \
          (t[0],t[1],t[2],t[3],t[4],t[5]) 

def _uuid_to_long(uuid):
    return long(uuid.replace('-',''), 16)
    
class NETAPPFVOL(object):
    def __init__(self, num, base):
        self.num = num
        self.uuid = base.split('/')[:-1]
        self.path = "%s_FV%d" % (base,num)
        self.physical_size = 0
        self.physical_utilisation = 0
        self.virtual_allocation = 0

class NETAPPLUN(object):
    def __init__(self, num, base, name, size, snapLUN, cmt, serial):
        self.fvnum = num
        self.name = name
        self.size = size
        path = "%s_FV%d" % (base,num)
        self.path = os.path.join(path, name)
        self.utilisation = size
        self.parentsnap = ""
        self.snap = snapLUN
        self.comment = cmt
        self.serial = serial
        
if __name__ == '__main__':
    SRCommand.run(NETAPPSR, DRIVER_INFO)
else:
    SR.registerSR(NETAPPSR)
