#!/usr/bin/python
# Copyright (C) 2006-2007 XenSource Ltd.
# Copyright (C) 2008-2009 Citrix Ltd.
#
# This program is free software; you can redistribute it and/or modify 
# it under the terms of the GNU Lesser General Public License as published 
# by the Free Software Foundation; version 2.1 only.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
# GNU Lesser General Public License for more details.

import os
import pexpect
import xs_errors
import re
import time, util

MAX_SSH_SESSIONS_ERR = "Error: Number of open cli sessions reached maximum allowed value of 7."
EQL_VERSION = 4.0

class LoginException(Exception):
    def __init__(self, code, reason='login failed'):
        self.code = code
        Exception.__init__(self, code)

class SnmpException(Exception):
    def __init__(self, code, reason='Invalid SNMP response'):
        self.code = code
        Exception.__init__(self, code)

class EofException(Exception):
    def __init__(self, code, reason='Unexpected  EOF received'):
        self.code = code
        Exception.__init__(self, code)

class SessionCred:
    def __init__(self, username, host):
        self.lockfilename = "EQL-%s-%s" % (username,host)
        self.lock = "/var/lock/" + self.lockfilename
        self.pid = 0
        if os.path.exists(self.lock):
            for file in filter(self.match_statefile, util.listdir("/var/lock")):
                self.pid = long(file.split("-")[-1])
                break
        if not self.pid:
            self.pid = os.getpid()
        self.statefile = self.lock + "-" + str(self.pid)

    def match_statefile(self, s):
        regex = re.compile("^%s-" % self.lockfilename)
        return regex.search(s, 0)
            
        
class SSHSession:
    def __init__(self, conn_cred):
        self.host = conn_cred[0]
        self.user = conn_cred[1]
        self.password = conn_cred[2]
        cred = SessionCred(self.user,self.host)
        self.lock = cred.lock
        self.statefile = cred.statefile

    def connect(self):
        util.SMlog("EQL: Logging into array (%s)...." % self.host)
        for i in range(0,50):
            try:
                self.login()
                self.prep_term()
                util.SMlog("EQL: Logged into array (%s)." % self.host)
                return
            except LoginException, inst:
                if (inst.code == 'ConnectFail'):
                    util.SMlog("EQL: Connect failed, entering delay, retry #%d" % i)
                    time.sleep(5)
                if (inst.code == 'PermDenied'):
                    util.SMlog("EQL: Permission denied; username/password invalid")
                    raise xs_errors.XenError('EQLPermDenied')
            except:
                if i == 0:
                    util._testHost(self.host, 22, 'EQLTargetPort')
                util.SMlog("EQL: Unexpected error; could not connect")
                # In case the connection was opened and we failed afterwards
                self.close()
                time.sleep(10)
        util.SMlog("EQL: Max. attempts reached, giving up")
        raise xs_errors.XenError('EQLConnectfail')

    def login(self):
        while True:
            self.child = pexpect.spawn("/usr/bin/ssh %s@%s"% \
                                   (self.user, self.host), maxread = 4000)
            self.sshpid = self.child.pid
            index = self.child.expect(['password:', r'yes/no'], timeout=120)
            if index == 0:
                self.child.write(self.password+ "\r\n")
            elif index == 1:
                self.child.write("yes\r\n")
                self.child.expect("password:", timeout=120)
                self.child.write(self.password + "\r\n")

            index = self.child.expect(['[\S]+>', 'Error:[\s]+.*$', 'Permission denied,[\s]+.*$'], timeout = 120)
            if index == 1:
                if (self.child.after[:-2] ==  MAX_SSH_SESSIONS_ERR):
                    raise LoginException('ConnectFail')
            elif index == 2:
                    raise LoginException('PermDenied')
            else:
                break
        self.child.write("\r\n")
        self.child.expect('[\S]+>')
        self.prompt = self.child.after

    def prep_term(self):
        self.child.write("cli-settings events off\r\n")
        self.child.write("grpparams info-messages disable\r\n")
        self.child.write("stty hardwrap off\r\n")
        self.child.write("cli-settings formatoutput off\r\n")
        self.child.write("cli-settings paging off\r\n")
        self.child.write("cli-settings confirmation off\r\n")
        self.child.write("cli-settings displayinMB on\r\n")
        self.child.write("\r\n")
        self.child.expect('[\S]+>')

    def reset_term(self):
        self.child.write("cli-settings formatoutput on\r\n")
        self.child.write("cli-settings paging on\r\n")
        self.child.write("cli-settings confirmation on\r\n")
        self.child.write("cli-settings displayinMB off\r\n")
        self.child.write("\r\n")
        self.child.expect('[\S]+>')
        self.child.write("logout\r\n")
        self.child.expect("\(y/n\)", timeout=30)
        self.child.write("y\r\n")
        self.child.expect('[\S]+>[\s]+Connection[\s\S.]*closed[\s.]*$')

    def _cmd_exec(self, command):
        """send a command and return response"""
        util.SMlog("EQL: %s" % command)
        self.error = 0
        self.errmsg = ""
        self.output = []
        self.errlog = []
        cmdargv = command.split()
        cmdargc = len(cmdargv)
        self.child.write("\r\n")
        self.child.write(command + "\r\n")
        self.child.write("\r\n")
        cmd_pattern = self.prompt + " " + command
        initstr = ""
        regex = re.compile(cmd_pattern)
        while True:
            output = self.child.readline()
            if output == '':
                # EOF signal, possible ssh connection failure
                util.SMlog("EQL: Unexpected EOF signal, received")
                raise EofException('InvalidEOFRecv')
            if (re.match('^[0-9]+:[0-9]+:[\s\S.]*:INFO:[\s\S.]*.*', output)):
                #INFO output
                continue
            cmdstr = initstr + output[:-2]
            if (regex.search(cmdstr,0)):
                break
            else:
                initstr = cmdstr
                continue
        err = False
        while True:
            output = self.child.readline()
            if output == '':
                # EOF signal, possible ssh connection failure
                util.SMlog("EQL: Unexpected EOF signal, received")
                raise EofException('InvalidEOFRecv')
            if (re.match('^[0-9]+:[0-9]+:[\s\S.]*:INFO:[\s\S.]*.*', output)):
                #INFO output
                continue
            if (re.match('^[\s]*$', output)):
                #blank line
                continue
            if (re.match('^[\s]*(%\s+)?Error[:]?[\s]+[-][\s]+Invalid[\s]+SNMP[\s]+response[\s]+.*$', output)):
                try:
                    self.child.write("\n")
                except:
                    pass
                util.SMlog("EQL: Unexpected Invalid SNMP response received")
                raise SnmpException('InvalidSNMPResp')
            if (re.match('^[\s]*(%\s+)?Error[:]?[\s]+.*$', output)):
                self.error = 1
                self.errmsg = output[:-2]
                self.child.write("\n")
                err = True
            else:
                if (re.match(self.prompt, output)):
                    break
                else:
                    output1 = output[:-2]
                    if err:
                        self.errlog.append(output1)
                    else:
                        self.output.append(output1)
        rc = self.error
        errmsg = self.errmsg
        result = self.output
        errlog = self.errlog
        if (rc):
            util.SMlog("EQL: %s failed: %s" % (command, errmsg))
        return (rc, result, errlog, errmsg)

    def command(self, command):
        reconnect = False
        SNMPexception = False
        EOFexception = False
        for i in range(0,20):
            try:
                if reconnect:
                    util.SMlog("EQL: Reconnecting to the array.")
                    self.login()
                    self.prep_term()
                    util.SMlog("EQL: Logged into array")
                return self._cmd_exec(command)
            except SnmpException, inst:
                if (inst.code == 'InvalidSNMPResp'):
                    SNMPexception = True
                util.SMlog("EQL: CMD failed, entering delay, retry #%d" % i)
                #self.close()
                #reconnect = True
                time.sleep(10)
            except EofException, inst:
                if (inst.code == 'InvalidEOFRecv'):
                    EOFexception = True
                util.SMlog("EQL: CMD failed, entering delay, retry #%d" % i)
                self.close()
                reconnect = True
                time.sleep(10)
            except:
                util.SMlog("EQL: CMD failed, entering delay, retry #%d" % i)
                reconnect = True
                time.sleep(10)
        util.SMlog("EQL: Max. CMD retry attempts reached, giving up")
        if (SNMPexception):
            raise xs_errors.XenError('EQLInvalidSNMPResp')
        elif (EOFexception):
            raise xs_errors.XenError('EQLInvalidEOFRecv')
        else:
            raise xs_errors.XenError('EQLConnectfail')
    
    def close(self):
        try:
            self.reset_term()
        except:
            util.SMlog("EQL: Reset_term failed, try to close connection anyway")
            pass
        if os.path.exists(self.statefile):
            os.remove(self.statefile)
        try:
            self.child.close(True)
        except:
            pass
        util.SMlog("EQL: Logged out from array (%s)" % self.host)

def vol_exists(conn, name):
        command = "vol select %s show"% name
        (rc, output, errlog, errmsg) = conn.command(command)
        if (rc):
                return False
        else:
                return True

def vol_create(conn, name, size, allocate, uuid, state, snapreserve, snapdepletion, poolname):
        command = "vol create %s %s description %s pool %s %s snap-depletion %s snap-reserve %d" % \
                  (name, size, uuid, poolname, state, snapdepletion, snapreserve)
        command += "%"
        if (allocate == 'thin'):
            command += " thin-provision"

        (rc, output, errlog, errmsg) = conn.command(command)
        outmsg = ""
        iSCSIname = ""
        if rc != 1:
                outmsg = output[0]
                iSCSIstr = "".join(output[1:])
                #iSCSIname = re.search('^(iSCSI target name is)[\s]*(.*)$', iSCSIstr).groups()[1]
        else:
                return (rc, outmsg, iSCSIname, errlog, errmsg)
        command1 = "vol select %s multihost-access enable" % (name)
        (rc, output, errlog, errmsg) = conn.command(command1)
        return (rc, outmsg, iSCSIname, errlog, errmsg)

def vol_delete(conn, name):
        command = "vol delete %s"% name
        (rc, output, errlog, errmsg) = conn.command(command)
        outmsg = ""
        if rc != 1:
                outmsg = output[0]
        return (rc, outmsg, errlog, errmsg)

def snap_delete(conn, volname, snapname):
        command = "vol select %s snapshot delete %s"% (volname, snapname)
        (rc, output, errlog, errmsg) = conn.command(command)
        outmsg = ""
        if rc != 1:
                outmsg = output[0]
        return (rc, outmsg, errlog, errmsg)

def list_SR_vols(conn, vol_pattern_str):
        vol_list = []
        vol_temp_list = []
        command = "vol show"
        (rc, output, errlog, errmsg) = conn.command(command)
        prev_match = 0
        if rc != 1:
                for index in range(2, len(output)):
                    if (re.match('^[\s]{2}', output[index])):
                        if (prev_match == 1):
                            partial_item = output[index].split()[0]
                            temp_item = vol_list[-1] + partial_item
                            vol_list[-1] = temp_item
                            continue
                        else:
                            partial_item = output[index].split()[0]
                            temp_item = vol_temp_list[-1] + partial_item
                            vol_temp_list[-1] = temp_item
                            continue
                    items = output[index].split()
                    if not len(items):
                        continue
                    if (len(vol_temp_list) != 0):
                        if re.match(vol_pattern_str, vol_temp_list[-1]):
                            vol_list.append(vol_temp_list[-1])
                            vol_temp_list = [] 
                    if re.match(vol_pattern_str, items[0]):
                        prev_match = 1
                        vol_list.append(items[0])
                    else:
                        prev_match = 0
                        vol_temp_list.append(items[0])
                if (len(vol_temp_list) != 0):
                    if re.match(vol_pattern_str, vol_temp_list[-1]):
                        vol_list.append(vol_temp_list[-1])
        return (rc, vol_list, errlog, errmsg)

def list_VDI_snaps(conn, vdi_name, vdi_pattern):
        snap_list = []
        vdi_snap_list = []
        command = "vol select %s snapshot show"% vdi_name
        (rc, output, errlog, errmsg) = conn.command(command)
        prev_match = 0
        if rc != 1:
                for index in range(2, len(output)):
                    if (re.match('^[\s]{2}', output[index])):
                        partial_item = output[index].split()[0]
                        temp_item = snap_list[-1] + partial_item
                        snap_list[-1] = temp_item
                        continue
                    else:
                        items = output[index].split()
                        snap_list.append(items[0])
                for id in range(len(snap_list)):
                    if (re.match(vdi_pattern, snap_list[id])):
                        vdi_snap_list.append(snap_list[id])
        return (rc, vdi_snap_list, errlog, errmsg)

def vol_online(conn, volname):
        command = "vol select %s online"% volname
        (rc, output, errlog, errmsg) = conn.command(command)
        return (rc, errlog, errmsg)

def vol_offline(conn, volname):
        command = "vol select %s offline"% volname
        (rc, output, errlog, errmsg) = conn.command(command)
        return (rc, errlog, errmsg)

def vol_set_deleted(conn, volname, voldesc):
        command = "vol select %s description %s"% (volname, voldesc)
        (rc, output, errlog, errmsg) = conn.command(command)
        return (rc, errlog, errmsg)

def snap_online(conn, volname, snapname):
        command = "vol select %s snapshot select %s online"% (volname, snapname)
        (rc, output, errlog, errmsg) = conn.command(command)
        return (rc, errlog, errmsg)

def snap_offline(conn, volname, snapname):
        command = "vol select %s snapshot select %s offline"% (volname, snapname)
        (rc, output, errlog, errmsg) = conn.command(command)
        return (rc, errlog, errmsg)

def vol_detail(conn, volname):
        command = "vol select %s show"% volname
        (rc, output, errlog, errmsg) = conn.command(command)
        vol_info = {}
        if not rc:
            for id in range(1, len(output)):
                if (re.match('^[_]+$', output[id])):
                    break
                else:
                    temp = output[id].split(":", 1)
                    if len(temp) > 1:
                        vol_info[temp[0]] = temp[1].lstrip()
        return (rc, vol_info, errlog, errmsg)

def snap_detail(conn, volname, snapname):
        command = "vol select %s snapshot select %s show"% (volname, snapname)
        (rc, output, errlog, errmsg) = conn.command(command)
        snap_info = {}
        for id in range(1, len(output)):
                if (re.match('^[_]+$', output[id])):
                        break
                else:
                        temp = output[id].split(":", 1)
                        if len(temp) > 1:
                            snap_info[temp[0]] = temp[1].lstrip()
        return (rc, snap_info, errlog, errmsg)

def snap_conn_detail(conn, volname, snapname):
        command = "vol select %s snapshot select %s show connections"% (volname, snapname)
        (rc, output, errlog, errmsg) = conn.command(command)
        snap_conn_info = output[2:]
        return (rc, snap_conn_info, errlog, errmsg)

def vol_create_access_rec_by_initiator(conn, volname, initiatorname):
        id_list = []
        initiator_list = []
        initiator_id = 0
        command = "vol select %s access show"% volname
        (rc, output, errlog, errmsg) = conn.command(command)
        if rc != 1:
                for index in range(2, len(output)):
                    if (re.match('^[\s]+$', output[index])):
                        continue
                    if (re.match('^[\s]{6}', output[index])):
                        partial_item = output[index].split()[0]
                        temp_item = initiator_list[-1] + partial_item
                        initiator_list[-1] = temp_item
                        continue
                    else:
                        items = output[index].split()
                        if (len(id_list) != 0):
                            if re.match(initiatorname, initiator_list[0]):
                                initiator_id = id_list[0]
                                break
                            else:
                                id_list = []
                                initiator_list = [] 
                        id_list.append(items[0])
                        initiator_list.append(items[1])

                if (len(id_list) != 0):
                    if re.match(initiatorname, initiator_list[0]):
                        initiator_id = id_list[0]
        else:
            return (rc, output, errlog, errmsg)

        if (initiator_id != 0):
            return (rc, output, errlog, errmsg)
        else:
            command = "vol select %s access create initiator %s"% (volname, initiatorname)
            (rc, output, errlog, errmsg) = conn.command(command)
            return (rc, output, errlog, errmsg)

def vol_delete_access_rec_by_initiator(conn, volname, initiator):
        id_list = []
        initiator_list = []
        id_to_remove = 0
        command = "vol select %s access show"% volname
        (rc, output, errlog, errmsg) = conn.command(command)
        if rc != 1:
                for index in range(2, len(output)):
                    if (re.match('^[\s]+$', output[index])):
                        continue
                    if (re.match('^[\s]{6}', output[index])):
                        partial_item = output[index].split()[0]
                        temp_item = initiator_list[-1] + partial_item
                        initiator_list[-1] = temp_item
                        continue
                    else:
                        items = output[index].split()
                        if (len(id_list) != 0):
                            if re.match(initiator, initiator_list[0]):
                                id_to_remove = id_list[0]
                                break
                            else:
                                id_list = []
                                initiator_list = [] 
                        id_list.append(items[0])
                        initiator_list.append(items[1])

                if (len(id_list) != 0):
                    if re.match(initiator, initiator_list[0]):
                        id_to_remove = id_list[0]
        else:
            return (rc, output, errlog, errmsg)

        if (id_to_remove == 0):
            return (rc, output, errlog, errmsg)
        else:
            command = "vol select %s access delete %s"% (volname, id_to_remove)
            (rc, output, errlog, errmsg) = conn.command(command)
            return (rc, output, errlog, errmsg)

def vol_create_snapshot(conn, volname, sr_uuid, snap_uuid, description):
        command = "vol select %s snapshot create-now description %s"% (volname, description)
        (rc, output, errlog, errmsg) = conn.command(command)
        outmsg = ""
        snapname = ""
        if rc != 1:
                outmsg = output[0]
                snapstr = "".join(output[1:])
                robj = re.search('^(Snapshot name is)[\s]*(.*)$', snapstr)
                if not robj:
                    raise xs_errors.XenError('VDISnapshot')
                snapname = robj.groups()[1]
        else:
            return (rc, errlog, errmsg)
        new_snap_name= "XenStorage" + sr_uuid.replace("-", "") + snap_uuid.replace("-", "")[:20]
        command1 = "vol select %s snapshot rename %s %s"% (volname, snapname, new_snap_name)
        (rc, output, errlog, errmsg) = conn.command(command1)
        if rc == 1:
            return (rc, errlog, errmsg)
        else:
            command1 = "vol select %s snapshot select %s multihost-access enable" % (volname, new_snap_name)
            (rc, output, errlog, errmsg) = conn.command(command1)
        return (rc, errlog, errmsg)



def vol_create_clone_from_snapshot(conn, volname, snap_name, sr_uuid, clone_uuid, clone_desc, snapreserve, snapdepletion):
        clone_name= "XenStorage" + sr_uuid.replace("-", "") + clone_uuid.replace("-", "")[:20]
        command = "vol select %s snapshot select %s clone %s description %s offline snap-depletion %s snap-reserve %d"% (volname, snap_name, clone_name, clone_desc, snapdepletion, snapreserve)
        command += "%"
        (rc, output, errlog, errmsg) = conn.command(command)
        outmsg = ""
        iscsiname = ""
        if rc != 1:
                outmsg = output[0]
                iscsistr = output[1]
                #iscsiname = re.search('^(iSCSI target name is)[\s]*(.*)$', iscsistr).groups()[1]
        else:
                return (rc, outmsg, iscsiname, errlog, errmsg)
        command1 = "vol select %s multihost-access enable" % (clone_name)
        (rc, output, errlog, errmsg) = conn.command(command1)
        return (rc, outmsg, iscsiname, errlog, errmsg)

def vol_create_clone(conn, volname, sr_uuid, clone_uuid, clone_desc, snapreserve, snapdepletion):
        clone_name= "XenStorage" + sr_uuid.replace("-", "") + clone_uuid.replace("-", "")[:20]
        command = "vol select %s clone %s description %s offline snap-depletion %s snap-reserve %d"% (volname, clone_name, clone_desc, snapdepletion, snapreserve)
        command += "%"
        (rc, output, errlog, errmsg) = conn.command(command)
        outmsg = ""
        iscsiname = ""
        if rc != 1:
                outmsg = output[0]
                iscsistr = output[1]
                #iscsiname = re.search('^(iSCSI target name is)[\s]*(.*)$', iscsistr).groups()[1]
        else:
                return (rc, outmsg, iscsiname, errlog, errmsg)
        command1 = "vol select %s multihost-access enable" % (clone_name)
        (rc, output, errlog, errmsg) = conn.command(command1)
        return (rc, outmsg, iscsiname, errlog, errmsg)

def vol_shrink(conn, volname, new_size):
        command = "vol select %s shrink %s no-snap"% (volname, new_size)
        (rc, output, errlog, errmsg) = conn.command(command)
        return (rc, errlog, errmsg)

def vol_grow(conn, volname, new_size):
        command = "vol select %s size %s no-snap"% (volname, new_size)
        (rc, output, errlog, errmsg) = conn.command(command)
        return (rc, errlog, errmsg)

def pool_detail(conn, poolname):
        command = "pool select %s show"% poolname
        (rc, output, errlog, errmsg) = conn.command(command)
        pool_info = {}
        for id in range(1, len(output)):
                if (re.match('^[_]+$', output[id])):
                        break
                else:
                        temp = output[id].split(":", 1)
                        pool_info[temp[0]] = temp[1].lstrip()
        return (rc, pool_info, errlog, errmsg)

def all_StoragePools(conn):
        command = "pool show"
        (rc, output, errlog, errmsg) = conn.command(command)
        pool_info = {}
        for id in range(2, len(output)):
            entry_list = []
            entry_list = output[id].split()
            member_str = ''
            for index in range(1, len(entry_list)):
                if index == (len(entry_list) -1):
                    member_str = member_str + entry_list[index]
                else:
                    member_str = member_str + entry_list[index] + "|"
            pool_info[entry_list[0]] = member_str
        return (rc, pool_info, errlog, errmsg)

def check_version(conn):
        fw_supported = False
        member_list = []
        command = "member show"
        (rc, output, errlog, errmsg) = conn.command(command)
        prev_match = 0
        if rc != 1:
                for index in range(2, len(output)):
                    if (re.match('^[\s]{2}', output[index])):
                        continue
                    items = output[index].split()
                    if not len(items) > 2:
                        continue
                    if (re.match('^V', items[2])):
                        try:
                            version = _sanitise_version(items[2])
                            util.SMlog("EQL: Array version [%.1f (%s)]" % (version,items[2]))
                            if not version >= EQL_VERSION:
                                return False
                            else:
                                fw_supported = True
                        except:
                            util.SMlog("Exception in version check")
                            return False
        else:
            util.SMlog("EQL: version check rc=1")
        if not fw_supported:
            util.SMlog("EQL: version check text '%s'" % (str(output)))
        return fw_supported

def grpparams_detail(conn):
        command = "show grpparams"
        (rc, output, errlog, errmsg) = conn.command(command)
        grpparams_info = {}
        if not rc:
            for id in range(1, len(output)):
                if (re.match('^[_]+$', output[id])):
                    break
                else:
                    temp = output[id].split(":", 1)
                    if len(temp) > 1:
                        grpparams_info[temp[0]] = temp[1].lstrip()
        return (rc, grpparams_info, errlog, errmsg)

def _sanitise_version(str):
    str = re.sub("^V","",str)
    arr = str.split(".")
    return float("%s.%s" % (arr[0],arr[1]))
